/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/box.c,v 1.2 90/05/16 08:30:24 william Exp $";
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "color.e"
#include "cursor.e"
#include "file.e"
#include "grid.e"
#include "obj.e"
#include "poly.e"
#include "pattern.e"
#include "raster.e"
#include "ruler.e"
#include "select.e"
#include "setup.e"

int	boxDrawn = FALSE;

static XPoint bv[5];

void MyBox (window, gc, x1, y1, x2, y2)
   /* Hollow box, solid outline with width=1 */
   Window	window;
   GC		gc;
   int		x1, y1, x2, y2;
{
   bv[0].x = (short)x1; bv[0].y = (short)y1;
   bv[1].x = (short)x1; bv[1].y = (short)y2;
   bv[2].x = (short)x2; bv[2].y = (short)y2;
   bv[3].x = (short)x2; bv[3].y = (short)y1;
   bv[4].x = (short)x1; bv[4].y = (short)y1;
   XDrawLines (mainDisplay, window, gc, bv, 5, CoordModeOrigin);
}
 
void MySolidBox (window, gc, x1, y1, x2, y2, pixel)
   Window	window;
   GC		gc;
   int		x1, y1, x2, y2, pixel;
{
   XGCValues	values;

   bv[0].x = (short)x1; bv[0].y = (short)y1;
   bv[1].x = (short)x1; bv[1].y = (short)y2;
   bv[2].x = (short)x2; bv[2].y = (short)y2;
   bv[3].x = (short)x2; bv[3].y = (short)y1;
   bv[4].x = (short)x1; bv[4].y = (short)y1;

   values.foreground = pixel;
   values.function = GXcopy;
   values.fill_style = FillSolid;
   XChangeGC (mainDisplay, gc,
      GCForeground | GCFunction | GCFillStyle, &values);

   XFillPolygon (mainDisplay, window, gc, bv, 5, Convex, CoordModeOrigin);
}

void DumpBoxObj (FP, ObjPtr)
   FILE			* FP;
   struct ObjRec	* ObjPtr;
{
   register int	ltx, lty, rbx, rby;
   int		fill, width, pen;

   ltx = ObjPtr->obbox.ltx;
   lty = ObjPtr->obbox.lty;
   rbx = ObjPtr->obbox.rbx;
   rby = ObjPtr->obbox.rby;

   fill = ObjPtr->detail.b->fill;
   pen = ObjPtr->detail.b->pen;
   width = ObjPtr->detail.b->width;

   if (fill == NONEPAT && pen == NONEPAT) return;

   switch (fill)
   {
      case NONEPAT: break;
      case SOLIDPAT:
         /* solid black object */
         fprintf (FP, "newpath\n   %1d %1d moveto\n", ltx, lty);
         fprintf (FP, "   %1d %1d lineto\n", rbx, lty);
         fprintf (FP, "   %1d %1d lineto\n", rbx, rby);
         fprintf (FP, "   %1d %1d lineto\n", ltx, rby);
         fprintf (FP, "closepath fill\n");
         break;
      case BACKPAT:
         /* solid white object */
         fprintf (FP, "newpath\n   %1d %1d moveto\n", ltx, lty);
         fprintf (FP, "   %1d %1d lineto\n", rbx, lty);
         fprintf (FP, "   %1d %1d lineto\n", rbx, rby);
         fprintf (FP, "   %1d %1d lineto\n", ltx, rby);
         fprintf (FP, "closepath 1 setgray fill 0 setgray\n");
         break;
      default:
         /* patterned */
         fprintf (FP, "gsave\n");
         fprintf (FP, "   pat%1d 8 1 0 72 300 32 div div setpattern\n", fill);
         fprintf (FP, "   newpath\n      %1d %1d moveto\n", ltx, lty);
         fprintf (FP, "      %1d %1d lineto\n", rbx, lty);
         fprintf (FP, "      %1d %1d lineto\n", rbx, rby);
         fprintf (FP, "      %1d %1d lineto\n", ltx, rby);
         fprintf (FP, "   closepath fill\n");
         fprintf (FP, "grestore\n");
         break;
   }

   fprintf (FP, "%1d setlinewidth\n", widthOfLine[width]);

   switch (pen)
   {
      case NONEPAT: break;
      case SOLIDPAT:
         /* solid black object */
         fprintf (FP, "newpath\n   %1d %1d moveto\n", ltx, lty);
         fprintf (FP, "   %1d %1d lineto\n", rbx, lty);
         fprintf (FP, "   %1d %1d lineto\n", rbx, rby);
         fprintf (FP, "   %1d %1d lineto\n", ltx, rby);
         fprintf (FP, "closepath stroke\n");
         break;
      case BACKPAT:
         /* solid white object */
         fprintf (FP, "newpath\n   %1d %1d moveto\n", ltx, lty);
         fprintf (FP, "   %1d %1d lineto\n", rbx, lty);
         fprintf (FP, "   %1d %1d lineto\n", rbx, rby);
         fprintf (FP, "   %1d %1d lineto\n", ltx, rby);
         fprintf (FP, "closepath 1 setgray stroke 0 setgray\n");
         break;
      default:
         /* patterned */
         fprintf (FP, "gsave\n");
         fprintf (FP, "   pat%1d 8 1 0 72 300 32 div div setpattern\n", pen);
         fprintf (FP, "   newpath\n      %1d %1d moveto\n", ltx, lty);
         fprintf (FP, "      %1d %1d lineto\n", rbx, lty);
         fprintf (FP, "      %1d %1d lineto\n", rbx, rby);
         fprintf (FP, "      %1d %1d lineto\n", ltx, rby);
         fprintf (FP, "   closepath stroke\n");
         fprintf (FP, "grestore\n");
         break;
   }
   fprintf (FP, "1 setlinewidth\n\n");
}

void DrawBoxObj (win, XOff, YOff, ObjPtr)
   Window		win;
   struct ObjRec	* ObjPtr;
{
   struct BoxRec	* box_ptr = ObjPtr->detail.b;
   int			fill, pen, pixel, func, ltx, lty, rbx, rby, width;
   int			real_x_off, real_y_off;
   char			s[80];
   XGCValues		values;

   pen = ObjPtr->detail.b->pen;
   fill = ObjPtr->detail.b->fill;
   width = ObjPtr->detail.b->width;
   pixel = colorPixels[ObjPtr->color];

   real_x_off = (XOff >> zoomScale) << zoomScale;
   real_y_off = (YOff >> zoomScale) << zoomScale;
   ltx = (ObjPtr->obbox.ltx - real_x_off) >> zoomScale;
   lty = (ObjPtr->obbox.lty - real_y_off) >> zoomScale;
   rbx = (ObjPtr->obbox.rbx - real_x_off) >> zoomScale;
   rby = (ObjPtr->obbox.rby - real_y_off) >> zoomScale;
   func = GXcopy;

   if (fill != 0)
   {
      values.foreground = (fill == 2) ? myBlackPixel : pixel;
      values.function = GXcopy;
      values.fill_style = FillOpaqueStippled;
      values.stipple = patPixmap[fill];
      XChangeGC (mainDisplay, drawGC,
            GCForeground | GCFunction | GCFillStyle | GCStipple, &values);
      XFillRectangle (mainDisplay, win, drawGC, ltx, lty, rbx-ltx, rby-lty);
   }

   if (pen != 0)
   {
      values.foreground = (pen == 2) ? myBlackPixel : pixel;
      values.function = GXcopy;
      values.fill_style = FillOpaqueStippled;
      values.stipple = patPixmap[pen];
      values.line_width = widthOfLine[width] >> zoomScale;
      XChangeGC (mainDisplay, drawGC,
            GCForeground | GCFunction | GCFillStyle | GCStipple | GCLineWidth,
            &values);
      XDrawRectangle (mainDisplay, win, drawGC, ltx, lty, rbx-ltx, rby-lty);
   }
}

static
void CreateBoxObj (X1, Y1, X2, Y2)
   int	X1, Y1, X2, Y2;
{
   struct BoxRec	* box_ptr;
   struct ObjRec	* obj_ptr;
   int			w, ltx, lty, rbx, rby;

   box_ptr = (struct BoxRec *) calloc (1, sizeof(struct BoxRec));
   box_ptr->fill = objFill;
   box_ptr->width = lineWidth;
   box_ptr->pen = penPat;

   obj_ptr = (struct ObjRec *) calloc (1, sizeof(struct ObjRec));

   if (X1 < X2)
      if (Y1 < Y2)
      {
         ltx = X1; lty = Y1; rbx = X2; rby = Y2;
      }
      else
      {
         ltx = X1; lty = Y2; rbx = X2; rby = Y1;
      }
   else
      if (Y1 < Y2)
      {
         ltx = X2; lty = Y1; rbx = X1; rby = Y2;
      }
      else
      {
         ltx = X2; lty = Y2; rbx = X1; rby = Y1;
      }

   obj_ptr->bbox.ltx = obj_ptr->obbox.ltx = obj_ptr->x = (ltx << zoomScale) +
         drawOrigX;
   obj_ptr->bbox.lty = obj_ptr->obbox.lty = obj_ptr->y = (lty << zoomScale) +
         drawOrigY;
   obj_ptr->bbox.rbx = obj_ptr->obbox.rbx = (rbx << zoomScale) + drawOrigX;
   obj_ptr->bbox.rby = obj_ptr->obbox.rby = (rby << zoomScale) + drawOrigY;
   w = widthOfLine[lineWidth];
   obj_ptr->bbox.ltx -= w;
   obj_ptr->bbox.lty -= w;
   obj_ptr->bbox.rbx += w;
   obj_ptr->bbox.rby += w;
   obj_ptr->type = OBJ_BOX;
   obj_ptr->color = colorIndex;
   obj_ptr->detail.b = box_ptr;
   AddObj (NULL, topObj, obj_ptr);
}
 
static
void ContinueBox (OrigX, OrigY)
int 	OrigX, OrigY;
{
   int 		end_x, end_y, grid_x, grid_y, saved_x, saved_y;
   int 		done = FALSE;
   int		pixel, xor_pixel;
   XGCValues	values;
   XEvent	input;
   XButtonEvent	* button_ev;
   XMotionEvent	* motion_ev;

   pixel = colorPixels[colorIndex];
   xor_pixel = xorColorPixels[colorIndex];

   values.foreground = xor_pixel;
   values.function = GXxor;
   values.fill_style = FillSolid;
   values.line_width = 0;

   XChangeGC (mainDisplay, drawGC,
         GCForeground | GCFunction | GCFillStyle | GCLineWidth, &values);

   saved_x = grid_x = OrigX;
   saved_y = grid_y = OrigY; 
   XGrabPointer (mainDisplay, drawWindow, FALSE,
         PointerMotionMask | ButtonReleaseMask,
         GrabModeAsync, GrabModeAsync, None, handCursor, CurrentTime);
   
   while (!done)
   {
      XNextEvent (mainDisplay, &input);
      if (input.type == ButtonRelease)
      {
         XUngrabPointer (mainDisplay, CurrentTime);
         MyBox (drawWindow, drawGC, OrigX, OrigY, saved_x, saved_y);
         done = TRUE;
      }
      else if (input.type == MotionNotify)
      {
         motion_ev = &(input.xmotion);
         end_x = motion_ev->x;
         end_y = motion_ev->y;
         GridXY (end_x, end_y, &grid_x, &grid_y);
         if (grid_x != saved_x || grid_y != saved_y)
         {
            MyBox (drawWindow, drawGC, OrigX, OrigY, saved_x, saved_y);
            saved_x = grid_x;
            saved_y = grid_y;
            MyBox (drawWindow, drawGC, OrigX, OrigY, saved_x, saved_y);
         }
         MarkRulers (grid_x, grid_y);
      }
   }
   if (OrigX != grid_x && OrigY != grid_y)
   {
      CreateBoxObj (OrigX, OrigY, grid_x, grid_y);
      DrawBoxObj (drawWindow, drawOrigX, drawOrigY, topObj);
      boxDrawn = TRUE;
      fileModified = TRUE;
   }
}

void DrawBox (input)
   XEvent		* input;
{
   XButtonEvent	* button_ev;
   Window	root_win, child_win;
   int		root_x, root_y, mouse_x, mouse_y, grid_x, grid_y;
   unsigned int	status;

   if (input->type != ButtonPress) return;

   button_ev = &(input->xbutton);
   if (button_ev->button == Button1)
   {
      XQueryPointer (mainDisplay, drawWindow, &root_win, &child_win,
            &root_x, &root_y, &mouse_x, &mouse_y, &status);
      GridXY (mouse_x, mouse_y, &grid_x, &grid_y);
      ContinueBox (grid_x, grid_y);
   }
}

void SaveBoxObj (FP, ObjPtr)
   FILE			* FP;
   struct ObjRec	* ObjPtr;
{
   fprintf (FP, "box(%s,", colorMenuItems[ObjPtr->color]);
   fprintf (FP, "%1d,%1d,%1d,%1d,%1d,%1d,%1d)", ObjPtr->obbox.ltx,
         ObjPtr->obbox.lty, ObjPtr->obbox.rbx, ObjPtr->obbox.rby,
         ObjPtr->detail.b->fill, ObjPtr->detail.b->width, ObjPtr->detail.b->pen);
}

void ReadBoxObj (Inbuf, ObjPtr)
   char			* Inbuf;
   struct ObjRec	* * ObjPtr;
{
   struct BoxRec	* box_ptr;
   char			color_str[20], * s;
   int			ltx, lty, rbx, rby, fill, width, pen, w;

   *ObjPtr = (struct ObjRec *) calloc (1, sizeof(struct ObjRec));
   s = FindChar ('(', Inbuf);
   s = ParseStr (s, ',', color_str);
   box_ptr = (struct BoxRec *) calloc (1, sizeof(struct BoxRec));

   if (fileVersion <= 5)
   {
      sscanf (s, "%d , %d , %d , %d , %d , %d , %d",
            &ltx, &lty, &rbx, &rby, &fill, &width, &pen);
      switch (width)
      {
         case 1: width = 3; break;
         case 2: width = 6; break;
      }
   }
   else
      sscanf (s, "%d , %d , %d , %d , %d , %d , %d",
            &ltx, &lty, &rbx, &rby, &fill, &width, &pen);

   box_ptr->fill = fill;
   box_ptr->width = width;
   box_ptr->pen = pen;
   (*ObjPtr)->x = ltx;
   (*ObjPtr)->y = lty;
   (*ObjPtr)->color = FindColorIndex (color_str);
   (*ObjPtr)->type = OBJ_BOX;
   (*ObjPtr)->obbox.ltx = ltx;
   (*ObjPtr)->obbox.lty = lty;
   (*ObjPtr)->obbox.rbx = rbx;
   (*ObjPtr)->obbox.rby = rby;
   w = widthOfLine[width];
   (*ObjPtr)->bbox.ltx = ltx - w;
   (*ObjPtr)->bbox.lty = lty - w;
   (*ObjPtr)->bbox.rbx = rbx + w;
   (*ObjPtr)->bbox.rby = rby + w;
   (*ObjPtr)->detail.b = box_ptr;
}

void FreeBoxObj (ObjPtr)
   struct ObjRec	* ObjPtr;
{
   cfree (ObjPtr->detail.b);
   cfree (ObjPtr);
}
