/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/choice.c,v 1.2 90/05/07 15:15:04 william Exp $";
#endif

#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "align.e"
#include "box.e"
#include "color.e"
#include "cursor.e"
#include "font.e"
#include "grid.e"
#include "mark.e"
#include "msg.e"
#include "oval.e"
#include "pattern.e"
#include "poly.e"
#include "polygon.e"
#include "raster.e"
#include "select.e"
#include "setup.e"
#include "text.e"

int	curChoice = NOTHING;

static GC	choiceGC;
static Pixmap	choiceBackingPixmap;

void InitChoice ()
{
   XGCValues	values;

   values.foreground = myWhitePixel;
   values.background = myBlackPixel;
   values.font = defaultFontPtr->fid;
   choiceGC = XCreateGC (mainDisplay, choiceWindow,
         GCForeground | GCBackground | GCFont, &values);

   choiceBackingPixmap = XCreatePixmap (mainDisplay, choiceWindow,
         choiceWindowW, choiceWindowH, mainDepth);
}

void CleanUpChoices ()
{
   XFreeGC (mainDisplay, choiceGC);
   XFreePixmap (mainDisplay, choiceBackingPixmap);
}

void ShowMode ()
{
   XSetStipple (mainDisplay, rasterGC, choicePixmap[curChoice]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, 0, 0, choiceImageW,
         choiceImageH);
}

void ShowColor ()
{
   XGCValues	values;

   if (colorDisplay)
   {
      values.foreground = colorPixels[colorIndex];
      values.function = GXcopy;
      values.fill_style = FillOpaqueStippled;
      values.stipple = patPixmap[1];
      XChangeGC (mainDisplay, patGC,
            GCForeground | GCFunction | GCFillStyle | GCStipple, &values);

      XFillRectangle (mainDisplay, choiceWindow, patGC,
            0, choiceImageH, choiceImageW, choiceImageH);
   }
}

void ShowHoriAlign ()
{
   XSetStipple (mainDisplay, rasterGC, alignHoriPixmap[horiAlign]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, choiceImageW, 0,
         choiceImageW, choiceImageH);
}

void ShowVertAlign ()
{
   XSetStipple (mainDisplay, rasterGC, alignVertPixmap[vertAlign]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, choiceImageW,
         choiceImageH, choiceImageW, choiceImageH);
}

void ShowJust ()
{
   XSetStipple (mainDisplay, rasterGC, justPixmap[textJust]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, 2*choiceImageW, 0,
         choiceImageW, choiceImageH);
}

void ShowCurFont ()
{
   register int	i, j;
   int		x, y, pixel, w, h, saved_x, saved_y;
   XRectangle	recs[1];
   XImage	* image;

   recs[0].x = 2*choiceImageW;
   recs[0].y = choiceImageH;
   recs[0].width = choiceImageW;
   recs[0].height = choiceImageH;

   w = XTextWidth (canvasFontPtr, "W", 1);
   h = canvasFontHeight;

   switch (curRotate)
   {
      case ROTATE0:
         x = (w > choiceImageW) ?
               2*choiceImageW - (w-choiceImageW)/2 :
               2*choiceImageW + (choiceImageW-w)/2;
         y = (h > choiceImageH) ?
               choiceImageH + (canvasFontAsc-(h-choiceImageH)/2) :
               choiceImageH + (canvasFontAsc+(choiceImageH-h)/2);
         break;
      case ROTATE90:
         x = (h > choiceImageW) ?
               2*choiceImageW + h - (h-choiceImageW)/2 :
               2*choiceImageW + h + (choiceImageW-h)/2;
         y = (w > choiceImageH) ?
               choiceImageH - (w-choiceImageH)/2 :
               choiceImageH + (choiceImageH-w)/2;
         break;
      case ROTATE180:
         x = (w > choiceImageW) ?
               2*choiceImageW + w - (w-choiceImageW)/2 :
               2*choiceImageW + w + (choiceImageW-w)/2;
         y = (h > choiceImageH) ?
               2*choiceImageH + (h-choiceImageH)/2 :
               2*choiceImageH - (choiceImageH-h)/2;
         break;
      case ROTATE270:
         x = (h > choiceImageW) ?
               2*choiceImageW - (h-choiceImageW)/2 :
               2*choiceImageW + (choiceImageW-h)/2;
         y = (w > choiceImageH) ?
               2*choiceImageH + (w-choiceImageH)/2 :
               2*choiceImageH - (choiceImageH-w)/2;
         break;
   }

   XClearArea (mainDisplay, choiceWindow, 2*choiceImageW, choiceImageH,
         choiceImageW, choiceImageH, FALSE);

   XSetFont (mainDisplay, choiceGC, canvasFontPtr->fid);
   if (curRotate == 0)
   {
      XSetClipRectangles (mainDisplay, choiceGC, 0, 0, recs, 1, YXBanded);
      XDrawString (mainDisplay, choiceWindow, choiceGC, x, y, "W", 1);
   }
   else
   {
      XDrawImageString (mainDisplay, choiceBackingPixmap, choiceGC, 0,
            canvasFontAsc, "W", 1);
      image = XGetImage (mainDisplay, choiceBackingPixmap, 0, 0, w, h, 1,
            XYPixmap);
      XSetClipRectangles (mainDisplay, choiceGC, 0, 0, recs, 1, YXBanded);
      saved_x = x;
      saved_y = y;
      for (i = 0; i < w; i++)
      {
         for (j = 0; j < h; j++)
         {
            if (XGetPixel (image, i, j) == 0)
            {
               switch (curRotate)
               {
                  case ROTATE90: x = saved_x-j; y = saved_y+i; break;
                  case ROTATE180: x = saved_x-i; y = saved_y-j; break;
                  case ROTATE270: x = saved_x+j; y = saved_y-i; break;
               }
/*             XDrawPoint (mainDisplay, choiceWindow, defaultGC, x, y); */
               /* Hack to run on RTs -- stupid IBM */
               XDrawLine (mainDisplay, choiceWindow, defaultGC, x, y, x, y);
            }
         }
      }
   }

   recs[0].x = 0;
   recs[0].y = 0;
   recs[0].width = choiceWindowW;
   recs[0].height = choiceWindowH;
   XSetClipRectangles (mainDisplay, choiceGC, 0, 0, recs, 1, YXBanded);
}

void ShowLineWidth ()
{
   XSetStipple (mainDisplay, rasterGC, shortLineWidthPixmap[lineWidth]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, 3*choiceImageW, 0,
         choiceImageW, choiceImageH);
}

void ShowLineStyle ()
{
   XSetStipple (mainDisplay, rasterGC, shortLineStylePixmap[lineStyle]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, 3*choiceImageW,
         choiceImageH, choiceImageW, choiceImageH);
}

void ShowWhereToPrint ()
{
   XSetStipple (mainDisplay, rasterGC, whereToPrintPixmap[whereToPrint]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, 4*choiceImageW, 0,
         choiceImageW, choiceImageH);
}

void ShowLineType ()
{
   XSetStipple (mainDisplay, rasterGC, shortLineTypePixmap[curSpline]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, 4*choiceImageW,
         choiceImageH, choiceImageW, choiceImageH);
}

void ShowFill ()
{
   XGCValues	values;

   values.foreground = myWhitePixel;
   values.background = myBlackPixel;
   values.function = GXcopy;
   values.fill_style = FillOpaqueStippled;
   values.stipple = patPixmap[objFill];
   XChangeGC (mainDisplay, patGC,
         GCForeground | GCBackground | GCFunction | GCFillStyle | GCStipple,
         &values);

   XFillRectangle (mainDisplay, choiceWindow, patGC,
         5*choiceImageW, 0, choiceImageW, choiceImageH);
}

void ShowPen ()
{
   XGCValues	values;

   values.foreground = myWhitePixel;
   values.background = myBlackPixel;
   values.function = GXcopy;
   values.fill_style = FillOpaqueStippled;
   values.stipple = patPixmap[penPat];
   XChangeGC (mainDisplay, patGC,
         GCForeground | GCBackground | GCFunction | GCFillStyle | GCStipple,
         &values);

   XFillRectangle (mainDisplay, choiceWindow, patGC,
         5*choiceImageW, choiceImageH, choiceImageW, choiceImageH);
}

void SetCurChoice (NewChoice)
   int	NewChoice;
{
   if (curChoice == NewChoice) return;

   if (NewChoice == DRAWTEXT && zoomScale != 0)
   {
      Msg ("Text can not be entered while zoomed-out.");
      return;
   }

   switch (curChoice)
   {
      case NOTHING:
         if (topSel != NULL)
         {
            HighLightReverse ();
            RemoveAllSel ();
         }
         break;
      case DRAWTEXT:
         CreateTextObj ();
         if (NewChoice == NOTHING && textDrawn) HighLightJustDrawnText ();
         textDrawn = FALSE;
         SetDefaultCursor (drawWindow);
         textCursorShown = FALSE;
         break;
      case DRAWBOX:
         if (NewChoice == NOTHING && boxDrawn) SelectTopObj ();
         boxDrawn = FALSE;
         break;
      case DRAWCIRCLE:
         if (NewChoice == NOTHING && ovalDrawn) SelectTopObj ();
         ovalDrawn = FALSE;
         break;
      case DRAWPOLY:
         if (NewChoice == NOTHING && polyDrawn) SelectTopObj ();
         polyDrawn = FALSE;
         break;
      case DRAWPOLYGON:
         if (NewChoice == NOTHING && polygonDrawn) SelectTopObj ();
         polygonDrawn = FALSE;
         break;
   }

   curChoice = NewChoice;

   if (curChoice == DRAWTEXT)
   {
      SetTextCursor (drawWindow);
      textCursorShown = FALSE;
   }

   ShowMode ();
}

void ChoiceEventHandler (input)
   XEvent	* input;
{
   XEvent	ev;
   XButtonEvent	* button_ev;
   int		xindex, yindex, delta = 0, root_x, root_y, x, y;
   Window	root_win, child_win;
   unsigned int	status;

   if (input->type == Expose)
   {
      XSync (mainDisplay, FALSE);
      while (XCheckWindowEvent (mainDisplay, choiceWindow, ExposureMask, &ev)) ;
      RedrawChoiceWindow ();
   }
   else if (input->type == ButtonPress)
   {
      button_ev = &(input->xbutton);
      switch (button_ev->button)
      {
         case Button1: delta = 1; break;
         case Button2:
            XQueryPointer (mainDisplay, rootWindow, &root_win, &child_win,
                  &root_x, &root_y, &x, &y, &status);
            break;
         case Button3: delta = -1; break;
         default: return;
      }
      xindex = (int)(button_ev->x / choiceImageW);
      yindex = (int)(button_ev->y / choiceImageH);
      switch (xindex)
      {
         case 0:
            if (yindex == 0)
            {
               if (delta == 0)
                  ModeMenu (x, y);
               else
                  SetCurChoice ((curChoice+MAXCHOICES+delta) % MAXCHOICES);
            }
            else if (colorDisplay)
            {
               if (delta == 0)
                  ColorMenu (x, y);
               else
               {
                  if (curChoice != NOTHING) SetCurChoice (NOTHING);
                  colorIndex = (colorIndex+maxColors+delta) % maxColors;
                  ShowColor ();
                  if (topSel != NULL) ChangeAllSelColor (colorIndex);
               }
            }
            break;
         case 1:
            if (yindex == 0)
            {
               if (delta != 0)
               {
                  horiAlign = (horiAlign+MAXALIGNS+delta) % MAXALIGNS;
                  ShowHoriAlign ();
               }
            }
            else
            {
               if (delta != 0)
               {
                  vertAlign = (vertAlign+MAXALIGNS+delta) % MAXALIGNS;
                  ShowVertAlign ();
               }
            }
            break;
         case 2:
            if (yindex == 0)
            {
               if (delta == 0)
                  StyleMenu (x, y);
               else
               {
                  if (curChoice != NOTHING) SetCurChoice (NOTHING);
                  textJust = (textJust+MAXJUSTS+delta) % MAXJUSTS;
                  ShowJust ();
                  if (topSel != NULL) ChangeFontJust (textJust);
               }
            }
            else
            {
               if (delta == 0)
                  FontMenu (x, y);
               else
               {
                  if (curChoice != NOTHING) SetCurChoice (NOTHING);
                  curRotate = (curRotate+4+delta) % 4;
                  ShowCurFont ();
                  if (topSel != NULL)
                  {
                     if (delta == 1)
                        RotateClockWise ();
                     else
                        RotateCounter ();
                  }
               }
            }
            break;
         case 3:
            if (yindex == 0)
            {
               if (delta == 0)
                  LineStyleMenu (x, y);
               else
               {
                  if (curChoice != NOTHING) SetCurChoice (NOTHING);
                  lineWidth = (lineWidth+MAXLINEWIDTHS+delta) % MAXLINEWIDTHS;
                  ShowLineWidth ();
                  if (topSel != NULL) ChangeAllSelLineWidth (lineWidth);
               }
            }
            else
            {
               if (delta == 0)
                  LineStyleMenu (x, y);
               else
               {
                  if (curChoice != NOTHING) SetCurChoice (NOTHING);
                  lineStyle = (lineStyle+MAXLINESTYLES+delta) % MAXLINESTYLES;
                  ShowLineStyle ();
                  if (topSel != NULL) ChangeAllSelLineStyle (lineStyle);
               }
            }
            break;
         case 4:
            if (yindex == 0)
            {
               if (delta == 0)
                  LayoutMenu (x, y);
               else
               {
                  whereToPrint = !whereToPrint;
                  ShowWhereToPrint ();
                  if (whereToPrint == PRINTER)
                     Msg ("Print device set to printer.");
                  else
                     Msg ("Print device set to LaTeX-Figure file.");
               }
            }
            else
            {
               if (delta == 0)
                  LineStyleMenu (x, y);
               else
               {
                  if (curChoice != NOTHING) SetCurChoice (NOTHING);
                  curSpline = (curSpline+MAXLINETYPES+delta) % MAXLINETYPES;
                  ShowLineType ();
                  if (topSel != NULL) ChangeAllSelLineType (curSpline);
               }
            }
            break;
         case 5:
            if (yindex == 0)
            {
               if (delta == 0)
                  FillMenu (x, y);
               else
               {
                  if (curChoice != NOTHING) SetCurChoice (NOTHING);
                  objFill = (objFill+MAXPATTERNS+delta) % MAXPATTERNS;
                  ShowFill ();
                  if (topSel != NULL) ChangeAllSelFill (objFill);
               }
            }
            else
            {
               if (delta == 0)
                  PenMenu (x, y);
               else
               {
                  if (curChoice != NOTHING) SetCurChoice (NOTHING);
                  penPat = (penPat+MAXPATTERNS+delta) % MAXPATTERNS;
                  ShowPen ();
                  if (topSel != NULL) ChangeAllSelPen (penPat);
               }
            }
            break;
      }
   }
}
