/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/color.c,v 1.6 90/05/17 08:21:55 william Exp $";
#endif

#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "choice.e"
#include "mark.e"
#include "menu.e"
#include "obj.e"
#include "raster.e"
#include "select.e"
#include "setup.e"
#include "text.e"

int	maxColors = MAXCOLORS;
int	defaultColorIndex = 4;
int	colorIndex = 0;
int	* colorPixels = NULL;
int	* xorColorPixels = NULL;
char	* * colorMenuItems = NULL;

static char	* defaultColorMenuItems[MAXCOLORS] =
{
   "magenta", "red", "green", "blue", "yellow", "pink", "cyan", "CadetBlue",
   "white", "DarkSlateGray"
};
static XColor	savedColorIndex;

void DefaultColorArrays (Entries, ForePixels, Valid)
   int	Entries, * * ForePixels, * * Valid;
{
   register int	i, * fore_pixels, pixel, * valid;

   pixel = myWhitePixel;
   *ForePixels = fore_pixels = (int *) calloc (Entries, sizeof(int));
   *Valid = valid = (int *) calloc (Entries, sizeof(int));
   for (i = 0; i < Entries; i++)
   {
      *fore_pixels++ = pixel;
      *valid++ = TRUE;
   }
}

int FindColorIndex (s)
   char	* s;
{
   register int i;

   if (colorMenuItems == NULL) return (0);

   for (i = 0; i < maxColors; i++)
      if (strcmp(s, colorMenuItems[i]) == 0)
         break;
   if (i == maxColors)
   {
      printf ("Can not find color `%s', use `%s' instead.\n", s,
            colorMenuItems[defaultColorIndex]);
      return (defaultColorIndex);
   }
   return (i);
}

void InitColor ()
{
   register int	i;
   XColor	color, exact_def;
   char		buf[80], * c_ptr;
   int		num, color_in_x_default = FALSE;
   int		black_allocated = FALSE, white_allocated = FALSE;

   if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, "MaxColors")) != NULL)
   {
      color_in_x_default = TRUE;
      maxColors = atoi (c_ptr);
      if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, "DefaultColorIndex")) !=
            NULL)
      {
         num = atoi (c_ptr);
         if (num < maxColors)
            defaultColorIndex = num;
         else
         {
            printf ("Warning:  DefaultColorIndex >= MaxColors, ");
            printf ("Use 0 for DefaultColorIndex\n");
            defaultColorIndex = 0;
         }
      }
      else
         defaultColorIndex = 0;
   }
   else
   {
      if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, "DefaultColorIndex")) !=
            NULL)
      {
         num = atoi (c_ptr);
         if (num < maxColors) defaultColorIndex = num;
      }
   }

   colorIndex = defaultColorIndex;

   colorPixels = (int *) calloc (maxColors, sizeof (int));
   xorColorPixels = (int *) calloc (maxColors, sizeof (int));
   colorMenuItems = (char * *) calloc (maxColors, sizeof (char *));

   if (color_in_x_default)
   {
      for (i = 0; i < maxColors; i++)
      {
         sprintf (buf, "Color%1d", i);
         if ((c_ptr = XGetDefault (mainDisplay, TOOL_NAME, buf)) != NULL)
         {
            colorMenuItems[i] = (char *) calloc (strlen(c_ptr),sizeof(char));
            strcpy (colorMenuItems[i], c_ptr);
         }
         else
         {
            printf ("Could not GetDefault %s*%s\n", TOOL_NAME, buf);
            exit (-1);
         }
      }
   }
   else
      for (i = 0; i < maxColors; i++)
         colorMenuItems[i] = defaultColorMenuItems[i];

   if (colorDisplay)
   {
      for (i = 0; i < maxColors; i++)
      {
         XAllocNamedColor(mainDisplay, mainColormap, colorMenuItems[i], &color,
               &exact_def);
         colorPixels[i] = color.pixel;
   
         if (strcmp (colorMenuItems[i], "white") == 0)
         {
            colorPixels[i] = myWhitePixel = color.pixel;
            white_allocated = TRUE;
         }
         else if (strcmp (colorMenuItems[i], "black") == 0)
         {
            colorPixels[i] = myBlackPixel = color.pixel;
            black_allocated = TRUE;
         }
      }

      if (!white_allocated)
      {
         XAllocNamedColor(mainDisplay, mainColormap, "white", &color,
               &exact_def);
         myWhitePixel = color.pixel;
      }
      if (!black_allocated)
      {
         XAllocNamedColor(mainDisplay, mainColormap, "black", &color,
               &exact_def);
         myBlackPixel = color.pixel;
      }

      for (i = 0; i < maxColors; i++)
         xorColorPixels[i] = colorPixels[i] ^ myBlackPixel;
   }
   else
   {
      XAllocNamedColor(mainDisplay, mainColormap, "white", &color, &exact_def);
      myWhitePixel = color.pixel;
      XAllocNamedColor(mainDisplay, mainColormap, "black", &color, &exact_def);
      myBlackPixel = color.pixel;

      for (i = 0; i < maxColors; i++)
      {
         colorPixels[i] = myWhitePixel;
         xorColorPixels[i] = myWhitePixel ^ myBlackPixel;
      }
      colorIndex = 0;
   }
}

static
int ChangeObjColor (ObjPtr, ColorIndex)
   struct ObjRec	* ObjPtr;
   int			ColorIndex;
{
   register struct ObjRec	* obj_ptr;
   int				changed = FALSE;

   for (obj_ptr = ObjPtr; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
      switch (obj_ptr->type)
      {
         case OBJ_POLY:
         case OBJ_BOX:
         case OBJ_OVAL:
         case OBJ_TEXT:
         case OBJ_POLYGON:
            if (obj_ptr->color != ColorIndex)
            {
               obj_ptr->color = ColorIndex;
               changed = TRUE;
            }
            break;

         case OBJ_GROUP:
         case OBJ_SYM:
         case OBJ_ICON:
            if (ChangeObjColor (obj_ptr->detail.r->last, ColorIndex))
               changed = TRUE;
            break;
      }
   return (changed);
}

void ChangeAllSelColor (ColorIndex)
   int	ColorIndex;
{
   register struct SelRec	* sel_ptr;
   register struct ObjRec	* obj_ptr;
   int				text_obj_created = FALSE, text_cursor_shown;
   int				changed = FALSE;

   if (topSel == NULL)
   {
      text_cursor_shown = textCursorShown;
      text_obj_created = TieLooseEnds ();
      colorIndex = ColorIndex;
      ShowColor ();
      if (!text_obj_created && curChoice == DRAWTEXT && text_cursor_shown)
      {
         NewCurText ();
         RedrawCurText ();
      }
      else
         textCursorShown = FALSE;
      return;
   }

   for (sel_ptr = botSel; sel_ptr != NULL; sel_ptr = sel_ptr->prev)
   {
      obj_ptr = sel_ptr->obj;
      switch (obj_ptr->type)
      {
         case OBJ_POLY:
         case OBJ_BOX:
         case OBJ_OVAL:
         case OBJ_TEXT:
         case OBJ_POLYGON:
            if (obj_ptr->color != ColorIndex)
            {
               obj_ptr->color = ColorIndex;
               changed = TRUE;
            }
            break;

         case OBJ_GROUP:
         case OBJ_SYM:
         case OBJ_ICON:
            if (ChangeObjColor (obj_ptr->detail.r->last, ColorIndex))
               changed = TRUE;
            break;
      }
   }

   if (changed)
   {
      HighLightReverse ();
      fileModified = TRUE;
      RedrawAnArea (botObj, selLtX-(1<<zoomScale), selLtY-(1<<zoomScale),
            selRbX+(1<<zoomScale), selRbY+(1<<zoomScale));
      HighLightForward ();
   }
}

void ColorMenu (X, Y)
   int	X, Y;
{
   register int	i, index, * fore_colors;
   Pixmap	* pixmap;
   int		w;

   if (!colorDisplay) return;

   pixmap = (Pixmap *) calloc (maxColors, sizeof (Pixmap));
   fore_colors = (int *) calloc (maxColors, sizeof(int));
   for (i = 0; i < maxColors; i++)
   {
      pixmap[i] = patPixmap[1];
      fore_colors[i] = colorPixels[i];
   }
   if ((maxColors % 10) == 0)
      w = (int)(maxColors / 10);
   else
      w = (int)(maxColors / 10) + 1;
   index = PxMpMenuLoop (X, Y, choiceImageW, choiceImageH, 10, w,
         maxColors, fore_colors, pixmap, MULTICOLOR);

   if (index != INVALID) ChangeAllSelColor (index);
   cfree (pixmap);
}
