/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/cursor.c,v 1.1 90/04/01 22:16:09 william Exp $";
#endif

#include <X11/Xlib.h>
#include <X11/cursorfont.h>
#include "types.h"
#include "const.h"

#include "choice.e"
#include "setup.e"

#include "bitmaps/nullmask.bm"
#include "bitmaps/text_cur_image.bm"

Cursor	nullCursor;
Cursor	cornerCursor;
Cursor	defaultCursor;
Cursor	handCursor;

static GC	textCursorGC;
static Cursor	textCursor;
static Pixmap	textPixmap;
static Pixmap	nullPixmap;

static XImage	* textCursorImage;

void SetTextCursor (window)
   Window	window;
{
   XDefineCursor (mainDisplay, window, textCursor);
}

void SetNullCursor (window)
   Window	window;
{
   XDefineCursor (mainDisplay, window, nullCursor);
}

void SetDefaultCursor (window)
   Window	window;
{
   XUndefineCursor (mainDisplay, window);
}

void ShowCursor ()
{
   if (curChoice == DRAWTEXT)
      SetTextCursor (drawWindow);
   else
      SetDefaultCursor (drawWindow);
}

void CreateCursor ()
{
   XGCValues	values;
   XColor	color;

   textPixmap = XCreateBitmapFromData (mainDisplay, mainWindow,
         text_cur_image_bits, text_cur_image_width, text_cur_image_height);

   values.foreground = myWhitePixel;
   values.background = myBlackPixel;
   values.fill_style = FillOpaqueStippled;
   values.stipple = textPixmap;
   textCursorGC = XCreateGC (mainDisplay, mainWindow,
         GCForeground | GCBackground | GCFillStyle | GCStipple, &values);

   textCursorImage = XCreateImage (mainDisplay, mainVisual, 1, XYBitmap, 0,
         text_cur_image_bits, text_cur_image_width, text_cur_image_height, 8,
         2);
   textCursorImage->byte_order = LSBFirst;
   textCursorImage->bitmap_bit_order = LSBFirst;
   textCursorImage->data = text_cur_image_bits;

   textCursor = XCreateFontCursor (mainDisplay, XC_xterm);
   cornerCursor = XCreateFontCursor (mainDisplay, XC_ul_angle);
   defaultCursor = XCreateFontCursor (mainDisplay, XC_arrow);
   handCursor = XCreateFontCursor (mainDisplay, XC_hand2);

   nullPixmap = XCreatePixmap (mainDisplay, mainWindow, nullmask_width,
         nullmask_height, 1);
   nullCursor = XCreatePixmapCursor (mainDisplay, nullPixmap, nullPixmap,
         &color, &color, 0, 0);
}

void PutCursor (window, x, y, foreground)
   Window	window;
   int		x, y, foreground;
{
   XSetForeground (mainDisplay, textCursorGC, foreground);
   XPutImage (mainDisplay, window, textCursorGC, textCursorImage, 0, 0, x, y,
         text_cur_image_width, text_cur_image_height);
}

void CleanUpCursors ()
{
   XFreePixmap (mainDisplay, textPixmap);
   XFreeGC (mainDisplay, textCursorGC);
/* XDestroyImage (textCursorImage); */

   XFreeCursor (mainDisplay, textCursor);
   XFreeCursor (mainDisplay, cornerCursor);
   XFreeCursor (mainDisplay, defaultCursor);
   XFreeCursor (mainDisplay, handCursor);

   XFreePixmap (mainDisplay, nullPixmap);
   XFreeCursor (mainDisplay, nullCursor);
}
