/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/edit.c,v 1.1 90/04/01 22:16:14 william Exp $";
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "align.e"
#include "color.e"
#include "drawing.e"
#include "dup.e"
#include "group.e"
#include "mark.e"
#include "obj.e"
#include "raster.e"
#include "select.e"
#include "setup.e"
#include "stretch.e"

#define EDIT_REDRAW 0
#define EDIT_DUP 1
#define EDIT_DELETE 2
#define EDIT_SELALL 3
#define EDIT_UNDODEL 4
#define MAXEDITMENUS 5

static char * editMenuStr[] =
      { "Redraw     ^R",
        "Duplicate  ^D",
        "Delete     ^X",
        "SelectAll  ^A",
        "UndoDelete #U"
      };

void EditMenu (X, Y)
   int	X, Y;
{
   int		index, * fore_colors, * valid;

   DefaultColorArrays (MAXEDITMENUS, &fore_colors, &valid);
   index = TextMenuLoop (X, Y, editMenuStr, MAXEDITMENUS, fore_colors, valid,
         SINGLECOLOR);

   switch (index)
   {
      case EDIT_REDRAW: ClearAndRedrawDrawWindow (); break;
      case EDIT_DUP: DupSelObj (); break;
      case EDIT_DELETE: DelAllSelObj (); break;
      case EDIT_SELALL: SelAllObj (); break;
      case EDIT_UNDODEL: UndoDelete (); break;
   }
}

void FrontProc ()
{
   if (topSel != NULL)
   {
      HighLightReverse ();
      MoveSelToTop ();
      RedrawAnArea (botObj, selLtX-(1<<zoomScale), selLtY-(1<<zoomScale),
            selRbX+(1<<zoomScale), selRbY+(1<<zoomScale));
      HighLightForward ();
      fileModified = TRUE;
   }
}

void BackProc ()
{
   if (topSel != NULL)
   {
      HighLightReverse ();
      MoveSelToBot ();
      RedrawAnArea (botObj, selLtX-(1<<zoomScale), selLtY-(1<<zoomScale),
            selRbX+(1<<zoomScale), selRbY+(1<<zoomScale));
      HighLightForward ();
      fileModified = TRUE;
   }
}

#define ARRANGE_FRONT 0
#define ARRANGE_BACK 1
#define ARRANGE_GROUP 2
#define ARRANGE_UNGROUP 3
#define ARRANGE_ALIGNOBJ 4
#define ARRANGE_ALIGNGRID 5
#define FLIP_HORIZONTAL 6
#define FLIP_VERTICAL 7
#define ROTATE_CLOCKWISE 8
#define ROTATE_COUNTER 9
#define MAXARRANGEMENUS 10

static char * arrangeMenuStr[] =
      { "Front           ^F",
        "Back            ^B",
        "Group           ^G",
        "UnGroup         ^U",
        "AlignObjs       ^L",
        "AlignToGrid     ^T",
        "FlipHorizontal  #H",
        "FlipVertical    #V",
        "RotateClockWise #W",
        "RotateCounter   #C",
      };

void ArrangeMenu (X, Y)
   int	X, Y;
{
   int		index, * fore_colors, * valid;

   DefaultColorArrays (MAXARRANGEMENUS, &fore_colors, &valid);
   index = TextMenuLoop (X, Y, arrangeMenuStr, MAXARRANGEMENUS, fore_colors,
         valid, SINGLECOLOR);
   switch (index)
   {
      case ARRANGE_FRONT: FrontProc (); break;
      case ARRANGE_BACK: BackProc (); break;
      case ARRANGE_GROUP: GroupSelObj (); break;
      case ARRANGE_UNGROUP: UngroupSelObj (); break;
      case ARRANGE_ALIGNOBJ: AlignSelObjs (); break;
      case ARRANGE_ALIGNGRID: AlignSelToGrid (); break;
      case FLIP_HORIZONTAL: FlipHorizontal (); break;
      case FLIP_VERTICAL: FlipVertical (); break;
      case ROTATE_CLOCKWISE: RotateClockWise (); break;
      case ROTATE_COUNTER: RotateCounter (); break;
   }
}
