/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/file.c,v 1.2 90/05/07 15:25:55 william Exp $";
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "align.e"
#include "attr.e"
#include "box.e"
#include "button.e"
#include "choice.e"
#include "color.e"
#include "dialog.e"
#include "drawing.e"
#include "font.e"
#include "grid.e"
#include "group.e"
#include "menu.e"
#include "msg.e"
#include "names.e"
#include "poly.e"
#include "polygon.e"
#include "obj.e"
#include "oval.e"
#include "pattern.e"
#include "rect.e"
#include "ruler.e"
#include "scroll.e"
#include "select.e"
#include "setup.e"
#include "stk.e"
#include "text.e"

#define CUR_VERSION 6

char	curFileName[80];
int	curFileDefined = FALSE;
int	fileVersion = INVALID;
int	importingFile = FALSE;

void ClearFileInfo ()
{
   curFileName[0] = '\0';
   curFileDefined = FALSE;
}

int OkayToCreateFile (FileName)
   char	* FileName;
{
   char s[256];

   if (fopen (FileName, "r") == NULL) return (TRUE);
   switch (YesNoCancel ("File exists, okay to overwrite? [ync](y)"))
   {
      case CONFIRM_YES: break;
      case CONFIRM_NO: return (FALSE);
      case CONFIRM_CANCEL: return (FALSE);
   }
   sprintf (s, "rm -f %s", FileName);
   system (s);
   return (TRUE);
}

void Save (FP, BotObjPtr, Level)
   FILE			* FP;
   struct ObjRec	* BotObjPtr;
   int			Level;
{
   struct ObjRec	* obj_ptr;

   TieLooseEnds ();

   if (Level == 0)
   {
      fprintf (FP, "state(%1d,%1d,", pageStyle, CUR_VERSION);
      fprintf (FP, "%1d,%1d,%1d,", drawOrigX, drawOrigY, zoomScale);
      fprintf (FP, "%1d,%1d,%1d,", xyGrid, gridOn, colorIndex);
      fprintf (FP, "%1d,%1d,%1d,", horiAlign, vertAlign, lineWidth);
      fprintf (FP, "%1d,%1d,%1d,%1d,", curSpline, lineStyle, objFill, penPat);
      fprintf (FP, "%1d,%1d,%1d,%1d", textJust, curFont, curStyle, curSize);
      fprintf (FP, ").\n");
   }

   for (obj_ptr = BotObjPtr; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      switch (obj_ptr->type)
      {
         case OBJ_POLY: SavePolyObj (FP, obj_ptr); break;
         case OBJ_BOX: SaveBoxObj (FP, obj_ptr); break;
         case OBJ_OVAL: SaveOvalObj (FP, obj_ptr); break;
         case OBJ_TEXT: SaveTextObj (FP, obj_ptr); break;
         case OBJ_POLYGON: SavePolygonObj (FP, obj_ptr); break;
         case OBJ_GROUP: SaveGroupObj (FP, obj_ptr, Level); break;
         case OBJ_SYM: SaveCompObj (FP, obj_ptr, Level); break;
         case OBJ_ICON: SaveIconObj (FP, obj_ptr, Level); break;
      }
      if (obj_ptr->prev == NULL)
         if (Level == 0)
            fprintf (FP, ".\n");
         else
            fprintf (FP, "\n");
      else
         if (Level == 0)
            fprintf (FP, ".\n");
         else
            fprintf (FP, ",\n");
   }
   SetCurChoice (NOTHING);
}

#define OBJ_FILE_SAVED 0
#define SYM_FILE_SAVED 1

static
int SaveTmpFile ()
   /* return TRUE if file successfully saved */
{
   char			new_file_name[80], s[80];
   FILE			* fp;
   int			count = 0, status = INVALID;
   struct ObjRec	* obj_ptr;

   strcpy (new_file_name, "tmpmodel");

   for (obj_ptr = topObj; obj_ptr != NULL; obj_ptr = obj_ptr->next)
      if (obj_ptr->type == OBJ_SYM) count++;

   switch (count)
   {
      case 0: strcat (new_file_name, ".obj"); status = OBJ_FILE_SAVED; break;
      case 1: strcat (new_file_name, ".sym"); status = SYM_FILE_SAVED; break;
      default:
         Msg ("TOO MANY SYMBOLS!  Symbol file not saved.");
         return (INVALID);
   }

   sprintf (s, "rm -f %s", new_file_name);
   system (s);

   if ((fp = fopen (new_file_name, "w")) == NULL)
   {
      sprintf (s, "Can not open '%s', file not saved", new_file_name);
      Msg (s);
      return (INVALID);
   }

   sprintf (s, "Saving temporary file '%s' ...", new_file_name);
   Msg (s);
   Save (fp, botObj, 0);
   fclose (fp);
   sprintf (s, "Temporary file '%s' saved.", new_file_name);
   Msg (s);

   chmod (new_file_name, 0777);
   return (status);
}

void SaveNewFile ()
{
   char			new_file_name[80], s[80], new_full_name[MAXPATHLENGTH];
   FILE			* fp;
   int			count = 0;
   struct ObjRec	* obj_ptr;

   Dialog ("Please Enter New File Name:", new_file_name);
   if (*new_file_name == '\0') return;

   for (obj_ptr = topObj; obj_ptr != NULL; obj_ptr = obj_ptr->next)
      if (obj_ptr->type == OBJ_SYM) count++;

   switch (count)
   {
      case 0: strcat (new_file_name, ".obj"); break;
      case 1: strcat (new_file_name, ".sym"); break;
      default:
         Msg ("TOO MANY SYMBOLS!  Symbol file not saved.");
         return;
   }

   strcpy (new_full_name, curDomainName);
   if (*curDomainName != '\0') strcat (new_full_name, "/");
   strcat (new_full_name, new_file_name);

   if (!OkayToCreateFile (new_full_name)) return;

   if ((fp = fopen (new_full_name, "w")) == NULL)
   {
      sprintf (s, "Can not open '%s', file not saved.", new_file_name);
      Msg (s);
      return;
   }

   strcpy (curFileName, new_file_name);
   curFileDefined = TRUE;
   sprintf (s, "Saving '%s' ...", new_file_name);
   Msg (s);
   Save (fp, botObj, 0);
   fclose (fp);
   sprintf (s, "File '%s' saved.", new_file_name);
   Msg (s);
   fileModified = FALSE;

   RedrawTitleWindow ();
}

void SaveFile ()
{
   int			i, len, count;
   struct ObjRec	* obj_ptr;
   FILE			* fp;
   char			ext[80], s[80], full_name[MAXPATHLENGTH];

   if (!curFileDefined)
   {
      SaveNewFile ();
      return;
   }

   len = strlen (curFileName);
   for (i = len-1; curFileName[i] != '.'; i--) ;
   strcpy (ext, &curFileName[i+1]);

   for (obj_ptr = topObj, count = 0; obj_ptr != NULL; obj_ptr = obj_ptr->next)
      if (obj_ptr->type == OBJ_SYM) count++;
   switch (count)
   {
      case 0:
         if (strcmp (ext, "sym") == 0)
         {
            Msg ("No SYMBOL defined in symbol file.  Symbol file not saved.");
            return;
         }
         break;
      case 1:
         if (strcmp (ext, "obj") == 0)
         {
            Msg ("One SYMBOL defined in OBJECT file.  Object file not saved.");
            return;
         }
         break;
      default:
         if (strcmp (ext, "sym") == 0)
         {
            Msg ("Too many SYMBOLS in symbol file!  Symbol file not saved.");
            return;
         }
         break;
   }

   strcpy (full_name, curDomainName);
   if (*curDomainName != '\0') strcat (full_name, "/");
   strcat (full_name, curFileName);
   if ((fp = fopen (full_name, "w")) == NULL)
   {
      sprintf (s, "Can not open %s, file not saved.", curFileName);
      Msg (s);
      return;
   }

   sprintf (s, "Saving '%s' ...", curFileName);
   Msg (s);

   Save (fp, botObj, 0);

   fclose (fp);
   sprintf (s, "File '%s' saved.", curFileName);
   Msg (s);
   fileModified = FALSE;
}

char * ParseStr (Str, C, Left)
   char	* Str, C, * Left;
{
   register char	* s = Str, * l = Left;

   while (*s != '\0' && *s != C) *l++ = *s++;

   if (*s == C) s++;
   *l = '\0';
   return (s);
}

char * FindChar (C, Str)
   char	C;
   char	* Str;
   /* returns the address of the character right after C of the string Str */
{
   register char	* s = Str;

   while (*s != '\0' && *s != C) s++;

   if (*s == C) s++;
   return (s);
}

void ReadState (Inbuf)
   char	* Inbuf;
{
   char	* s;
   int	page_style, size;

   s = FindChar ('(', Inbuf);
   sscanf (s, "%d", &page_style);
   s = FindChar (',', s);
   if (*s == '\0')
      fileVersion = INVALID;
   else
      sscanf (s, "%d", &fileVersion);

   if (!importingFile)
   {
      pageStyle = page_style;
      if (fileVersion >= 2)
      {
         s = FindChar (',', s);
         if (fileVersion <= 3)
         {
            sscanf (s, "%d , %d , %d , %d , %d , %d , %d , %d , %d , %d , \
                  %d , %d , %d , %d , %d , %d",
                  &drawOrigX, &drawOrigY, &zoomScale, &xyGrid, &gridOn,
                  &colorIndex, &horiAlign, &vertAlign, &lineWidth, &lineStyle,
                  &objFill, &penPat, &textJust, &curFont, &curStyle, &curSize);
            if (lineWidth == LINE_CURVED)
            {
               lineWidth = 0;
               curSpline = LT_SPLINE;
            }
            else
               curSpline = LT_STRAIGHT;
         }
         else
         {
            sscanf (s, "%d , %d , %d , %d , %d , %d , %d , %d , %d , \
                  %d , %d , %d , %d , %d , %d , %d , %d",
                  &drawOrigX, &drawOrigY, &zoomScale, &xyGrid, &gridOn,
                  &colorIndex, &horiAlign, &vertAlign, &lineWidth, &curSpline,
                  &lineStyle, &objFill, &penPat, &textJust, &curFont,
                  &curStyle, &curSize);
         }
         if (colorIndex >= maxColors)
         {
            printf ("Can not find color `%s', use `%s' instead.\n", s,
                  colorMenuItems[defaultColorIndex]);
            colorIndex = defaultColorIndex;
         }
         SetCanvasFont ();
      }

      UpdDrawWinWH ();
      UpdPageStyle (pageStyle);
      UpdDrawWinBBox ();

      DrawPaperBoundary ();
      RedrawGridLines ();
      RedrawRulers ();
      RedrawChoiceWindow ();
   }
}

int ReadObj (FP, ObjPtr)
   FILE			* FP;
   struct ObjRec	* * ObjPtr;
{
   char			inbuf[256], obj_name[10], tmp_str[256], * s, * s1;
   char			* line = NULL, * c_ptr;
   int			len, id, old_len, cur_size, done = FALSE;
   int			allocated = FALSE;
   struct ObjRec	* tmp_obj_ptr;
   struct ObjRec	* tmp_top_obj = NULL, * tmp_bot_obj = NULL;
   struct AttrRec       * top_attr = NULL, * bot_attr = NULL, * attr_ptr;

   while (fgets (inbuf, 256, FP) != NULL)
   {
      if (inbuf[0] == ']') return (FALSE);

      len = strlen(inbuf);
      if (inbuf[len-1] != '\r' && inbuf[len-1] != '\n')
      {  /* line longer than 256 characters */
         /* inbuf[255] == '\0' and len == 255 now */
         cur_size = 510;
         allocated = TRUE;
         line = (char *) calloc (cur_size, sizeof(char));
         strcpy (line, inbuf);
         c_ptr = &(line[255]);
         while (fgets (inbuf, 256, FP) != NULL && !done)
         {
            len = strlen(inbuf);
            if (inbuf[len-1] == '\r' || inbuf[len-1] == '\n')
            {
               done = TRUE;
               inbuf[len-1] = '\0';
               strcpy (c_ptr, inbuf);
            }
            else
            {
               cur_size += 255;
               line = (char *) realloc (line, cur_size);
               strcpy (c_ptr, inbuf);
               c_ptr += 255;
            }
         }
         if (!done)
         {
            if (allocated) cfree (line);
            return (FALSE);
         }
      }
      else
      {
         line = inbuf;
         line[len-1] = '\0';
      }

      ParseStr (line, '(', obj_name);
      if (strcmp (obj_name, "poly") == 0)
      {
         ReadPolyObj (line, ObjPtr);
         if (fileVersion != INVALID)
            while (ReadAttr (FP, &attr_ptr))
            {
               attr_ptr->owner = *ObjPtr;
               attr_ptr->prev = NULL;
               attr_ptr->next = top_attr;
               if (top_attr == NULL)
                  bot_attr = attr_ptr;
               else
                  top_attr->prev = attr_ptr;
               top_attr = attr_ptr;
            }
         if (bot_attr != NULL) bot_attr->next = NULL;
         (*ObjPtr)->detail.p->fattr = top_attr;
         (*ObjPtr)->detail.p->lattr = bot_attr;
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "box") == 0)
      {
         ReadBoxObj (line, ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "oval") == 0)
      {
         ReadOvalObj (line, ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "text") == 0)
      {
         ReadTextObj (FP, line, ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "polygon") == 0)
      {
         ReadPolygonObj (line, ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "group") == 0)
      {
         ReadGroupObj (FP, ObjPtr);
         if (fileVersion != INVALID)
            while (ReadAttr (FP, &attr_ptr))
            {
               attr_ptr->owner = *ObjPtr;
               attr_ptr->prev = NULL;
               attr_ptr->next = top_attr;
               if (top_attr == NULL)
                  bot_attr = attr_ptr;
               else
                  top_attr->prev = attr_ptr;
               top_attr = attr_ptr;
            }
         if (bot_attr != NULL) bot_attr->next = NULL;
         (*ObjPtr)->detail.r->fattr = top_attr;
         (*ObjPtr)->detail.r->lattr = bot_attr;
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "sym") == 0)
      {
         ReadGroupObj (FP, ObjPtr);
         (*ObjPtr)->type = OBJ_SYM;
         if (fileVersion != INVALID)
            while (ReadAttr (FP, &attr_ptr))
            {
               attr_ptr->owner = *ObjPtr;
               attr_ptr->prev = NULL;
               attr_ptr->next = top_attr;
               if (top_attr == NULL)
                  bot_attr = attr_ptr;
               else
                  top_attr->prev = attr_ptr;
               top_attr = attr_ptr;
            }
         if (bot_attr != NULL) bot_attr->next = NULL;
         (*ObjPtr)->detail.r->fattr = top_attr;
         (*ObjPtr)->detail.r->lattr = bot_attr;
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "icon") == 0)
      {
         ReadGroupObj (FP, ObjPtr);
         (*ObjPtr)->type = OBJ_ICON;
         if (fgets (line, 256, FP) == NULL)
         {
            if (allocated) cfree (line);
            return (FALSE);
         }

         strcpy(tmp_str, FindChar ('"', line));
         s = FindChar ('"', tmp_str);
         if (fileVersion != INVALID)
         {
            s1 = FindChar (',', s);
            sscanf (s1, "%d", &id);
            if (id >= objId) objId = id+1;
            (*ObjPtr)->detail.r->id = id;
         }
         (*ObjPtr)->detail.r->dirty = FALSE;
         *(--s) = '\0';
         strcpy ((*ObjPtr)->detail.r->s, tmp_str);
         if (fileVersion != INVALID)
            while (ReadAttr (FP, &attr_ptr))
            {
               attr_ptr->owner = *ObjPtr;
               attr_ptr->prev = NULL;
               attr_ptr->next = top_attr;
               if (top_attr == NULL)
                  bot_attr = attr_ptr;
               else
                  top_attr->prev = attr_ptr;
               top_attr = attr_ptr;
            }
         if (bot_attr != NULL) bot_attr->next = NULL;
         (*ObjPtr)->detail.r->fattr = top_attr;
         (*ObjPtr)->detail.r->lattr = bot_attr;
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "state") == 0)
      {
         ReadState (line);
         *ObjPtr = NULL;
         if (allocated) cfree (line);
         return (TRUE);
      }
   }
   if (allocated) cfree (line);
   return (FALSE);
}

void ChangeDomain ()
{
   char 	domain_name[80], s[80];
   register int	i;

   if (SelectDomain (domain_name) == INVALID) return;

   if (strcmp (domain_name, "..") == 0)
   {
      if (*curDomainName == '\0')
         Msg ("Already at the top.  Domain not changed.");
      else
      {
         i = strlen (curDomainName) - 1;
         for ( ; i >= 0 && curDomainName[i] != '/'; i--) ;
         if (i >= 0)
         {
            curDomainName[i] = '\0';
            sprintf (s, "Current domain is %s.", curDomainName);
            Msg (s);
         }
         else
         {
            curDomainName[0] = '\0';
            Msg ("At the top of the domain hierarchy.");
         }
      }
   }
   else if (strcmp (domain_name, ".") == 0)
   {
      sprintf (s, "Current domain is %s.", curDomainName);
      Msg (s);
   }
   else
   {
      if (*curDomainName != '\0') strcat (curDomainName, "/");
      strcat (curDomainName, domain_name);
      sprintf (s, "Current domain is %s.", curDomainName);
      Msg (s);
   }
}

void ImportFile ()
{
   struct ObjRec	* obj_ptr, * saved_top_obj, * saved_bot_obj;
   char 		file_name[80], s[80], full_name[MAXPATHLENGTH];
   FILE			* fp;

   if (SelectFileName ("obj", "Please Select A File To IMPORT ...",
         file_name) == INVALID) return;

   strcat (file_name, ".obj");
   strcpy (full_name, curDomainName);
   if (*curDomainName != '\0') strcat (full_name, "/");
   strcat (full_name, file_name);

   if ((fp = fopen (full_name, "r")) == NULL)
   {
      sprintf (s, "Can not import '%s'.", file_name);
      Msg (s);
      return;
   }

   TieLooseEnds ();
   if (topSel != NULL) { HighLightReverse (); RemoveAllSel (); }
   saved_top_obj = topObj;
   saved_bot_obj = botObj;
   topObj = botObj = NULL;

   importingFile = TRUE;
   sprintf (s, "Importing '%s' ...", file_name);
   Msg (s);

   while (ReadObj (fp, &obj_ptr))
      if (obj_ptr != NULL)
         AddObj (NULL, topObj, obj_ptr);

   fclose (fp);
   importingFile = FALSE;
   if (topObj != NULL) fileModified = TRUE;

   RedrawDrawWindow (botObj);
   SelAllObj ();

   if (botObj != NULL)
      botObj->next = saved_top_obj;
   else
      topObj = saved_top_obj;

   if (saved_top_obj != NULL)
   {
      saved_top_obj->prev = botObj;
      botObj = saved_bot_obj;
   }

   sprintf (s, "'%s' imported.", file_name);
   Msg (s);
}

void LoadFile ()
{
   struct ObjRec	* obj_ptr;
   char 		file_name[80], s[80], full_name[MAXPATHLENGTH];
   FILE			* fp;

   if (SelectFileName ("obj", "Please Select A File To OPEN ...",
         file_name) == INVALID) return;

   strcat (file_name, ".obj");
   strcpy (full_name, curDomainName);
   if (*curDomainName != '\0') strcat (full_name, "/");
   strcat (full_name, file_name);

   if ((fp = fopen (full_name, "r")) == NULL)
   {
      sprintf (s, "Can not open '%s'.", file_name);
      Msg (s);
      return;
   }

   TieLooseEnds ();
   CleanUpDrawingWindow ();
   sprintf (s, "Loading '%s' ...", file_name);
   Msg (s);

   XClearWindow (mainDisplay, drawWindow);

   while (ReadObj (fp, &obj_ptr))
      if (obj_ptr != NULL)
      {
         AddObj (NULL, topObj, obj_ptr);
         if (PointInBBox (obj_ptr->x, obj_ptr->y, drawWinBBox) ||
               BBoxIntersect (obj_ptr->bbox, drawWinBBox))
            DrawObj (drawWindow, obj_ptr);
      }

   if (topObj == NULL)
   {
      DrawPaperBoundary ();
      RedrawGridLines ();
   }

   fclose (fp);
   strcpy (curFileName, file_name);
   curFileDefined = TRUE;

   fileModified = FALSE;
   sprintf (s, "Current file is '%s'.", file_name);
   Msg (s);

   RedrawTitleWindow ();
}

void DumpSymOutline (FP, ObjPtr)
   FILE				* FP;
   register struct ObjRec	* ObjPtr;
{
   int  ltx, lty, rbx, rby;

   ltx = ObjPtr->obbox.ltx - QUARTER_INCH + 1;
   lty = ObjPtr->obbox.lty - QUARTER_INCH + 1;
   rbx = ObjPtr->obbox.rbx + QUARTER_INCH - 1;
   rby = ObjPtr->obbox.rby + QUARTER_INCH - 1;

   fprintf (FP, "gsave\n");
   fprintf (FP, "   [4 4] 0 setdash\n");
   fprintf (FP, "   newpath\n   %1d %1d moveto\n", ltx, lty);
   fprintf (FP, "      %1d %1d lineto\n", rbx, lty);
   fprintf (FP, "      %1d %1d lineto\n", rbx, rby);
   fprintf (FP, "      %1d %1d lineto\n", ltx, rby);
   fprintf (FP, "   closepath stroke\n");
   fprintf (FP, "grestore\n");
}

void DumpAttrs (FP, AttrPtr)
   FILE				* FP;
   register struct AttrRec	* AttrPtr;
{
   for ( ; AttrPtr != NULL; AttrPtr = AttrPtr->prev)
      if (AttrPtr->shown)
         DumpTextObj (FP, AttrPtr->obj);
}

void DumpAllObj (FP, ObjPtr)
   FILE				* FP;
   register struct ObjRec	* ObjPtr;
{
   register struct ObjRec	* obj_ptr;

   switch (ObjPtr->type)
   {
      case OBJ_POLY:
         DumpPolyObj (FP, ObjPtr);
         DumpAttrs (FP, ObjPtr->detail.p->lattr);
         break;
      case OBJ_BOX: DumpBoxObj (FP, ObjPtr); break;
      case OBJ_OVAL: DumpOvalObj (FP, ObjPtr); break;
      case OBJ_TEXT: DumpTextObj (FP, ObjPtr); break;
      case OBJ_POLYGON: DumpPolygonObj (FP, ObjPtr); break;
      case OBJ_SYM:
      case OBJ_ICON:
      case OBJ_GROUP:
         obj_ptr = ObjPtr->detail.r->last;
         for ( ; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
            DumpAllObj (FP, obj_ptr);
         DumpAttrs (FP, ObjPtr->detail.r->lattr);
         if (ObjPtr->type == OBJ_SYM) DumpSymOutline (FP, ObjPtr);
         break;
   }
}

void Dump ()
{
   register struct ObjRec	* obj_ptr;
   char				cmd[256], tmp_str[256];
   int				i, len;
   FILE				* fp;

   if (botObj == NULL) { Msg ("No object to print."); return; }

   Msg ("Generating print file ...");

   system ("rm -f /tmp/DRAW.PS /tmp/DUMP.PS");
   if ((fp = fopen ("/tmp/DUMP.PS", "w")) == NULL)
   {
      Msg ("Can not create /tmp/DUMP.PS, print aborted.");
      return;
   }

   for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
      DumpAllObj (fp, obj_ptr);

   fprintf (fp, "grestore\n\n");
   fprintf (fp, "showpage\n");
   fclose (fp);

   switch (pageStyle)
   {
      case PORTRAIT:
         sprintf (cmd, "cat %s/.psmac %s/.port /tmp/DUMP.PS > /tmp/DRAW.PS",
               drawPath, drawPath);
         break;
      case LANDSCAPE:
         sprintf (cmd, "cat %s/.psmac %s/.land /tmp/DUMP.PS > /tmp/DRAW.PS",
               drawPath, drawPath);
         break;
      case HIGHPORT:
         sprintf (cmd, "cat %s/.psmac %s/.hiport /tmp/DUMP.PS > /tmp/DRAW.PS",
               drawPath, drawPath);
         break;
      case HIGHLAND:
         sprintf (cmd, "cat %s/.psmac %s/.hiland /tmp/DUMP.PS > /tmp/DRAW.PS",
               drawPath, drawPath);
         break;
      case SLIDEPORT:
         sprintf (cmd, "cat %s/.psmac %s/.slideport /tmp/DUMP.PS > /tmp/DRAW.PS",
               drawPath, drawPath);
         break;
      case SLIDELAND:
         sprintf (cmd, "cat %s/.psmac %s/.slideland /tmp/DUMP.PS > /tmp/DRAW.PS",
               drawPath, drawPath);
         break;
   }
   system (cmd);
   chmod ("/tmp/DRAW.PS", 0777);
   chmod ("/tmp/DUMP.PS", 0777);

   switch (whereToPrint)
   {
      case PRINTER:
         system ("lpr /tmp/DRAW.PS &");
         Msg ("Printing in background.");
         break;
      case LATEX_FIG:
         if (!curFileDefined)
         {
            Dialog ("No current file.  Can not generate LaTeX output!", cmd);
            return;
         }

         strcpy (tmp_str, curDomainName);
         if (*curDomainName != '\0') strcat (tmp_str, "/");
         strcat (tmp_str, curFileName);
         len = strlen (tmp_str);
         for (i = len-1; tmp_str[i] != '.'; i--) ;
         strcpy (&tmp_str[i], ".ps");
         if ((fp = fopen (tmp_str, "w")) == NULL)
         {
            sprintf (cmd, "Can not open %s, file not saved.", tmp_str);
            Msg (cmd);
            return;
         }
         fclose (fp);
/*       tmp_str[i] = '\0'; */
         sprintf (cmd, "Printing into '%s'.", tmp_str);
         Msg (cmd);
         sprintf (cmd, "rm -f %s; tgif2ps < /tmp/DRAW.PS > %s; chmod 660 %s",
               tmp_str, tmp_str, tmp_str);
         system (cmd);
         break;
   }
}

void NewProc ()
{
   char	s[80];

   if (fileModified)
   {
      switch (YesNoCancel ("File modified, save file before clear? [ync](y)"))
      {
         case CONFIRM_YES: SaveFile (); break;
         case CONFIRM_NO: TieLooseEnds (); break;
         case CONFIRM_CANCEL: return;
      }
      fileModified = FALSE;
   }
   CleanUpDrawingWindow ();
   ClearFileInfo ();
   ClearAndRedrawDrawWindow ();
   Msg ("Editing no file.");
   objId = 0;
   RedrawTitleWindow ();
}

void OpenProc ()
{
   if (fileModified)
   {
      switch (YesNoCancel ("File modified, save file before open? [ync](y)"))
      {
         case CONFIRM_YES: SaveFile (); break;
         case CONFIRM_NO: break;
         case CONFIRM_CANCEL: return;
      }
   }
   LoadFile ();
}

#define FILE_NEW 0
#define FILE_OPEN 1
#define FILE_SAVE 2
#define FILE_SAVENEW 3
#define FILE_IMPORT 4
#define FILE_DOMAIN 5
#define FILE_DUMP 6
#define FILE_SOLVE 7
#define FILE_SIMULATE 8
#define FILE_PROBE 9
#define FILE_ANIMATE 10
#define FILE_ESCAPE 11
#define FILE_QUIT 12

#define FILEMENUENTRIES 13

char * fileMenuStr[] =
      { "New          ^N",
        "Open         ^O",
        "Save         ^S",
        "SaveNew     ^#S",
        "Import       #P",
        "ChangeDomain ^C",
        "Print        ^P",
        "Solve        #S",
        "Simulate     #Y",
        "Probe        #B",
        "Animate      ^Z",
        "Escape       #X",
        "Quit         ^Q"
      };

int QuitProc ()
{
   if (fileModified)
   {
      switch (YesNoCancel ("File modified, save file before quit? [ync](y)"))
      {
         case CONFIRM_YES: SaveFile (); break;
         case CONFIRM_NO: break;
         case CONFIRM_CANCEL: return (INVALID);
      }
   }
   if (AncesterModified ())
   {
      switch (YesNoCancel ("Ancester file modified, still quitting? [ync](y)"))
      {
         case CONFIRM_YES: return (FILE_QUIT);
         case CONFIRM_NO: return (INVALID);
         case CONFIRM_CANCEL: return (INVALID);
      }
   }
   return (FILE_QUIT);
}

int SolveProc ()
{
   switch (SaveTmpFile ())
   {
      case OBJ_FILE_SAVED: return (FILE_SOLVE);
      case SYM_FILE_SAVED: return (INVALID);
      case INVALID: return (INVALID);
   }
}

int SimulateProc ()
{
   switch (SaveTmpFile ())
   {
      case OBJ_FILE_SAVED: return (FILE_SIMULATE);
      case SYM_FILE_SAVED: return (INVALID);
      case INVALID: return (INVALID);
   }
}

int ProbeProc ()
{
   switch (SaveTmpFile ())
   {
      case OBJ_FILE_SAVED: return (FILE_PROBE);
      case SYM_FILE_SAVED: return (INVALID);
      case INVALID: return (INVALID);
   }
}

int AnimateProc ()
{
   switch (SaveTmpFile ())
   {
      case OBJ_FILE_SAVED: return (FILE_ANIMATE);
      case SYM_FILE_SAVED: return (INVALID);
      case INVALID: return (INVALID);
   }
}

int EscapeProc ()
{
   return (FILE_ESCAPE);
}

int FileMenu (X, Y)
   int	X, Y;
{
   register int index;
   int		* fore_colors, * valid;

   DefaultColorArrays (FILEMENUENTRIES, &fore_colors, &valid);
   index = TextMenuLoop (X, Y, fileMenuStr, FILEMENUENTRIES, fore_colors,
         valid, SINGLECOLOR);
   switch (index)
   {
      case FILE_NEW: NewProc (); break;
      case FILE_OPEN: OpenProc (); break;
      case FILE_SAVE: SaveFile (); break;
      case FILE_SAVENEW: SaveNewFile (); break;
      case FILE_IMPORT: ImportFile (); break;
      case FILE_DOMAIN: ChangeDomain (); break;
      case FILE_DUMP: Dump (); break;
      case FILE_SOLVE: return (SolveProc ()); break;
      case FILE_SIMULATE: return (SimulateProc ()); break;
      case FILE_PROBE: return (ProbeProc ()); break;
      case FILE_ANIMATE: return (AnimateProc ()); break;
      case FILE_ESCAPE: return (EscapeProc ()); break;
      case FILE_QUIT: return (QuitProc ());
   }
   return (INVALID);
}
