/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/grid.c,v 1.1 90/04/01 22:16:18 william Exp $";
#endif

#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "choice.e"
#include "color.e"
#include "drawing.e"
#include "dup.e"
#include "msg.e"
#include "obj.e"
#include "raster.e"
#include "ruler.e"
#include "scroll.e"
#include "select.e"
#include "setup.e"

#define OFFSET_X(x) (((x) - drawOrigX) >> zoomScale)
#define OFFSET_Y(y) (((y) - drawOrigY) >> zoomScale)

int	gridOn = TRUE;
int	xyGrid = DEFAULT_GRID;
int	curScale = 1;
int	pageStyle = PORTRAIT;
int	whereToPrint = PRINTER;

static int	gridShown = TRUE;

void MyHDotLine (Win, Y, XStart, XEnd, Pixel)
   Window	Win;
   int		Y, XStart, XEnd, Pixel;
{
   register int	x;

   for (x = XStart; x < XEnd; x += 8)
/*    XDrawPoint (mainDisplay, Win, defaultGC, x, Y); */
/* Hack to run on RTs -- stupid IBM */
      XDrawLine (mainDisplay, Win, defaultGC, x, Y, x, Y);
}

void MyVDotLine (Win, X, YStart, YEnd, Pixel)
   Window	Win;
   int		X, YStart, YEnd, Pixel;
{
   register int	y;

   for (y = YStart; y < YEnd; y += 8)
/*    XDrawPoint (mainDisplay, Win, defaultGC, X, y); */
/* Hack to run on RTs -- stupid IBM */
      XDrawLine (mainDisplay, Win, defaultGC, X, y, X, y);
}

void RedrawGridLines ()
{
   register int	i, inc = HALF_INCH, abs_grid = HALF_INCH<<zoomScale;
   int		x_start, y_start, x_end, y_end, x_grid_start, y_grid_start;
   int		pixel = myWhitePixel;

   if (!gridShown) return;

   if (drawOrigX % abs_grid == 0)
      x_start = 0;
   else
      x_start = OFFSET_X(((int)(drawOrigX / abs_grid) + 1) * abs_grid);
   if (drawOrigY % abs_grid == 0)
      y_start = 0;
   else
      y_start = OFFSET_Y(((int)(drawOrigY / abs_grid) + 1) * abs_grid);

   x_end = min(OFFSET_X(paperWidth), OFFSET_X(drawOrigX+drawWinW));
   y_end = min(OFFSET_Y(paperHeight), OFFSET_Y(drawOrigY+drawWinH));

   if (drawOrigX % 8 == 0)
      x_grid_start = 0;
   else
      x_grid_start = ((int)(drawOrigX / 8) + 1) * 8 - drawOrigX;
   if (drawOrigY % 8 == 0)
      y_grid_start = 0;
   else
      y_grid_start = ((int)(drawOrigY / 8) + 1) * 8 - drawOrigY;

   for (i = x_start; i < x_end; i += inc)
      MyVDotLine (drawWindow, i, y_grid_start, y_end, pixel);
   for (i = y_start; i < y_end; i += inc)
      MyHDotLine (drawWindow, i, x_grid_start, x_end, pixel);
}

void DrawGridLines (Win, LtX, LtY, RbX, RbY)
   Window	Win;
   int 		LtX, LtY, RbX, RbY; /* absolute coordinates */
{
   register int	i, inc = HALF_INCH, abs_grid = HALF_INCH<<zoomScale;
   int		x_start, y_start, x_end, y_end;
   int		x_grid_start, y_grid_start, x_grid_end, y_grid_end;
   int		pixel = myWhitePixel;

   if (!gridShown) return;

   if (LtX > paperWidth || LtY > paperHeight) return;

   if (LtX % abs_grid == 0)
      x_start = OFFSET_X(LtX);
   else
      x_start = OFFSET_X(((int)(LtX / abs_grid) + 1) * abs_grid);
   if (LtY % abs_grid == 0)
      y_start = OFFSET_Y(LtY);
   else
      y_start = OFFSET_Y(((int)(LtY / abs_grid) + 1) * abs_grid);

   if (RbX % abs_grid == 0)
      x_end = min(OFFSET_X(paperWidth),OFFSET_X(RbX));
   else
      x_end = min(OFFSET_X(paperWidth),
            OFFSET_X(((int)(RbX / abs_grid) + 1) * abs_grid));
   if (RbY % abs_grid == 0)
      y_end = min(OFFSET_Y(paperHeight),OFFSET_Y(RbY));
   else
      y_end = min(OFFSET_Y(paperHeight),
            OFFSET_Y(((int)(RbY / abs_grid) + 1) * abs_grid));

   if (LtX % 8 == 0)
      x_grid_start = OFFSET_X(LtX);
   else
      x_grid_start = ((int)(OFFSET_X(LtX) / 8) + 1) * 8;
   if (LtY % 8 == 0)
      y_grid_start = OFFSET_Y(LtY);
   else
      y_grid_start = ((int)(OFFSET_Y(LtY) / 8) + 1) * 8;

   if (RbX % 8 == 0)
      x_grid_end = min(OFFSET_X(paperWidth), OFFSET_X(RbX));
   else
      x_grid_end = min(OFFSET_X(paperWidth), ((int)(OFFSET_X(RbX)/8)+1)*8);
   if (RbY % 8 == 0)
      y_grid_end = min(OFFSET_Y(paperHeight), OFFSET_Y(RbY));
   else
      y_grid_end = min(OFFSET_Y(paperHeight), ((int)(OFFSET_Y(RbY)/8)+1)*8);

   for (i = x_start; i < x_end; i += inc)
      MyVDotLine (Win, i, y_grid_start, y_grid_end, pixel);
   for (i = y_start; i < y_end; i += inc)
      MyHDotLine (Win, i, x_grid_start, x_grid_end, pixel);
}

#define LAYOUT_INCGRID 0
#define LAYOUT_DECGRID 1
#define LAYOUT_GRID 2
#define LAYOUT_SNAP 3
#define LAYOUT_ZOOMIN 4
#define LAYOUT_ZOOMOUT 5
#define LAYOUT_LAND 6
#define LAYOUT_PORT 7
#define LAYOUT_HILAND 8
#define LAYOUT_HIPORT 9
#define LAYOUT_SLIDELAND 10
#define LAYOUT_SLIDEPORT 11
#define LAYOUT_TO_LPR 12
#define LAYOUT_TO_TEXFIG 13

#define MAXLAYOUTMENUS 14

char * layoutMenuStr[] =
{  "+Grid     #I",
   "-Grid     #D",
   "ShowGrid  #G",
   "SnapOff  ^#G",
   "ZoomIn    #Z",
   "ZoomOut   #O",
   "Landscape",
   "Portrait",
   "HighLand",
   "HighPort",
   "SlideLand",
   "SlidePort",
   "Printer",
   "LaTeX-Figure"
};
#define MAX_LAYOUT_STR 13 /* longest string length in layoutMenuStr[] */

static char showGridStr[] = "ShowGrid  #G";
static char hideGridStr[] = "HideGrid  #G";
static char snapOnStr[]   = "SnapOn   ^#G";
static char snapOffStr[]  = "SnapOff  ^#G";

void IncGrid ()
{
   if (!gridOn)
      Msg ("Snap is not on, grid size not changed.");
   else if (xyGrid < HALF_INCH)
   {
      xyGrid *= 2;
      RedrawRulers ();
   }
   else
      Msg ("Already at maximun grid, grid size not changed.");
}

void DecGrid ()
{
   if (!gridOn)
      Msg ("Snap is not on, grid size not changed.");
   else if (xyGrid > 4)
   {
      xyGrid /= 2;
      RedrawRulers ();
   }
   else
      Msg ("Already at minimum grid, grid size not changed.");
}

void ToggleGridShown ()
{
   gridShown = !gridShown;
   ClearAndRedrawDrawWindow ();
}

void ToggleSnapOn ()
{
   gridOn = !gridOn;
   if (gridOn)
      Msg ("Snapping to grid point activated.");
   else
      Msg ("Snapping to grid point disabled.");
   RedrawRulers ();
}

void ZoomIn ()
{
   if (zoomScale == 0)
   {
      Msg ("Already at highest magnification, can no longer zoom in.");
      return;
   }

   zoomScale--;
   drawWinW >>= 1;
   drawWinH >>= 1;

   if (topSel != NULL)
   {
      drawOrigX = min(max(0,paperWidth-drawWinW), max(0,
            ((int)(selObjLtX/HALF_INCH))*HALF_INCH-(PIX_PER_INCH<<zoomScale)));
      drawOrigY = min(max(0,paperHeight-drawWinH), max(0,
            ((int)(selObjLtY/HALF_INCH))*HALF_INCH-(PIX_PER_INCH<<zoomScale)));
   }

   UpdDrawWinBBox ();
   AdjSplineVs ();
   ClearAndRedrawDrawWindow ();
   RedrawRulers ();
   RedrawScrollBars ();
   justDupped = FALSE;
}

void ZoomOut ()
{
   if (paperWidth <= drawWinW && paperHeight <= drawWinH)
   {
      Msg ("Already at paper boundaries, can no longer zoom out.");
      return;
   }

   TieLooseEnds ();
   SetCurChoice (NOTHING);

   zoomScale++;
   drawWinW <<= 1;
   drawWinH <<= 1;

   if (topSel != NULL)
   {
      drawOrigX = min(max(0,paperWidth-drawWinW), max(0,
            ((int)(selObjLtX/HALF_INCH))*HALF_INCH-(PIX_PER_INCH<<zoomScale)));
      drawOrigY = min(max(0,paperHeight-drawWinH), max(0,
            ((int)(selObjLtY/HALF_INCH))*HALF_INCH-(PIX_PER_INCH<<zoomScale)));
   }
   else
   {
      if (drawOrigX + drawWinW > paperWidth)
      {
         if (drawWinW > paperWidth)
            drawOrigX = 0;
         else
            drawOrigX = max(0,paperWidth-drawWinW);
      }
      if (drawOrigY + drawWinH > paperHeight)
      {
         if (drawWinH > paperHeight)
            drawOrigY = 0;
         else
            drawOrigY = max(0,paperHeight-drawWinH);
      }
   }

   UpdDrawWinBBox ();
   AdjSplineVs ();
   ClearAndRedrawDrawWindow ();
   RedrawRulers ();
   RedrawScrollBars ();
   justDupped = FALSE;
}

int UpdPageStyle (NewPageStyle)
   int	NewPageStyle;
{
   int	changed = FALSE, old_w = paperWidth, old_h = paperHeight;

   switch (NewPageStyle)
   {
      case LANDSCAPE:
         pageStyle = LANDSCAPE;
         paperWidth = 10 * PIX_PER_INCH;
         paperHeight = 7.5 * PIX_PER_INCH;
         if (drawOrigX + drawWinW > paperWidth || drawOrigX + drawWinW > old_w)
         {
            if (drawOrigX + drawWinW > paperWidth)
               drawOrigX = max(0,paperWidth-drawWinW);
            changed = TRUE;
         }
         if (drawOrigY + drawWinH > paperHeight || drawOrigY + drawWinH > old_h)
         {
            if (drawOrigY + drawWinH > paperHeight)
               drawOrigY = max(0,paperHeight-drawWinH);
            changed = TRUE;
         }
         RedrawScrollBars ();
         break;
      case PORTRAIT:
         pageStyle = PORTRAIT;
         paperWidth = 7.5 * PIX_PER_INCH;
         paperHeight = 10 * PIX_PER_INCH;
         if (drawOrigX + drawWinW > paperWidth || drawOrigX + drawWinW > old_w)
         {
            if (drawOrigX + drawWinW > paperWidth)
               drawOrigX = max(0,paperWidth-drawWinW);
            changed = TRUE;
         }
         if (drawOrigY + drawWinH > paperHeight || drawOrigY + drawWinH > old_h)
         {
            if (drawOrigY + drawWinH > paperHeight)
               drawOrigY = max(0,paperHeight-drawWinH);
            changed = TRUE;
         }
         RedrawScrollBars ();
         break;
      case HIGHLAND:
         pageStyle = HIGHLAND;
         paperWidth = 21 * PIX_PER_INCH;
         paperHeight = 15.5 * PIX_PER_INCH;
         if (drawOrigX + drawWinW > paperWidth || drawOrigX + drawWinW > old_w)
         {
            if (drawOrigX + drawWinW > paperWidth)
               drawOrigX = max(0,paperWidth-drawWinW);
            changed = TRUE;
         }
         if (drawOrigY + drawWinH > paperHeight || drawOrigY + drawWinH > old_h)
         {
            if (drawOrigY + drawWinH > paperHeight)
               drawOrigY = max(0,paperHeight-drawWinH);
            changed = TRUE;
         }
         RedrawScrollBars ();
         break;
      case HIGHPORT:
         pageStyle = HIGHPORT;
         paperWidth = 15.5 * PIX_PER_INCH;
         paperHeight = 21 * PIX_PER_INCH;
         if (drawOrigX + drawWinW > paperWidth || drawOrigX + drawWinW > old_w)
         {
            if (drawOrigX + drawWinW > paperWidth)
               drawOrigX = max(0,paperWidth-drawWinW);
            changed = TRUE;
         }
         if (drawOrigY + drawWinH > paperHeight || drawOrigY + drawWinH > old_h)
         {
            if (drawOrigY + drawWinH > paperHeight)
               drawOrigY = max(0,paperHeight-drawWinH);
            changed = TRUE;
         }
         RedrawScrollBars ();
         break;
      case SLIDELAND:
         pageStyle = SLIDELAND;
         paperWidth = 5 * PIX_PER_INCH;
         paperHeight = 3.75 * PIX_PER_INCH;
         if (drawOrigX + drawWinW > paperWidth || drawOrigX + drawWinW > old_w)
         {
            if (drawOrigX + drawWinW > paperWidth)
               drawOrigX = max(0,paperWidth-drawWinW);
            changed = TRUE;
         }
         if (drawOrigY + drawWinH > paperHeight || drawOrigY + drawWinH > old_h)
         {
            if (drawOrigY + drawWinH > paperHeight)
               drawOrigY = max(0,paperHeight-drawWinH);
            changed = TRUE;
         }
         RedrawScrollBars ();
         break;
      case SLIDEPORT:
         pageStyle = SLIDEPORT;
         paperWidth = 3.75 * PIX_PER_INCH;
         paperHeight = 5 * PIX_PER_INCH;
         if (drawOrigX + drawWinW > paperWidth || drawOrigX + drawWinW > old_w)
         {
            if (drawOrigX + drawWinW > paperWidth)
               drawOrigX = max(0,paperWidth-drawWinW);
            changed = TRUE;
         }
         if (drawOrigY + drawWinH > paperHeight || drawOrigY + drawWinH > old_h)
         {
            if (drawOrigY + drawWinH > paperHeight)
               drawOrigY = max(0,paperHeight-drawWinH);
            changed = TRUE;
         }
         RedrawScrollBars ();
         break;
   }
   return (changed);
}

void LayoutMenu (X, Y)
   int	X, Y;
{
   int		index, changed = FALSE, * fore_colors, * valid;
   register int	i;
   char		* menu_strings[MAXLAYOUTMENUS], * str_ptr;

   str_ptr = (char *) calloc (MAX_LAYOUT_STR*MAXLAYOUTMENUS, sizeof(char));
   for (i = 0; i < MAXLAYOUTMENUS; i++, str_ptr += MAX_LAYOUT_STR)
   {
      menu_strings[i] = str_ptr;
      strcpy (menu_strings[i], layoutMenuStr[i]);
   }

   if (gridShown)
      strcpy (menu_strings[LAYOUT_GRID], hideGridStr);
   else
      strcpy (menu_strings[LAYOUT_GRID], showGridStr);
   if (gridOn)
      strcpy (menu_strings[LAYOUT_SNAP], snapOffStr);
   else
      strcpy (menu_strings[LAYOUT_SNAP], snapOnStr);

   DefaultColorArrays (MAXLAYOUTMENUS, &fore_colors, &valid);
   index = TextMenuLoop (X, Y, menu_strings, MAXLAYOUTMENUS,
         fore_colors, valid, SINGLECOLOR);

   cfree (menu_strings[0]);

   switch (index)
   {
      case LAYOUT_INCGRID: IncGrid (); break;
      case LAYOUT_DECGRID: DecGrid (); break;
      case LAYOUT_GRID: ToggleGridShown (); break;
      case LAYOUT_SNAP: ToggleSnapOn (); break;
      case LAYOUT_ZOOMIN: ZoomIn (); break;
      case LAYOUT_ZOOMOUT: ZoomOut (); break;
      case LAYOUT_LAND:
         if (pageStyle != LANDSCAPE)
         {
            if (UpdPageStyle (LANDSCAPE))
            {
               UpdDrawWinBBox ();
               ClearAndRedrawDrawWindow ();
            }
            RedrawRulers ();
            fileModified = TRUE;
            Msg ("Page style changed to Landscape.");
         }
         break;
      case LAYOUT_PORT:
         if (pageStyle != PORTRAIT)
         {
            if (UpdPageStyle (PORTRAIT))
            {
               UpdDrawWinBBox ();
               ClearAndRedrawDrawWindow ();
            }
            RedrawRulers ();
            fileModified = TRUE;
            Msg ("Page style changed to Portrait.");
         }
         break;
      case LAYOUT_HILAND:
         if (pageStyle != HIGHLAND)
         {
            if (UpdPageStyle (HIGHLAND))
            {
               UpdDrawWinBBox ();
               ClearAndRedrawDrawWindow ();
            }
            RedrawRulers ();
            fileModified = TRUE;
            Msg ("Page style changed to HighLand.");
         }
         break;
      case LAYOUT_HIPORT:
         if (pageStyle != HIGHPORT)
         {
            if (UpdPageStyle (HIGHPORT))
            {
               UpdDrawWinBBox ();
               ClearAndRedrawDrawWindow ();
            }
            RedrawRulers ();
            fileModified = TRUE;
            Msg ("Page style changed to HighPort.");
         }
         break;
      case LAYOUT_SLIDELAND:
         if (pageStyle != SLIDELAND)
         {
            if (UpdPageStyle (SLIDELAND))
            {
               UpdDrawWinBBox ();
               ClearAndRedrawDrawWindow ();
            }
            RedrawRulers ();
            fileModified = TRUE;
            Msg ("Page style changed to SlideLand.");
         }
         break;
      case LAYOUT_SLIDEPORT:
         if (pageStyle != SLIDEPORT)
         {
            if (UpdPageStyle (SLIDEPORT))
            {
               UpdDrawWinBBox ();
               ClearAndRedrawDrawWindow ();
            }
            RedrawRulers ();
            fileModified = TRUE;
            Msg ("Page style changed to SlidePort.");
         }
         break;
      case LAYOUT_TO_LPR:
         if (whereToPrint != PRINTER)
         {
            whereToPrint = PRINTER;
            ShowWhereToPrint ();
            Msg ("Print device set to printer.");
         }
         break;
      case LAYOUT_TO_TEXFIG:
         if (whereToPrint != LATEX_FIG)
         {
            whereToPrint = LATEX_FIG;
            ShowWhereToPrint ();
            Msg ("Print device set to LaTeX-Figure file.");
         }
         break;
   }
}

void GridXY (X, Y, GridX, GridY)
   int	X, Y, * GridX, * GridY;
{
   register int	dx, dy;

   if (gridOn)
   {
      dx = (X+drawOrigX) % xyGrid;
      if (dx < xyGrid / 2)
         *GridX = X - dx;
      else
         *GridX = X - dx + xyGrid;
      dy = (Y+drawOrigY) % xyGrid;
      if (dy < xyGrid / 2)
         *GridY = Y - dy;
      else
         *GridY = Y - dy + xyGrid;
   }
   else
   {
      *GridX = X;
      *GridY = Y;
   }
}
