/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/mark.c,v 1.1 90/04/01 22:16:21 william Exp $";
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "raster.e"
#include "setup.e"
#include "select.e"

#define OFFSET_X(x) (((x) - drawOrigX) >> zoomScale)
#define OFFSET_Y(y) (((y) - drawOrigY) >> zoomScale)

#define FORWARD 0
#define REVERSE 1

#define Mark(X,Y) \
      XFillRectangle(mainDisplay,drawWindow,revDefaultGC,(X)-2,(Y)-2,5,5)

static
void MarkPoly (NumPts, V)
   int		NumPts;
   XPoint	* V;
{
   register int	i;

   for (i = 0; i < NumPts; i++)
      Mark (OFFSET_X(V[i].x), OFFSET_Y(V[i].y));
}

static
void Mark4Corners (BBox)
   struct BBRec	BBox;
{
   Mark (OFFSET_X(BBox.ltx), OFFSET_Y(BBox.lty));
   Mark (OFFSET_X(BBox.ltx), OFFSET_Y(BBox.rby));
   Mark (OFFSET_X(BBox.rbx), OFFSET_Y(BBox.lty));
   Mark (OFFSET_X(BBox.rbx), OFFSET_Y(BBox.rby));
}

static
void Mark8Places (BBox)
   struct BBRec	BBox;
{
   register int	xmid, ymid;

   if ((BBox.rbx - BBox.ltx) >> zoomScale >= 10)
   {
      xmid = (BBox.ltx+BBox.rbx) / 2;
      Mark (OFFSET_X(xmid), OFFSET_Y(BBox.lty));
      Mark (OFFSET_X(xmid), OFFSET_Y(BBox.rby));
   }
   if ((BBox.rby - BBox.lty) >> zoomScale >= 10)
   {
      ymid = (BBox.lty+BBox.rby) / 2;
      Mark (OFFSET_X(BBox.ltx), OFFSET_Y(ymid));
      Mark (OFFSET_X(BBox.rbx), OFFSET_Y(ymid));
   }
   Mark4Corners (BBox);
}

void HighLightAnObj (ObjPtr)
   register struct ObjRec	* ObjPtr;
{
   switch (ObjPtr->type)
   {
      case OBJ_POLY:
         MarkPoly (ObjPtr->detail.p->n, ObjPtr->detail.p->vlist); break;
      case OBJ_POLYGON:
         MarkPoly (ObjPtr->detail.g->n-1, ObjPtr->detail.g->vlist); break;

      case OBJ_BOX:
      case OBJ_OVAL:
      case OBJ_GROUP:
      case OBJ_SYM: Mark8Places (ObjPtr->obbox); break;

      case OBJ_TEXT:
      case OBJ_ICON: Mark4Corners (ObjPtr->obbox); break;
   }
}

static
void HighLight (Dir)
   int	Dir;
{
   register struct SelRec	* sel_ptr;
   register struct ObjRec	* obj_ptr;

   switch (Dir)
   {
      case FORWARD: sel_ptr = botSel; break;
      case REVERSE: sel_ptr = topSel; break;
   }
   while (sel_ptr != NULL)
   {
      obj_ptr = sel_ptr->obj;
      switch (obj_ptr->type)
      {
         case OBJ_POLY:
            MarkPoly (obj_ptr->detail.p->n, obj_ptr->detail.p->vlist);
            break;
         case OBJ_POLYGON:
            MarkPoly (obj_ptr->detail.g->n-1, obj_ptr->detail.g->vlist);
            break;

         case OBJ_BOX:
         case OBJ_OVAL:
         case OBJ_GROUP:
         case OBJ_SYM: Mark8Places (obj_ptr->obbox); break;

         case OBJ_TEXT:
         case OBJ_ICON: Mark4Corners (obj_ptr->obbox); break;
      }
      switch (Dir)
      {
         case FORWARD: sel_ptr = sel_ptr->prev; break;
         case REVERSE: sel_ptr = sel_ptr->next; break;
      }
   }
}

void HighLightForward ()
{
   HighLight (FORWARD);
}
 
void HighLightReverse ()
{
   HighLight (REVERSE);
}
