/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */

#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/move.c,v 1.2 90/05/17 09:17:01 william Exp $";
#endif
#include <stdio.h>
#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "attr.e"
#include "cursor.e"
#include "drawing.e"
#include "dup.e"
#include "grid.e"
#include "obj.e"
#include "oval.e"
#include "raster.e"
#include "ruler.e"
#include "select.e"
#include "setup.e"

#define OFFSET_X(x) (((x) - drawOrigX) >> zoomScale)
#define OFFSET_Y(y) (((y) - drawOrigY) >> zoomScale)

static
void MovePoly (ObjPtr, Dx, Dy)
   struct ObjRec	* ObjPtr;
   int			Dx, Dy;
{
   register int	i;

   for (i = 0; i < ObjPtr->detail.p->n; i++)
   {
      ObjPtr->detail.p->vlist[i].x += Dx;
      ObjPtr->detail.p->vlist[i].y += Dy;
   }
   if (ObjPtr->detail.p->curved)
      for (i = 0; i < ObjPtr->detail.p->sn; i++)
      {
         ObjPtr->detail.p->svlist[i].x += (Dx>>zoomScale);
         ObjPtr->detail.p->svlist[i].y += (Dy>>zoomScale);
      }
}

static
void MovePolygon (ObjPtr, Dx, Dy)
   struct ObjRec	* ObjPtr;
   int			Dx, Dy;
{
   register int	i;

   for (i = 0; i < ObjPtr->detail.g->n; i++)
   {
      ObjPtr->detail.g->vlist[i].x += Dx;
      ObjPtr->detail.g->vlist[i].y += Dy;
   }
   if (ObjPtr->detail.g->curved)
      for (i = 0; i < ObjPtr->detail.g->sn; i++)
      {
         ObjPtr->detail.g->svlist[i].x += Dx;
         ObjPtr->detail.g->svlist[i].y += Dy;
      }
}

void MoveObj (ObjPtr, Dx, Dy)
   struct ObjRec	* ObjPtr;
   int			Dx, Dy;
{
   struct ObjRec	* ptr;

   ObjPtr->x += Dx;
   ObjPtr->y += Dy;
   ObjPtr->bbox.ltx += Dx;
   ObjPtr->bbox.lty += Dy;
   ObjPtr->bbox.rbx += Dx;
   ObjPtr->bbox.rby += Dy;
   ObjPtr->obbox.ltx += Dx;
   ObjPtr->obbox.lty += Dy;
   ObjPtr->obbox.rbx += Dx;
   ObjPtr->obbox.rby += Dy;
   switch (ObjPtr->type)
   {
      case OBJ_POLY:
         MoveAttrs(ObjPtr->detail.p->fattr, Dx,Dy);
         MovePoly (ObjPtr, Dx, Dy); 
         break;
      case OBJ_BOX : break;
      case OBJ_OVAL: break;
      case OBJ_TEXT: break;
      case OBJ_POLYGON: MovePolygon (ObjPtr, Dx, Dy); break;
      case OBJ_SYM:
      case OBJ_ICON:
      case OBJ_GROUP:
         MoveAttrs(ObjPtr->detail.r->fattr, Dx,Dy);
         for (ptr = ObjPtr->detail.r->first; ptr != NULL; ptr = ptr->next)
            MoveObj (ptr, Dx, Dy);
         break;
   }
}

void MoveAllSel (Dx, Dy)
   register int	Dx, Dy;
{
   register struct ObjRec	* obj_ptr;
   struct SelRec		* sel_ptr;

   for (sel_ptr = topSel; sel_ptr != NULL; sel_ptr = sel_ptr->next)
      MoveObj (sel_ptr->obj, Dx, Dy);
}

void MoveSel (OrigX, OrigY, ObjPtr)
   int 			OrigX, OrigY;
   struct ObjRec	* ObjPtr;
{
   register int		i, num_pts;
   struct OvalRec	* oval_ptr;
   struct PolyRec	* poly_ptr;
   struct PolygonRec	* polygon_ptr;
   struct BBRec		bbox, o_bbox;
   Window		root_win, child_win;
   int			root_x, root_y;
   XPoint		* v;
   unsigned int		status;
   XButtonEvent		input;
   int 			ltx, lty, rbx, rby, sel_ltx, sel_lty, sel_rbx, sel_rby;
   int			x, y, moving = TRUE, dx, dy;
   int			grid_x = OrigX, grid_y = OrigY, new_grid_x, new_grid_y;

   sel_ltx = OFFSET_X(selLtX) - 1; sel_lty = OFFSET_Y(selLtY) - 1;
   sel_rbx = OFFSET_X(selRbX) + 1; sel_rby = OFFSET_Y(selRbY) + 1;

   SelBox (drawWindow, revDefaultGC, sel_ltx, sel_lty, sel_rbx, sel_rby);

   switch (ObjPtr->type)
   {
      case OBJ_BOX:
      case OBJ_TEXT:
         ltx = OFFSET_X(ObjPtr->obbox.ltx); lty = OFFSET_Y(ObjPtr->obbox.lty);
         rbx = OFFSET_X(ObjPtr->obbox.rbx); rby = OFFSET_Y(ObjPtr->obbox.rby);
         SelBox (drawWindow, revDefaultGC, ltx, lty, rbx, rby);
         break;
      case OBJ_OVAL:
         oval_ptr = ObjPtr->detail.o;
         num_pts = 13;
         bbox.ltx = OFFSET_X(ObjPtr->obbox.ltx);
         bbox.lty = OFFSET_Y(ObjPtr->obbox.lty);
         bbox.rbx = OFFSET_X(ObjPtr->obbox.rbx);
         bbox.rby = OFFSET_Y(ObjPtr->obbox.rby);
         MyOval (drawWindow, revDefaultGC, bbox);
         break;
      case OBJ_POLY:
         poly_ptr = ObjPtr->detail.p;
         num_pts = poly_ptr->n;
         v = (XPoint *) calloc (poly_ptr->n, sizeof(XPoint));
         for (i = 0; i < num_pts; i++)
         {
            v[i].x = OFFSET_X(poly_ptr->vlist[i].x);
            v[i].y = OFFSET_Y(poly_ptr->vlist[i].y);
         }
         XDrawLines (mainDisplay, drawWindow, revDefaultGC, v, num_pts,
               CoordModeOrigin);
         break;
      case OBJ_POLYGON:
         polygon_ptr = ObjPtr->detail.g;
         num_pts = polygon_ptr->n;
         v = (XPoint *) calloc (polygon_ptr->n, sizeof(XPoint));
         for (i = 0; i < num_pts; i++)
         {
            v[i].x = OFFSET_X(polygon_ptr->vlist[i].x);
            v[i].y = OFFSET_Y(polygon_ptr->vlist[i].y);
         }
         XDrawLines (mainDisplay, drawWindow, revDefaultGC, v, num_pts,
               CoordModeOrigin);
         break;
      case OBJ_GROUP:
      case OBJ_SYM:
      case OBJ_ICON:
         ltx = OFFSET_X(ObjPtr->obbox.ltx); lty = OFFSET_Y(ObjPtr->obbox.lty);
         rbx = OFFSET_X(ObjPtr->obbox.rbx); rby = OFFSET_Y(ObjPtr->obbox.rby);
         SelBox (drawWindow, revDefaultGC, ltx, lty, rbx, rby);
         break;
   }

   XGrabPointer (mainDisplay, drawWindow, FALSE, ButtonReleaseMask,
         GrabModeAsync, GrabModeAsync, None, defaultCursor, CurrentTime);

   dx = dy = 0;

   while (moving)
   {
      XQueryPointer (mainDisplay, drawWindow, &root_win, &child_win,
            &root_x, &root_y, &x, &y, &status);

      GridXY (x, y, &new_grid_x, &new_grid_y);
      if (new_grid_x==grid_x && new_grid_y==grid_y && (status & Button1Mask)!=0)
         continue;

      grid_x = new_grid_x;
      grid_y = new_grid_y;

      switch (ObjPtr->type)
      {
         case OBJ_BOX:
         case OBJ_TEXT:
            SelBox (drawWindow, revDefaultGC, ltx+dx, lty+dy, rbx+dx, rby+dy);
            break;
         case OBJ_OVAL:
            o_bbox.ltx = bbox.ltx + dx; o_bbox.lty = bbox.lty + dy;
            o_bbox.rbx = bbox.rbx + dx; o_bbox.rby = bbox.rby + dy;
            MyOval (drawWindow, revDefaultGC, o_bbox);
            break;
         case OBJ_POLY:
         case OBJ_POLYGON:
            for (i = 0; i < num_pts; i++)
            {
               v[i].x += dx;
               v[i].y += dy;
            }
            XDrawLines (mainDisplay, drawWindow, revDefaultGC, v, num_pts,
                  CoordModeOrigin);
            for (i = 0; i < num_pts; i++)
            {
               v[i].x -= dx;
               v[i].y -= dy;
            }
            break;
         case OBJ_GROUP:
         case OBJ_SYM:
         case OBJ_ICON:
            SelBox (drawWindow, revDefaultGC, ltx+dx, lty+dy, rbx+dx, rby+dy);
            break;
      }
      SelBox (drawWindow, revDefaultGC, sel_ltx+dx, sel_lty+dy, sel_rbx+dx,
            sel_rby+dy);

      dx = grid_x - OrigX;
      dy = grid_y - OrigY;

      if ((status & Button1Mask) == 0)
      {
         /* mouse is up */
         XUngrabPointer (mainDisplay, CurrentTime);
         MarkRulers (grid_x, grid_y);
         moving = FALSE;
         switch (ObjPtr->type)
         {
            case OBJ_POLY:
            case OBJ_POLYGON: cfree(v); break;
         }
         if (dx != 0 || dy != 0)
         {
            HighLightReverse ();
            dx <<= zoomScale;
            dy <<= zoomScale;
            MoveAllSel (dx, dy);
            RedrawAreas (botObj, selLtX-(1<<zoomScale), selLtY-(1<<zoomScale),
                  selRbX+(1<<zoomScale), selRbY+(1<<zoomScale),
                  selLtX-(1<<zoomScale)+dx, selLtY-(1<<zoomScale)+dy,
                  selRbX+(1<<zoomScale)+dx, selRbY+(1<<zoomScale)+dy);
            HighLightForward ();
            UpdSelBBox ();
            if (justDupped)
            {
               dupDx += dx;
               dupDy += dy;
            }
            fileModified = TRUE;
         }
      }
      else
      {
         MarkRulers (grid_x, grid_y);
         SelBox (drawWindow, revDefaultGC, sel_ltx+dx, sel_lty+dy, sel_rbx+dx,
               sel_rby+dy);
         switch (ObjPtr->type)
         {
            case OBJ_BOX:
            case OBJ_TEXT:
               SelBox (drawWindow, revDefaultGC, ltx+dx, lty+dy, rbx+dx,
                     rby+dy);
               break;
            case OBJ_OVAL:
               o_bbox.ltx = bbox.ltx + dx; o_bbox.lty = bbox.lty + dy;
               o_bbox.rbx = bbox.rbx + dx; o_bbox.rby = bbox.rby + dy;
               MyOval (drawWindow, revDefaultGC, o_bbox);
               break;
            case OBJ_POLY:
            case OBJ_POLYGON:
               for (i = 0; i < num_pts; i++)
               {
                  v[i].x += dx;
                  v[i].y += dy;
               }
               XDrawLines (mainDisplay, drawWindow, revDefaultGC, v, num_pts,
                     CoordModeOrigin);
               for (i = 0; i < num_pts; i++)
               {
                  v[i].x -= dx;
                  v[i].y -= dy;
               }
               break;
            case OBJ_GROUP:
            case OBJ_SYM:
            case OBJ_ICON:
               SelBox (drawWindow, revDefaultGC, ltx+dx, lty+dy, rbx+dx,
                     rby+dy);
               break;
         }
      }
   }
}
