/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/names.c,v 1.1 90/04/01 22:16:26 william Exp $";
#endif

#include <sys/types.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "const.h"
#include "types.h"

#include "box.e"
#include "button.e"
#include "cursor.e"
#include "font.e"
#include "mainloop.e"
#include "raster.e"
#include "scroll.e"
#include "setup.e"

#include "bitmaps/uparrow.bm"
#include "bitmaps/downarrow.bm"

#define ITEM_DSPED 10
#define ITEM_LEN 30
#define ROW_HEIGHT (defaultFontHeight+1)

char	curDomainName[MAXPATHLENGTH];

static Window		nameBaseWin;
static Window		nameDspWin;
static Window		nameScrollWin;
static int		nameEntries;
static char		* * nameDspPtr;
static int		nameFirst;
static int		nameMarked;
static struct BBRec	buttonBBox[2];

static int	nameDspH;		/* height of display area */
static int	nameDspW;		/* width of display area */
static int	nameDspWinH;		/* height of display window */
static int	nameDspWinW;		/* width of display window */
static int	nameScrollAreaH;	/* heighr of the grey scroll area */

static GC	nameGC;
static GC	revNameGC;

void InitNames ()
{
   XGCValues	values;

   values.foreground = myWhitePixel;
   values.background = myBlackPixel;
   values.fill_style = FillSolid;
   values.font = defaultFontPtr->fid;
   nameGC = XCreateGC (mainDisplay, rootWindow,
         GCForeground | GCBackground | GCFillStyle | GCFont, &values);

   values.foreground = myBlackPixel;
   values.background = myWhitePixel;
   revNameGC = XCreateGC (mainDisplay, rootWindow,
         GCForeground | GCBackground | GCFillStyle | GCFont, &values);

   nameDspW = ITEM_LEN * defaultFontWidth;
   nameDspH = ITEM_DSPED * ROW_HEIGHT;
   nameDspWinW = nameDspW + 2 * brdrW;
   nameDspWinH = nameDspH + 2 * brdrW;
   nameScrollAreaH = nameDspH - 2 * uparrow_height;

   *curDomainName = '\0';
}

void CleanUpNames ()
{
   XFreeGC (mainDisplay, nameGC);
   XFreeGC (mainDisplay, revNameGC);
}

static
void RedrawNameScrollWin ()
{
   double        frac, start_frac;
   int          block_h, block_start;
   XGCValues	values;
   XEvent	ev;

   XSync (mainDisplay, FALSE);
   while (XCheckWindowEvent (mainDisplay, vSBarWindow, ExposureMask, &ev)) ;

   start_frac = (double)((double)(nameFirst)/(double)(nameEntries));
   block_start = (int)(nameScrollAreaH * start_frac); /* starting pixel */

   if (nameEntries > ITEM_DSPED)
      frac = (double)((double)ITEM_DSPED / (double)(nameEntries));
   else
      frac = 1.0;

   block_h = (int)(nameScrollAreaH * frac); /* number of pixels */

   values.foreground = myBlackPixel;
   values.background = myWhitePixel;
   values.function = GXcopy;
   values.fill_style = FillSolid;
   XChangeGC (mainDisplay, scrollGC,
         GCForeground | GCBackground | GCFunction | GCFillStyle,
         &values);
   XFillRectangle (mainDisplay, nameScrollWin, scrollGC, 0, scrollBarW,
         scrollBarW, nameScrollAreaH);

   values.foreground = myWhitePixel;
   values.background = myBlackPixel;
   values.fill_style = FillOpaqueStippled;
   values.stipple = patPixmap[SCROLLPAT];
   XChangeGC (mainDisplay, scrollGC,
         GCForeground | GCBackground | GCFillStyle | GCStipple,
         &values);
   XFillRectangle (mainDisplay, nameScrollWin, scrollGC, 0,
         scrollBarW+block_start, scrollBarW, block_h);

   scrollImage->data = upData;
   XPutImage (mainDisplay, nameScrollWin, scrollGC, scrollImage, 0, 0, 0, 0,
         scrollBarW, scrollBarW);
   scrollImage->data = downData;
   XPutImage (mainDisplay, nameScrollWin, scrollGC, scrollImage, 0, 0,
         0, scrollBarW+nameScrollAreaH, scrollBarW, scrollBarW);
}

static
void RedrawDspWindow ()
{
   register int	i;
   int		top, len, end, pixel;
   DspItem	* diptr;
   XGCValues	values;

   top = defaultFontAsc+1;

   if (nameFirst+ITEM_DSPED > nameEntries)
      end = nameEntries;
   else
      end = nameFirst + ITEM_DSPED;

   XFillRectangle (mainDisplay, nameDspWin, revNameGC, 0, 0,
         ITEM_LEN*defaultFontWidth, ITEM_DSPED*ROW_HEIGHT);

   for (i = nameFirst; i < end; i++)
   {
      len = strlen (nameDspPtr[i]);
      if (i == nameMarked)
      {
         XFillRectangle (mainDisplay, nameDspWin, nameGC,
               0, (i-nameFirst)*ROW_HEIGHT, ITEM_LEN*defaultFontWidth,
               ROW_HEIGHT);
         XDrawString (mainDisplay, nameDspWin, revNameGC,
               0, (i-nameFirst)*ROW_HEIGHT+top, nameDspPtr[i], len);
      }
      else
      {
         XFillRectangle (mainDisplay, nameDspWin, revNameGC,
               0, (i-nameFirst)*ROW_HEIGHT, ITEM_LEN*defaultFontWidth,
               ROW_HEIGHT);
         XDrawString (mainDisplay, nameDspWin, nameGC,
               0, (i-nameFirst)*ROW_HEIGHT+top, nameDspPtr[i], len);
      }
   }
}

static
void RedrawNameBaseWindow (Str, str_start, button_start)
   char	* Str;
   int	str_start, button_start;
{
   int	top = defaultFontAsc+2;

   XDrawString (mainDisplay, nameBaseWin, nameGC, str_start,
         ROW_HEIGHT+top, Str, strlen(Str));

   buttonBBox[0].lty = buttonBBox[1].lty = (ITEM_DSPED+4) * ROW_HEIGHT;
   buttonBBox[0].ltx = button_start;
   DisplayButton (nameBaseWin, "OK", 8, &buttonBBox[0], BUTTON_NORMAL);
   buttonBBox[1].ltx = buttonBBox[CONFIRM_YES].rbx + 1 + defaultFontWidth;
   DisplayButton (nameBaseWin, "CANCEL", 8, &buttonBBox[1],
         BUTTON_NORMAL);
}

static
char * * MakeNameDspItemArray (Entries, DLPtr)
   int		Entries;
   DspList	* DLPtr;
{
   register int	i, j, len;
   char		* * dsp_ptr, * c_ptr;

   dsp_ptr = (char * *) calloc (Entries, sizeof(char *));
   c_ptr = (char *) calloc (Entries*MAXPATHLENGTH, sizeof(char));
   for (i = 0; i < Entries; i++, DLPtr = DLPtr->next)
   {
      dsp_ptr[i] = c_ptr;
      len = strlen (DLPtr->itemstr);
      for (j = len; j >= 0 && DLPtr->itemstr[j] != '/'; j--) ;
      if (j >= 0)
         strcpy (c_ptr, (&DLPtr->itemstr[j])+1);
      else
         strcpy (c_ptr, DLPtr->itemstr);
      c_ptr += MAXPATHLENGTH;
   }
   return (dsp_ptr);
}

static
int GetNameEntryNum (RowOffset)
   int		RowOffset;
{
   register int	index;

   index = nameFirst + RowOffset;
   if (index >= nameEntries) return (INVALID);
   return (index);
}

static
void NameScrollHandler (button_ev)
   XButtonEvent	* button_ev;
{
   double	frac, start_frac;
   int		block_h, block_start;

   if (button_ev->y < scrollBarW)
   {
      /* clicked in the uparrow */
      if (nameFirst != 0)
      {
         nameFirst--;
         RedrawNameScrollWin ();
         RedrawDspWindow ();
         return;
      }
   }
   else if (button_ev->y >= scrollBarW+nameScrollAreaH)
   {
      /* clicked in the downarrow */
      if (nameEntries <= ITEM_DSPED) return;

      if (nameFirst+ITEM_DSPED != nameEntries)
      {
         nameFirst++;
         RedrawNameScrollWin ();
         RedrawDspWindow ();
         return;
      }
   }
   else /* clicked in the middle region */
   {
      if (nameEntries <= ITEM_DSPED) return;

      frac = (double)((double)ITEM_DSPED / (double)(nameEntries));
      block_h = (int)(nameScrollAreaH * frac);
      block_start = button_ev->y - scrollBarW;
      start_frac = (double)((double)(block_start)/(double)(nameScrollAreaH));
      if (block_start+block_h >= nameScrollAreaH)
      {
         nameFirst = nameEntries - ITEM_DSPED;
         RedrawNameScrollWin ();
         RedrawDspWindow ();
         return;
      }
      else
      {
         nameFirst = (int)(nameEntries * start_frac);
         RedrawNameScrollWin ();
         RedrawDspWindow ();
         return;
      }
   }
}

static
void NameDspHandler (button_ev)
   XButtonEvent	* button_ev;
{
   int		row_offset, len, top;

   top = defaultFontAsc+1;

   row_offset = (int)(button_ev->y / ROW_HEIGHT);
   if (nameMarked != INVALID && nameMarked == row_offset+nameFirst) return;

   if (nameMarked != INVALID &&
         nameMarked >= nameFirst && nameMarked < nameFirst+ITEM_DSPED)
   {
      len = strlen (nameDspPtr[nameMarked]);
      XFillRectangle (mainDisplay, nameDspWin, revNameGC, 0,
            (nameMarked - nameFirst)*ROW_HEIGHT,
            ITEM_LEN*defaultFontWidth, ROW_HEIGHT);
      XDrawString (mainDisplay, nameDspWin, nameGC, 0,
            (nameMarked - nameFirst)*ROW_HEIGHT+top,
            nameDspPtr[nameMarked], len);
   }

   nameMarked = GetNameEntryNum (row_offset);
   if (nameMarked != INVALID)
   {
      len = strlen (nameDspPtr[nameMarked]);
      XFillRectangle (mainDisplay, nameDspWin, nameGC, 0,
            (nameMarked - nameFirst)*ROW_HEIGHT,
            ITEM_LEN*defaultFontWidth, ROW_HEIGHT);
      XDrawString (mainDisplay, nameDspWin, revNameGC, 0,
            (nameMarked - nameFirst)*ROW_HEIGHT+top,
            nameDspPtr[nameMarked], len);
   }
}

static char * okCancelStr[] = { "OK", "CANCEL" };

static
int Names (Str)
   char	* Str;
{
   int		x, y, pixel, button_widths, str_width, graph_width, w, h;
   int		str_start, button_start, graph_start, index;
   int		dsp_w, dsp_h;
   XEvent	input, ev;
   int		i, len, looping = TRUE, grabbed = FALSE;
   char		buf[80];
   XKeyEvent	* key_ev;
   XButtonEvent	* button_ev;
   KeySym	key_sym;
   XComposeStatus	c_stat;
   XSetWindowAttributes	win_attrs;
   int		win_x, win_y, win_w, win_h, win_d, win_brdr_w;
   Window	root_win;

   dsp_w = DisplayWidth (mainDisplay, mainScreen);
   dsp_h = DisplayHeight (mainDisplay, mainScreen);

   button_widths = ButtonWidth("OK", 8) + ButtonWidth("CANCEL", 8) +
         defaultFontWidth;
   str_width = defaultFontWidth * strlen (Str);
   graph_width = nameDspWinW + scrollBarW + 2 * brdrW;

   if (str_width > graph_width)
   {
      w = str_width + 4 * defaultFontWidth;
      str_start = 2 * defaultFontWidth;
      graph_start = (w - graph_width) / 2;
   }
   else
   {
      w = graph_width + 4 * defaultFontWidth;
      str_start = (w - str_width) / 2;
      graph_start = 2 * defaultFontWidth;
   }
   button_start = (w - button_widths) / 2;
   h = (6 + ITEM_DSPED) * ROW_HEIGHT;

   win_x = (w > dsp_w) ? 0 : (dsp_w - w)/2;
   win_y = (h > dsp_h) ? 0 : (dsp_h - h)/3;

   if ((nameBaseWin = XCreateSimpleWindow (mainDisplay, rootWindow,
         win_x, win_y, w, h, brdrW, myWhitePixel, myBlackPixel)) == 0)
   { printf ("Could not create desired popup window!\n"); exit (-1); }

   XDefineCursor (mainDisplay, nameBaseWin, defaultCursor);

   if ((nameDspWin = XCreateSimpleWindow (mainDisplay, nameBaseWin, graph_start,
         3*ROW_HEIGHT, nameDspW, nameDspH, brdrW, myWhitePixel,
         myBlackPixel)) == 0)
   { printf ("Could not create desired popup window!\n"); exit (-1); }

   if ((nameScrollWin = XCreateSimpleWindow (mainDisplay, nameBaseWin,
         graph_start+nameDspWinW, 3*ROW_HEIGHT, scrollBarW, nameDspH,
         brdrW, myWhitePixel, myBlackPixel)) == 0)
   { printf ("Could not create desired popup scroll window!\n"); exit (-1); }

   win_attrs.save_under = True;
   XChangeWindowAttributes (mainDisplay, nameBaseWin, CWSaveUnder, &win_attrs);

   XSetTransientForHint (mainDisplay, nameBaseWin, mainWindow);

   XMapWindow (mainDisplay, nameBaseWin);
   XSelectInput (mainDisplay, nameBaseWin,
         KeyPressMask | ButtonPressMask | ExposureMask);

   XMapWindow (mainDisplay, nameDspWin);
   XSelectInput (mainDisplay, nameDspWin,
         KeyPressMask | ButtonPressMask | ExposureMask);

   XMapWindow (mainDisplay, nameScrollWin);
   XSelectInput (mainDisplay, nameScrollWin,
         KeyPressMask | ButtonPressMask | ExposureMask);

   while (looping)
   {
      XNextEvent (mainDisplay, &input);
      if (input.type == Expose)
      {
         if (input.xany.window == nameBaseWin)
            RedrawNameBaseWindow (Str, str_start, button_start);
         else if (input.xany.window == nameScrollWin)
            RedrawNameScrollWin ();
         else if (input.xany.window == nameDspWin)
            RedrawDspWindow ();

         if (!grabbed)
         {
            XGrabKeyboard (mainDisplay, nameBaseWin, False,
                  GrabModeAsync, GrabModeAsync, CurrentTime);
            grabbed = TRUE;
         }

         continue;
      }
      else if (input.type == KeyPress)
      {
         key_ev = &(input.xkey);
         XLookupString (key_ev, buf, 80, &key_sym, &c_stat);

         if (buf[0] == '\r' || buf[0] == '\n')
         {
            looping = FALSE;
            index = nameMarked;
         }
         else if (buf[0] == '\033')
         {
            looping = FALSE;
            index = INVALID;
         }
         else if (key_sym>='\040' && key_sym<='\177')
         {
            if (buf[0] == '$')
               i = (nameEntries == 0) ? 0 : nameEntries-1;
            else
               for (i = 0; i < nameEntries; i++)
                  if (*nameDspPtr[i] == buf[0]) break;
            if (i < nameEntries)
            {
               if (i < nameFirst)
                  nameFirst = i;
               else if (i >= nameFirst+ITEM_DSPED)
               {
                  if (i < nameEntries-ITEM_DSPED)
                     nameFirst = i;
                  else
                     nameFirst = nameEntries-ITEM_DSPED;
               }
               nameMarked = i;
               RedrawNameScrollWin ();
               RedrawDspWindow ();
            }
         }
      }
      else if (input.type == ButtonPress)
      {
         button_ev = &(input.xbutton);
         if (button_ev->window == nameBaseWin)
         {
            if (PointInBBox (button_ev->x, button_ev->y, buttonBBox[0]))
            {
               looping = FALSE;
               index = nameMarked;
            }
            else if (PointInBBox (button_ev->x, button_ev->y, buttonBBox[1]))
            {
               looping = FALSE;
               index = INVALID;
            }
         }
         else if (button_ev->window == nameScrollWin)
            NameScrollHandler (button_ev);
         else if (button_ev->window == nameDspWin)
            NameDspHandler (button_ev);
      }
   }
   if (grabbed) XUngrabKeyboard (mainDisplay, CurrentTime);

   if (index == INVALID)
      DisplayButton (nameBaseWin, okCancelStr[1], 8, &buttonBBox[1],
            BUTTON_INVERT);
   else
      DisplayButton (nameBaseWin, okCancelStr[0], 8, &buttonBBox[0],
            BUTTON_INVERT);

   XDestroyWindow (mainDisplay, nameBaseWin);
   XSync (mainDisplay, FALSE);
   while (XCheckMaskEvent (mainDisplay, ExposureMask, &input)) ;
   return (index);
}

static
int	LargerStr (S1, S2)
   register char	* S1, * S2;
   /* returns TRUE if S1 > S2 */
{
   while (*S1 == *S2 && *S1 != '\0' && *S2 != '\0') { S1++; S2++; }

   return (*S1 > *S2);
}

static
DspList * DirListing (Path, Extension, Entries)
   char	* Path, * Extension;
   int	* Entries;
{
   DspList		* dsp_ptr = NULL, * head_ptr, * tail_ptr, * p, * p1;
   DIR			* dirp;
   struct direct	* d;
   int			len, ext_len, count = 0;
   char			ext[80];

   if (*Path == '\0')
   {
      if ((dirp = opendir (".")) == NULL) return (NULL);
   }
   else
   {
      if ((dirp = opendir (Path)) == NULL) return (NULL);
   }

   strcpy (ext, ".");
   strcat (ext, Extension);
   ext_len = strlen (ext);
   head_ptr = tail_ptr = NULL;

   while ((d = readdir (dirp)) != NULL)
   {
      len = strlen (d->d_name);
      if (len > ext_len && (strcmp (ext, &d->d_name[len-ext_len]) == 0))
      {
         d->d_name[len-ext_len] = '\0';
         dsp_ptr = (DspList *) calloc (1, sizeof(DspList));
         strcpy (dsp_ptr->itemstr, d->d_name);
         if (head_ptr == NULL)
            head_ptr = tail_ptr = dsp_ptr;
         else
         {
            p1 = NULL;
            for (p = head_ptr; p != NULL; p = p->next)
               if (LargerStr (d->d_name, p->itemstr))
                  p1 = p;
               else
                  break;

            dsp_ptr->next = p;
            if (p == NULL)
            {  /* dsp_ptr has the largest element */
               tail_ptr->next = dsp_ptr;
               tail_ptr = dsp_ptr;
            }
            else if (p1 == NULL)
               head_ptr = dsp_ptr;
            else
               p1->next = dsp_ptr;
         }
         count++;
      }
   }

   closedir (dirp);
   *Entries = count;
   return (head_ptr);
}

int SelectFileName (Extension, Str, SelStr)
   char	* Extension, * Str, * SelStr;
{
   register int	i, index;
   char		s[80], full_name[MAXPATHLENGTH];
   DspList	* dsp_ptr;

   sprintf (s, "Generating list of file names, please wait ...");
   Msg (s);
   strcpy (full_name, curDomainName);

   if ((dsp_ptr = DirListing (curDomainName, Extension, &nameEntries)) == NULL)
   {
      strcpy (s, "No appropriate files to select from.");
      Msg (s);
      *SelStr = '\0';
      return (INVALID);
   }

   nameDspPtr = MakeNameDspItemArray (nameEntries, dsp_ptr);
   nameFirst = 0;
   nameMarked = 0;

   if ((index = Names (Str)) == INVALID)
      *SelStr = '\0';
   else
      strcpy (SelStr, nameDspPtr[index]);

   for ( ; dsp_ptr != NULL; dsp_ptr = dsp_ptr->next) cfree (dsp_ptr);
   cfree (nameDspPtr[0]);
   cfree (nameDspPtr);

   s[0] = '\0';
   Msg (s);

   return (index);
}

static
DspList * DomainListing (Path, Entries)
   char	* Path;
   int	* Entries;
{
   DspList		* dsp_ptr = NULL, * head_ptr, * tail_ptr, * p, * p1;
   DIR			* dirp, * dp;
   struct direct	* d;
   struct stat		stat_buf;
   int			count = 0;
   char			path[255], s[255];

   if (*Path == '\0')
      strcpy (path, ".");
   else
      strcpy (path, Path);

   if ((dirp = opendir (path)) == NULL) return (NULL);

   strcat (path, "/");
   head_ptr = tail_ptr = NULL;

   while ((d = readdir (dirp)) != NULL)
   {
      strcpy (s, path);
      strcat (s, d->d_name);

      stat (s, &stat_buf);
      if (stat_buf.st_mode & S_IFDIR)
      {
         dsp_ptr = (DspList *) calloc (1, sizeof(DspList));
         strcpy (dsp_ptr->itemstr, d->d_name);
         if (head_ptr == NULL)
            head_ptr = tail_ptr = dsp_ptr;
         else
         {
            p1 = NULL;
            for (p = head_ptr; p != NULL; p = p->next)
               if (LargerStr (d->d_name, p->itemstr))
                  p1 = p;
               else
                  break;

            dsp_ptr->next = p;
            if (p == NULL)
            {  /* dsp_ptr has the largest element */
               tail_ptr->next = dsp_ptr;
               tail_ptr = dsp_ptr;
            }
            else if (p1 == NULL)
               head_ptr = dsp_ptr;
            else
               p1->next = dsp_ptr;
         }
         count++;
      }
   }

   closedir (dirp);
   *Entries = count;
   return (head_ptr);
}

int SelectDomain (SelStr)
   char	* SelStr;
{
   register int		i, index;
   char			s[80];
   DspList		* dsp_ptr;

   Msg ("Generating list of domain names, please wait ...");

   if ((dsp_ptr = DomainListing (curDomainName, &nameEntries)) == NULL)
   {
      *SelStr = '\0';
      return (INVALID);
   }

   nameDspPtr = MakeNameDspItemArray (nameEntries, dsp_ptr);
   nameFirst = 0;
   nameMarked = 0;

   if ((index = Names ("PLEASE SELECT NEW DOMAIN ...")) == INVALID)
      *SelStr = '\0';
   else
      strcpy (SelStr, nameDspPtr[index]);

   for ( ; dsp_ptr != NULL; dsp_ptr = dsp_ptr->next) cfree (dsp_ptr);
   cfree (nameDspPtr[0]);
   cfree (nameDspPtr);

   s[0] = '\0';
   Msg (s);

   return (index);
}
