/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/oval.c,v 1.2 90/05/16 08:30:59 william Exp $";
#endif

#include <math.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "color.e"
#include "cursor.e"
#include "file.e"
#include "grid.e"
#include "menu.e"
#include "obj.e"
#include "pattern.e"
#include "poly.e"
#include "raster.e"
#include "ruler.e"
#include "select.e"
#include "setup.e"

int	ovalDrawn = FALSE;

static XPoint ov[13];

XPoint * MakeOvalVertex (BBox)
   struct BBRec	BBox;
{
   register XPoint	* v;
   register int		i, xc, yc;
   int			a, b;

   a = (BBox.rbx - BBox.ltx) / 2; xc = BBox.ltx + a;
   b = (BBox.rby - BBox.lty) / 2; yc = BBox.lty + b;

   v = (XPoint *) calloc (13, sizeof(XPoint));

   v[0].x = (short)xc;			v[0].y = (short)(yc-b);
   v[1].x = (short)(xc-a/2);		v[1].y = (short)(yc-b*0.866);
   v[2].x = (short)(xc-a*0.866);	v[2].y = (short)(yc-b/2);
   v[3].x = (short)(xc-a);		v[3].y = (short)yc;
   v[4].x = (short)(xc-a*0.866);	v[4].y = (short)(yc+b/2);
   v[5].x = (short)(xc-a/2);		v[5].y = (short)(yc+b*0.866);
   v[6].x = (short)xc;			v[6].y = (short)(yc+b);
   v[7].x = (short)(xc+a/2);		v[7].y = (short)(yc+b*0.866);
   v[8].x = (short)(xc+a*0.866);	v[8].y = (short)(yc+b/2);
   v[9].x = (short)(xc+a);		v[9].y = (short)yc;
   v[10].x = (short)(xc+a*0.866);	v[10].y = (short)(yc-b/2);
   v[11].x = (short)(xc+a/2);		v[11].y = (short)(yc-b*0.866);
   v[12].x = (short)xc;			v[12].y = (short)(yc-b);

   return (v);
}
 
void MyOval (window, gc, bbox)
   Window	window;
   GC		gc;
   struct BBRec	bbox;
{
   register int		ltx, lty, w, h;

   if (bbox.ltx > bbox.rbx)
   {
      ltx = bbox.rbx; w = bbox.ltx - ltx;
   }
   else
   {
      ltx = bbox.ltx; w = bbox.rbx - ltx;
   }

   if (bbox.lty > bbox.rby)
   {
      lty = bbox.rby; h = bbox.lty - lty;
   }
   else
   {
      lty = bbox.lty; h = bbox.rby - lty;
   }

   XDrawArc (mainDisplay, window, gc, ltx, lty, w, h, 0, 360*64);
}
 
void DumpOvalObj (FP, ObjPtr)
   FILE			* FP;
   struct ObjRec	* ObjPtr;
{
   int			ltx, lty, rbx, rby, xc, yc, a, b, fill, width, pen;

   ltx = ObjPtr->obbox.ltx; lty = ObjPtr->obbox.lty;
   rbx = ObjPtr->obbox.rbx; rby = ObjPtr->obbox.rby;
   a = (rbx - ltx) / 2; xc = ltx + a;
   b = (rby - lty) / 2; yc = lty + b;

   fill = ObjPtr->detail.o->fill;
   width = ObjPtr->detail.o->width;
   pen = ObjPtr->detail.o->pen;

   if (fill == NONEPAT && pen == NONEPAT) return;

   switch (fill)
   {
      case NONEPAT: break;
      case SOLIDPAT:
         /* solid black oval */
         fprintf (FP, "newpath %1d %1d %1d %1d ellipse fill\n", xc, yc, a, b);
         break;
      case BACKPAT:
         /* solid white oval */
         fprintf (FP, "newpath %1d %1d %1d %1d ellipse\n", xc, yc, a, b);
         fprintf (FP, "closepath 1 setgray fill 0 setgray\n");
         break;
      default:
         /* patterned */
         fprintf (FP, "gsave\n");
         fprintf (FP, "   pat%1d 8 1 0 72 300 32 div div setpattern\n", fill);
         fprintf (FP, "   newpath %1d %1d %1d %1d ellipse fill\n", xc, yc, a, b);
         fprintf (FP, "grestore\n");
         break;
   }

   fprintf (FP, "%1d setlinewidth\n", widthOfLine[width]);

   switch (pen)
   {
      case NONEPAT: break;
      case SOLIDPAT:
         fprintf (FP, "newpath %1d %1d %1d %1d ellipse stroke\n", xc, yc, a, b);
         break;
      case BACKPAT:
         fprintf (FP, "newpath %1d %1d %1d %1d ellipse\n", xc, yc, a, b);
         fprintf (FP, "closepath 1 setgray stroke 0 setgray\n");
         break;
      default:
         fprintf (FP, "gsave\n");
         fprintf (FP, "   pat%1d 8 1 0 72 300 32 div div setpattern\n", pen);
         fprintf (FP, "   newpath %1d %1d %1d %1d ellipse stroke\n",
               xc, yc, a, b);
         fprintf (FP, "grestore\n");
         break;
   }
   fprintf (FP, "1 setlinewidth\n\n");
}
 
void DrawOvalObj (window, XOff, YOff, ObjPtr)
   Window		window;
   int			XOff, YOff;
   struct ObjRec	* ObjPtr;
{
   struct OvalRec	* oval_ptr = ObjPtr->detail.o;
   int			fill, width, pen, pixel, real_x_off, real_y_off;
   char			s[80];
   struct BBRec		bbox;
   XGCValues		values;

   real_x_off = (XOff >> zoomScale) << zoomScale;
   real_y_off = (YOff >> zoomScale) << zoomScale;
   bbox.ltx = (ObjPtr->obbox.ltx - real_x_off) >> zoomScale;
   bbox.lty = (ObjPtr->obbox.lty - real_y_off) >> zoomScale;
   bbox.rbx = (ObjPtr->obbox.rbx - real_x_off) >> zoomScale;
   bbox.rby = (ObjPtr->obbox.rby - real_y_off) >> zoomScale;

   fill = ObjPtr->detail.o->fill;
   width = ObjPtr->detail.o->width;
   pen = ObjPtr->detail.o->pen;
   pixel = colorPixels[ObjPtr->color];

   if (fill != 0)
   {
      values.foreground = (fill == 2) ? myBlackPixel : pixel;
      values.function = GXcopy;
      values.fill_style = FillOpaqueStippled;
      values.stipple = patPixmap[fill];
      XChangeGC (mainDisplay, drawGC,
            GCForeground | GCFunction | GCFillStyle | GCStipple, &values);
      XFillArc (mainDisplay, window, drawGC, bbox.ltx, bbox.lty,
            bbox.rbx-bbox.ltx, bbox.rby-bbox.lty, 0, 360*64);
   }
   if (pen != 0)
   {
      values.foreground = (pen == 2) ? myBlackPixel : pixel;
      values.function = GXcopy;
      values.fill_style = FillOpaqueStippled;
      values.stipple = patPixmap[pen];
      values.line_width = widthOfLine[width] >> zoomScale;
      XChangeGC (mainDisplay, drawGC,
            GCForeground | GCFunction | GCFillStyle | GCStipple | GCLineWidth,
            &values);
      XDrawArc (mainDisplay, window, drawGC, bbox.ltx, bbox.lty,
            bbox.rbx-bbox.ltx, bbox.rby-bbox.lty, 0, 360*64);
   }
}

static
void CreateOvalObj (BBox)
   struct BBRec	BBox;
{
   struct OvalRec	* oval_ptr;
   struct ObjRec	* obj_ptr;
   int			w;

   oval_ptr = (struct OvalRec *) calloc (1, sizeof(struct OvalRec));
   oval_ptr->fill = objFill;
   oval_ptr->width = lineWidth;
   oval_ptr->pen = penPat;

   obj_ptr = (struct ObjRec *) calloc (1, sizeof(struct ObjRec));

   obj_ptr->bbox.ltx = obj_ptr->obbox.ltx = obj_ptr->x =
         (BBox.ltx << zoomScale) + drawOrigX;
   obj_ptr->bbox.lty = obj_ptr->obbox.lty = obj_ptr->y =
         (BBox.lty << zoomScale) + drawOrigY;
   obj_ptr->bbox.rbx = obj_ptr->obbox.rbx = (BBox.rbx << zoomScale) + drawOrigX;
   obj_ptr->bbox.rby = obj_ptr->obbox.rby = (BBox.rby << zoomScale) + drawOrigY;
   w = widthOfLine[lineWidth];
   obj_ptr->bbox.ltx -= w;
   obj_ptr->bbox.lty -= w;
   obj_ptr->bbox.rbx += w;
   obj_ptr->bbox.rby += w;
   obj_ptr->type = OBJ_OVAL;
   obj_ptr->color = colorIndex;
   obj_ptr->detail.o = oval_ptr;
   AddObj (NULL, topObj, obj_ptr);
}
 
static
void ContinueOval (OrigX, OrigY)
   int	OrigX, OrigY;
{
   int 		end_x, end_y, grid_x, grid_y;
   int 		done = FALSE;
   int		pixel, xor_pixel;
   struct BBRec	bbox;
   XGCValues	values;
   XEvent	input;
   XButtonEvent	* button_ev;
   XMotionEvent	* motion_ev;

   bbox.ltx = bbox.rbx = OrigX;
   bbox.lty = bbox.rby = OrigY;

   pixel = colorPixels[colorIndex];
   xor_pixel = xorColorPixels[colorIndex];

   values.foreground = xor_pixel;
   values.function = GXxor;
   values.fill_style = FillSolid;
   values.line_width = 0;
   XChangeGC (mainDisplay, drawGC,
      GCForeground | GCFunction | GCFillStyle | GCLineWidth, &values);

   grid_x = end_x = OrigX;
   grid_y = end_y = OrigY; 
   XGrabPointer (mainDisplay, drawWindow, FALSE,
         PointerMotionMask | ButtonReleaseMask,
         GrabModeAsync, GrabModeAsync, None, handCursor, CurrentTime);
   
   while (!done)
   {
      XNextEvent (mainDisplay, &input);
      if (input.type == ButtonRelease)
      {
         XUngrabPointer (mainDisplay, CurrentTime);
         MyOval (drawWindow, drawGC, bbox);
         done = TRUE;
      }
      else if (input.type == MotionNotify)
      {
         motion_ev = &(input.xmotion);
         end_x = motion_ev->x;
         end_y = motion_ev->y;
         GridXY (end_x, end_y, &grid_x, &grid_y);
         if (grid_x != bbox.rbx || grid_y != bbox.rby)
         {
            MyOval (drawWindow, drawGC, bbox);
            bbox.rbx = grid_x;
            bbox.rby = grid_y;
            MyOval (drawWindow, drawGC, bbox);
         }
         MarkRulers (grid_x, grid_y);
      }
   }
   if (OrigX != grid_x && OrigY != grid_y)
   {
      if (bbox.ltx > bbox.rbx)
      {
         end_x = bbox.ltx; bbox.ltx = bbox.rbx; bbox.rbx = end_x;
      }
      if (bbox.lty > bbox.rby)
      {
         end_y = bbox.lty; bbox.lty = bbox.rby; bbox.rby = end_y;
      }
      CreateOvalObj (bbox);
      DrawOvalObj (drawWindow, drawOrigX, drawOrigY, topObj);
      ovalDrawn = TRUE;
      fileModified = TRUE;
   }
}

void DrawOval (input)
   XEvent	* input;
{
   XButtonEvent	* button_ev;
   Window	root_win, child_win;
   int		root_x, root_y;
   unsigned int	status;
   int		mouse_x, mouse_y, grid_x, grid_y;

   if (input->type != ButtonPress) return;

   button_ev = &(input->xbutton);
   if (button_ev->button == Button1)
   {
      XQueryPointer (mainDisplay, drawWindow, &root_win, &child_win,
            &root_x, &root_y, &mouse_x, &mouse_y, &status);
      GridXY (mouse_x, mouse_y, &grid_x, &grid_y);
      ContinueOval (grid_x, grid_y);
   }
}

void SaveOvalObj (FP, ObjPtr)
   FILE			* FP;
   struct ObjRec	* ObjPtr;
{
   fprintf (FP, "oval(%s,", colorMenuItems[ObjPtr->color]);
   fprintf (FP, "%1d,%1d,%1d,%1d,%1d,%1d,%1d)", ObjPtr->obbox.ltx,
         ObjPtr->obbox.lty, ObjPtr->obbox.rbx, ObjPtr->obbox.rby,
         ObjPtr->detail.o->fill, ObjPtr->detail.o->width, ObjPtr->detail.o->pen);
}

void ReadOvalObj (Inbuf, ObjPtr)
   char			* Inbuf;
   struct ObjRec	* * ObjPtr;
{
   struct OvalRec	* oval_ptr;
   char			color_str[20], * s;
   int			ltx, lty, rbx, rby, fill, width, pen, w;

   * ObjPtr = (struct ObjRec *) calloc (1, sizeof(struct ObjRec));
   s = FindChar ('(', Inbuf);
   s = ParseStr (s, ',', color_str);
   oval_ptr = (struct OvalRec *) calloc (1, sizeof(struct OvalRec));

   if (fileVersion <= 5)
   {
      sscanf (s, "%d , %d, %d , %d , %d , %d , %d",
            &ltx, &lty, &rbx, &rby, &fill, &width, &pen);
      switch (width)
      {
         case 1: width = 3; break;
         case 2: width = 6; break;
      }
   }
   else
      sscanf (s, "%d , %d, %d , %d , %d , %d , %d",
            &ltx, &lty, &rbx, &rby, &fill, &width, &pen);

   oval_ptr->fill = fill;
   oval_ptr->width = width;
   oval_ptr->pen = pen;
   (*ObjPtr)->x = ltx;
   (*ObjPtr)->y = lty;
   (*ObjPtr)->color = FindColorIndex (color_str);
   (*ObjPtr)->type = OBJ_OVAL;
   (*ObjPtr)->obbox.ltx = ltx;
   (*ObjPtr)->obbox.lty = lty;
   (*ObjPtr)->obbox.rbx = rbx;
   (*ObjPtr)->obbox.rby = rby;
   if (width == 0)
   {
      (*ObjPtr)->bbox.ltx = ltx;
      (*ObjPtr)->bbox.lty = lty;
      (*ObjPtr)->bbox.rbx = rbx;
      (*ObjPtr)->bbox.rby = rby;
   }
   else
   {
      w = widthOfLine[width];
      (*ObjPtr)->bbox.ltx = ltx - w;
      (*ObjPtr)->bbox.lty = lty - w;
      (*ObjPtr)->bbox.rbx = rbx + w;
      (*ObjPtr)->bbox.rby = rby + w;
   }
   (*ObjPtr)->detail.o = oval_ptr;
}

void FreeOvalObj (ObjPtr)
   struct ObjRec	* ObjPtr;
{
   cfree (ObjPtr->detail.o);
   cfree (ObjPtr);
}
