/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/rect.c,v 1.1 90/04/01 22:16:37 william Exp $";
#endif

#include <X11/Xlib.h>
#include <math.h>
#include "const.h"
#include "types.h"

#include "oval.e"
#include "poly.e"
#include "setup.e"
#include "spline.e"

int BBoxIntersect (Rect1, Rect2)
   struct BBRec	Rect1, Rect2;
{
   if (Rect1.ltx < Rect2.ltx)
      if (Rect1.lty < Rect2.lty)
         return (Rect1.rbx >= Rect2.ltx && Rect1.rby >= Rect2.lty);
      else
         return (Rect1.rbx >= Rect2.ltx && Rect1.lty <= Rect2.rby);
   else
      if (Rect1.lty < Rect2.lty)
         return (Rect1.ltx <= Rect2.rbx && Rect1.rby >= Rect2.lty);
      else
         return (Rect1.ltx <= Rect2.rbx && Rect1.lty <= Rect2.rby);
}

int PointInBBox (X, Y, Rect)
   int	X, Y;
   struct BBRec	Rect;
{
   return (X >= Rect.ltx && Y >= Rect.lty && X <= Rect.rbx && Y <= Rect.rby);
}

int PointInPolygon (X, Y, NumPts, V)
   int		X, Y, NumPts;
   XPoint	* V;
{
   register double	x1, x2, y1, y2;
   double		m, y_int;
   int			n, count = 0;

   x2 = (double)V[0].x;
   y2 = (double)V[0].y;
   for (n = 0; n < NumPts-1; n++)
   {
      x1 = x2;
      y1 = y2;
      x2 = (double)V[n+1].x;
      y2 = (double)V[n+1].y;
      if (x2 == x1 && X == x1)
      {
         if (Y <= min(y1,y2)) count++;
         continue;
      }
      if (x2 > x1)
      {
         if (X >= x2 || X < x1) continue;
      }
      else
      {
         if (X > x1 || X <= x2) continue;
      }

      m = (y1 - y2) / (x1 - x2);
      y_int = m * X + (y1 - m * x1);
      if (Y <= y_int) count++;
   }
   return (count & 0x1);
}

int PointInPoly (X, Y, NumPts, V, W)
   int		X, Y, NumPts, W;
   XPoint	* V;
{
   register double	x1, x2, y1, y2;
   double		m, m1, b, b1, x_int, y_int;
   int			n;

   x2 = (double)V[0].x;
   y2 = (double)V[0].y;
   for (n = 0; n < NumPts-1; n++)
   {
      x1 = x2;
      y1 = y2;
      x2 = (double)V[n+1].x;
      y2 = (double)V[n+1].y;
      if (x2 >= x1)
      {
         if (X < x1-W || X > x2+W) continue;
      }
      else
      {
         if (X < x2-W || X > x1+W) continue;
      }
      if (y2 >= y1)
      {
         if (Y < y1-W || Y > y2+W) continue;
      }
      else
      {
         if (Y < y2-W || Y > y1+W) continue;
      }

      if (x1 == x2 || y1 == y2)
         return (TRUE);
      else
      {
         m = (y1 - y2) / (x1 - x2);
         b = y1 - m * x1;
         m1 = (x2 - x1) / (y1 - y2);
         b1 = Y - m1 * X;
         x_int = (b1 - b) / (m - m1);
         y_int = m * x_int + b;
         if (sqrt ((double)(x_int-X)*(x_int-X)+(y_int-Y)*(y_int-Y)) <= (double)W)
            return (TRUE);
      }
   }
   return (FALSE);
}

int FindGoodBox (X, Y, BoxObj)
   int			X, Y;
   struct ObjRec	* BoxObj;
   /* X and Y are absolute coordinates */
{
   struct BBRec	bbox;
   int		w;

   if (BoxObj->detail.b->fill != 0) return (TRUE);

   w = (widthOfLine[BoxObj->detail.b->width]+3) << zoomScale;
   bbox.ltx = BoxObj->obbox.ltx + w;
   bbox.lty = BoxObj->obbox.lty + w;
   bbox.rbx = BoxObj->obbox.rbx - w;
   bbox.rby = BoxObj->obbox.rby - w;
   return (!PointInBBox (X, Y, bbox));
}

int FindGoodOval (X, Y, OvalObj)
   int			X, Y;
   struct ObjRec	* OvalObj;
   /* X and Y are absolute coordinates */
{
   struct OvalRec	* oval_ptr = OvalObj->detail.o;
   int			w, ltx, lty, rbx, rby, delta;
   double		cx, cy, rx, ry, x1, x2, y1, y2, tmp_x, tmp_y;
   int			fill = oval_ptr->fill, pen = oval_ptr->pen;

   if (fill == NONEPAT && pen == NONEPAT) return (FALSE);

   ltx = OvalObj->obbox.ltx; lty = OvalObj->obbox.lty;
   rbx = OvalObj->obbox.rbx; rby = OvalObj->obbox.rby;

   cx = (ltx + rbx) / 2; cy = (lty + rby) / 2;
   rx = (rbx - ltx) / 2; ry = (rby - lty) / 2;

   if (rx >= ry)
   {  /* flat oval */
      tmp_y = sqrt ((double)(ry*ry*(1-(X-cx)*(X-cx)/rx/rx)));
      y1 = cy - tmp_y;
      y2 = cy + tmp_y;
      delta = (int)(3.0*(ry/rx));
   }
   else
   {  /* tall oval */
      tmp_x = sqrt ((double)(rx*rx*(1-(Y-cy)*(Y-cy)/ry/ry)));
      x1 = cx - tmp_x;
      x2 = cx + tmp_x;
      delta = (int)(3.0*(rx/ry));
   }

   if (fill != NONEPAT)
   {
      if (rx >= ry)
      {  /* flat oval */
         if (Y >= y1 && y2 >= Y) return (TRUE);
      }
      else
      {  /* tall oval */
         if (X >= x1 && x2 >= X) return (TRUE);
      }
   }
   w = (widthOfLine[oval_ptr->width]+3+delta) << zoomScale;
   if (pen != NONEPAT)
   {
      if (rx >= ry)
         return ((fabs ((double)(Y-y1)) <= w) || (fabs ((double)(Y-y2)) <= w));
      else
         return ((fabs ((double)(X-x1)) <= w) || (fabs ((double)(X-x2)) <= w));
   }
   else
      return (FALSE);
}

int FindGoodPoly (X, Y, PolyObj)
   int			X, Y;
   struct ObjRec	* PolyObj;
   /* X and Y are absolute coordinates */
{
   int			w, n;
   XPoint		* v;
   struct PolyRec	* poly_ptr;

   poly_ptr = PolyObj->detail.p;
   w = (arrowHeadH[poly_ptr->width]+3) << zoomScale;
   if (poly_ptr->curved == LT_SPLINE)
      return (PointInSplinePoly (X, Y, poly_ptr, w));
   else
   {
      if (poly_ptr->fill != NONEPAT)
      {
         n = poly_ptr->n;
         v = poly_ptr->vlist;
         v[n].x = v[0].x; v[n].y = v[0].y;
         if (PointInPolygon (X, Y, n+1, v))
            return (TRUE);
      }
      return (PointInPoly (X, Y, poly_ptr->n, poly_ptr->vlist, w));
   }
}

int FindGoodPolygon (X, Y, PolygonObj)
   int			X, Y;
   struct ObjRec	* PolygonObj;
   /* X and Y are absolute coordinates */
{
   int			w;
   struct PolygonRec	* polygon_ptr;

   polygon_ptr = PolygonObj->detail.g;
   w = (widthOfLine[polygon_ptr->width]+3) << zoomScale;
   if (polygon_ptr->curved == LT_SPLINE)
      return (PointInSplinePolygon (X, Y, polygon_ptr, w));
   else
   {
      if (polygon_ptr->fill != NONEPAT)
         if (PointInPolygon (X, Y, polygon_ptr->n, polygon_ptr->vlist))
            return (TRUE);
      return (PointInPoly (X, Y, polygon_ptr->n, polygon_ptr->vlist, w));
   }
}

int FindGoodObj (X, Y, FirstObj)
   int			X, Y;
   struct ObjRec	* FirstObj;
   /* X and Y are absolute coordinates */
{
   register struct ObjRec	* obj_ptr;

   for (obj_ptr = FirstObj; obj_ptr != NULL; obj_ptr = obj_ptr->next)
      if (X >= obj_ptr->bbox.ltx && X <= obj_ptr->bbox.rbx &&
            Y >= obj_ptr->bbox.lty && Y <= obj_ptr->bbox.rby)
      {
         switch (obj_ptr->type)
         {
            case OBJ_TEXT: return (TRUE);
            case OBJ_BOX:
               if (FindGoodBox (X, Y, obj_ptr)) return (TRUE);
               break;
            case OBJ_OVAL:
               if (FindGoodOval (X, Y, obj_ptr)) return (TRUE);
               break;
            case OBJ_POLY:
               if (FindGoodPoly (X, Y, obj_ptr)) return (TRUE);
               break;
            case OBJ_POLYGON:
               if (FindGoodPolygon (X, Y, obj_ptr)) return (TRUE);
               break;

            case OBJ_GROUP:
            case OBJ_SYM:
            case OBJ_ICON:
               if (FindGoodObj (X, Y, obj_ptr->detail.r->first)) return (TRUE);
               break;
         }
      }
   return (FALSE);
}
