/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/scroll.c,v 1.1 90/04/01 22:16:39 william Exp $";
#endif

#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "drawing.e"
#include "msg.e"
#include "obj.e"
#include "raster.e"
#include "ruler.e"
#include "setup.e"
#include "text.e"

#include "bitmaps/uparrow.bm"
#include "bitmaps/downarrow.bm"
#include "bitmaps/rightarrow.bm"
#include "bitmaps/leftarrow.bm"

char	* upData, * downData, * rightData, * leftData;

GC	scrollGC;
Pixmap	scrollPixmap;
XImage	* scrollImage;

static int	scrollAreaH, scrollAreaW;

void UpdScrollWinWH ()
{
   scrollAreaH = vSBarH - 2 * scrollBarW;
   scrollAreaW = hSBarW - 2 * scrollBarW;
}

void InitScroll ()
{
   XGCValues	values;

   upData = uparrow_bits;
   downData = downarrow_bits;
   rightData = rightarrow_bits;
   leftData = leftarrow_bits;

   UpdScrollWinWH ();

   scrollPixmap = XCreatePixmap (mainDisplay, mainWindow, scrollBarW,
         scrollBarW, mainDepth);
   values.foreground = myWhitePixel;
   values.background = myBlackPixel;
   values.fill_style = FillTiled;
   values.tile = scrollPixmap;
   scrollGC = XCreateGC (mainDisplay, mainWindow,
         GCForeground | GCBackground | GCFillStyle | GCTile, &values);
   scrollImage = XCreateImage (mainDisplay, mainVisual, 1, XYBitmap,
         0, uparrow_bits, scrollBarW, scrollBarW, 8, 2);
   scrollImage->byte_order = LSBFirst;
   scrollImage->bitmap_bit_order = LSBFirst;
}

static
void RedrawVScrollWindow ()
{
   double	frac, start_frac;
   int		block_h, block_start, pixel, func;
   XGCValues	values;
   XEvent	ev;

   XSync (mainDisplay, FALSE);
   while (XCheckWindowEvent (mainDisplay, vSBarWindow, ExposureMask, &ev)) ;

   start_frac = (double)((double)(drawOrigY)/(double)(paperHeight));
   block_start = (int)(scrollAreaH * start_frac);

   if (paperHeight > drawWinH)
      frac = (double)((double)drawWinH / (double)(paperHeight));
   else
      frac = 1.0;

   block_h = (int)(scrollAreaH * frac);

   values.foreground = myBlackPixel;
   values.background = myWhitePixel;
   values.function = GXcopy;
   values.fill_style = FillSolid;
   XChangeGC (mainDisplay, scrollGC,
         GCForeground | GCBackground | GCFunction | GCFillStyle,
         &values);
   XFillRectangle (mainDisplay, vSBarWindow, scrollGC, 0, scrollBarW,
         scrollBarW, scrollAreaH);

   values.foreground = myWhitePixel;
   values.background = myBlackPixel;
   values.fill_style = FillOpaqueStippled;
   values.stipple = patPixmap[SCROLLPAT];
   XChangeGC (mainDisplay, scrollGC,
         GCForeground | GCBackground | GCFillStyle | GCStipple,
         &values);
   XFillRectangle (mainDisplay, vSBarWindow, scrollGC, 0,
         scrollBarW+block_start, scrollBarW, block_h);

   scrollImage->data = upData;
   XPutImage (mainDisplay, vSBarWindow, scrollGC, scrollImage, 0, 0, 0, 0,
         scrollBarW, scrollBarW);
   scrollImage->data = downData;
   XPutImage (mainDisplay, vSBarWindow, scrollGC, scrollImage, 0, 0,
         0, scrollBarW+scrollAreaH, scrollBarW, scrollBarW);
}

static
void RedrawHScrollWindow ()
{
   double	frac, start_frac;
   int		block_w, block_start, pixel, func;
   XGCValues	values;
   XEvent	ev;

   XSync (mainDisplay, FALSE);
   while (XCheckWindowEvent (mainDisplay, hSBarWindow, ExposureMask, &ev)) ;

   start_frac = (double)((double)(drawOrigX)/(double)(paperWidth));
   block_start = (int)(scrollAreaW * start_frac);

   if (paperWidth > drawWinW)
      frac = (double)((double)drawWinW / (double)(paperWidth));
   else
      frac = 1.0;

   block_w = (int)(scrollAreaW * frac);

   values.foreground = myBlackPixel;
   values.background = myWhitePixel;
   values.function = GXcopy;
   values.fill_style = FillSolid;
   XChangeGC (mainDisplay, scrollGC,
         GCForeground | GCBackground | GCFunction | GCFillStyle,
         &values);
   XFillRectangle (mainDisplay, hSBarWindow, scrollGC, scrollBarW, 0,
         scrollAreaW, scrollBarW);

   values.foreground = myWhitePixel;
   values.background = myBlackPixel;
   values.fill_style = FillOpaqueStippled;
   values.stipple = patPixmap[SCROLLPAT];
   XChangeGC (mainDisplay, scrollGC,
         GCForeground | GCBackground | GCFillStyle | GCStipple,
         &values);
   XFillRectangle (mainDisplay, hSBarWindow, scrollGC, scrollBarW+block_start,
         0, block_w, scrollBarW);

   scrollImage->data = leftData;
   XPutImage (mainDisplay, hSBarWindow, scrollGC, scrollImage, 0, 0, 0, 0,
         scrollBarW, scrollBarW);
   scrollImage->data = rightData;
   XPutImage (mainDisplay, hSBarWindow, scrollGC, scrollImage, 0, 0,
         scrollBarW+scrollAreaW, 0, scrollBarW, scrollBarW);
}

void RedrawScrollBars ()
{
   RedrawVScrollWindow ();
   RedrawHScrollWindow ();
}

static
void VSBarHandler (button_ev)
   XButtonEvent	* button_ev;
{
   double	frac, start_frac;
   int		block_h, block_start, adjustment, new_start;

   if (button_ev->button == Button1)
   {
      if (button_ev->y < scrollBarW)
      {
         /* clicked in the uparrow */
         if (drawOrigY != 0)
         {
            drawOrigY -= HALF_INCH;
            RedrawVScrollWindow ();
            UpdDrawWinBBox ();
            AdjSplineVs ();
            AdjustCurText (0, HALF_INCH);
            ClearAndRedrawDrawWindow ();
            XClearWindow (mainDisplay, vRuleWindow);
            RedrawVRuler ();
            return;
         }
      }
      else if (button_ev->y >= scrollBarW+scrollAreaH)
      {
         /* clicked in the downarrow */
         if (paperHeight <= drawWinH) return;

         if (drawOrigY+drawWinH < paperHeight)
         {
            drawOrigY += HALF_INCH;
            RedrawVScrollWindow ();
            UpdDrawWinBBox ();
            AdjSplineVs ();
            AdjustCurText (0, -HALF_INCH);
            ClearAndRedrawDrawWindow ();
            XClearWindow (mainDisplay, vRuleWindow);
            RedrawVRuler ();
            return;
         }
      }
      else /* clicked in the middle region */
      {
         if (paperHeight <= drawWinH) return;

         frac = (double)((double)drawWinH / (double)(paperHeight));
         block_h = (int)(scrollAreaH * frac);
         block_start = button_ev->y - scrollBarW;
         start_frac = (double)((double)(block_start)/(double)(scrollAreaH));
         if (block_start+block_h >= scrollAreaH)
         {
            adjustment = drawOrigY - (paperHeight - drawWinH);
            if (adjustment != 0)
            {
               drawOrigY = paperHeight - drawWinH;
               RedrawVScrollWindow ();
               UpdDrawWinBBox ();
               AdjSplineVs ();
               AdjustCurText (0, adjustment);
               ClearAndRedrawDrawWindow ();
               XClearWindow (mainDisplay, vRuleWindow);
               RedrawVRuler ();
            }
            return;
         }
         else
         {
            new_start = ((int)(paperHeight*start_frac/HALF_INCH))*HALF_INCH;
            adjustment = drawOrigY - new_start;
            if (adjustment != 0)
            {
               drawOrigY = new_start;
               RedrawVScrollWindow ();
               UpdDrawWinBBox ();
               AdjSplineVs ();
               AdjustCurText (0, adjustment);
               ClearAndRedrawDrawWindow ();
               XClearWindow (mainDisplay, vRuleWindow);
               RedrawVRuler ();
            }
            return;
         }
      }
   }
}

static
void HSBarHandler (button_ev)
   XButtonEvent	* button_ev;
{
   double	frac, start_frac;
   int		block_w, block_start, adjustment, new_start;

   if (button_ev->button == Button1)
   {
      if (button_ev->x < scrollBarW)
      {
         /* clicked in the leftarrow */
         if (drawOrigX != 0)
         {
            drawOrigX -= HALF_INCH;
            RedrawHScrollWindow ();
            UpdDrawWinBBox ();
            AdjSplineVs ();
            AdjustCurText (HALF_INCH, 0);
            ClearAndRedrawDrawWindow ();
            XClearWindow (mainDisplay, hRuleWindow);
            RedrawHRuler ();
            return;
         }
      }
      else if (button_ev->x >= scrollBarW+scrollAreaW)
      {
         /* clicked in the rightarrow */
         if (paperWidth <= drawWinW) return;

         if (drawOrigX+drawWinW < paperWidth)
         {
            drawOrigX += HALF_INCH;
            RedrawHScrollWindow ();
            UpdDrawWinBBox ();
            AdjSplineVs ();
            AdjustCurText (-HALF_INCH, 0);
            ClearAndRedrawDrawWindow ();
            XClearWindow (mainDisplay, hRuleWindow);
            RedrawHRuler ();
            return;
         }
      }
      else /* clicked in the middle region */
      {
         if (paperWidth <= drawWinW) return;

         frac = (double)((double)drawWinW / (double)(paperWidth));
         block_w = (int)(scrollAreaW * frac);
         block_start = button_ev->x - scrollBarW;
         start_frac = (double)((double)(block_start)/(double)(scrollAreaW));
         if (block_start+block_w >= scrollAreaW)
         {
            adjustment = drawOrigX - (paperWidth - drawWinW);
            if (adjustment != 0)
            {
               drawOrigX = paperWidth - drawWinW;
               RedrawHScrollWindow ();
               UpdDrawWinBBox ();
               AdjSplineVs ();
               AdjustCurText (adjustment, 0);
               ClearAndRedrawDrawWindow ();
               XClearWindow (mainDisplay, hRuleWindow);
               RedrawHRuler ();
            }
            return;
         }
         else
         {
            new_start = ((int)(paperWidth*start_frac/HALF_INCH))*HALF_INCH;
            adjustment = drawOrigX - new_start;
            if (adjustment != 0)
            {
               drawOrigX = new_start;
               RedrawHScrollWindow ();
               UpdDrawWinBBox ();
               AdjSplineVs ();
               AdjustCurText (adjustment, 0);
               ClearAndRedrawDrawWindow ();
               XClearWindow (mainDisplay, hRuleWindow);
               RedrawHRuler ();
            }
            return;
         }
      }
   }
}

void ScrollEventHandler (input)
   XEvent	* input;
{
   int		row_offset, len, index, x, y, func, level;
   char		new_file_name[80];
   XButtonEvent	* button_ev;

   if (input->xany.window == vSBarWindow)
   {
      if (input->type == Expose)
      {
         XSync (mainDisplay, FALSE);
         RedrawVScrollWindow ();
         return;
      }

      Msg (""); 
      VSBarHandler (&(input->xbutton));
      return;
   }
   else if (input->xany.window == hSBarWindow)
   {
      if (input->type == Expose)
      {
         XSync (mainDisplay, FALSE);
         RedrawHScrollWindow ();
         return;
      }

      Msg (""); 
      HSBarHandler (&(input->xbutton));
      return;
   }
}

void CleanUpScrolls ()
{
   XFreePixmap (mainDisplay, scrollPixmap);
   XFreeGC (mainDisplay, scrollGC);
/* XDestroyImage (scrollImage); */
}
