/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/select.c,v 1.1 90/04/01 22:16:40 william Exp $";
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "choice.e"
#include "cursor.e"
#include "drawing.e"
#include "dup.e"
#include "grid.e"
#include "group.e"
#include "mark.e"
#include "move.e"
#include "obj.e"
#include "raster.e"
#include "rect.e"
#include "ruler.e"
#include "stretch.e"
#include "setup.e"

#define FORWARD 0
#define REVERSE 1

int		selLtX, selLtY, selRbX, selRbY;
int		selObjLtX, selObjLtY, selObjRbX, selObjRbY;
struct SelRec	* topSel = NULL, * botSel = NULL;

static struct SelRec	* topCutSel = NULL, * botCutSel = NULL;

void CalcBBox (X1, Y1, X2, Y2, LtX, LtY, RbX, RbY)
   int X1, Y1, X2, Y2, * LtX, * LtY, * RbX, * RbY;
{
   if (X1 < X2)
      if (Y1 < Y2)
      {
         *LtX = X1; *LtY = Y1; *RbX = X2; *RbY = Y2;
      }
      else
      {
         *LtX = X1; *LtY = Y2; *RbX = X2; *RbY = Y1;
      }
   else
      if (Y1 < Y2)
      {
         *LtX = X2; *LtY = Y1; *RbX = X1; *RbY = Y2;
      }
      else
      {
         *LtX = X2; *LtY = Y2; *RbX = X1; *RbY = Y1;
      }
}

void RemoveAllSel ()
{
   while (topSel != NULL)
   {
      cfree (topSel);
      topSel = topSel->next;
   }
   botSel = NULL;
}

static
struct ObjRec * FindAnObj (X, Y)
   int	X, Y;
   /* X and Y are absolute coordinates */
{
   register struct ObjRec	* obj_ptr;

   for (obj_ptr = topObj; obj_ptr != NULL; obj_ptr = obj_ptr->next)
      if (X >= obj_ptr->bbox.ltx-(3<<zoomScale) &&
            X <= obj_ptr->bbox.rbx+(3<<zoomScale) &&
            Y >= obj_ptr->bbox.lty-(3<<zoomScale) &&
            Y <= obj_ptr->bbox.rby+(3<<zoomScale))
      {
         switch (obj_ptr->type)
         {
            case OBJ_TEXT: return (obj_ptr);
            case OBJ_BOX:
               if (FindGoodBox (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_OVAL:
               if (FindGoodOval (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_POLY:
               if (FindGoodPoly (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_POLYGON:
               if (FindGoodPolygon (X, Y, obj_ptr)) return (obj_ptr); break;

            case OBJ_GROUP:
            case OBJ_SYM:
            case OBJ_ICON:
               if (FindGoodObj (X, Y, obj_ptr->detail.r->first))
                  return (obj_ptr);
         }
      }
   return (NULL);
}

static
struct SelRec * AlreadySelected (ObjPtr)
   register struct ObjRec	* ObjPtr;
{
   register struct SelRec	* sel_ptr;

   for (sel_ptr = topSel; sel_ptr != NULL; sel_ptr = sel_ptr->next)
      if (sel_ptr->obj == ObjPtr)
         return (sel_ptr);
   return (NULL);
}

void AddSel (PrevPtr, NextPtr, SelPtr)
   struct SelRec	* PrevPtr, * NextPtr, * SelPtr;
   /* add SelPtr between PrevPtr and NextPtr */
{
   SelPtr->prev = PrevPtr;
   SelPtr->next = NextPtr;

   if (PrevPtr == NULL)
      topSel = SelPtr;
   else
      PrevPtr->next = SelPtr;

   if (NextPtr == NULL)
      botSel = SelPtr;
   else
      NextPtr->prev = SelPtr;
}

void AddNewSelObj (ObjPtr)
   register struct ObjRec	* ObjPtr;
{
   register struct ObjRec	* obj_ptr = topObj;
   register struct SelRec	* sel_ptr = topSel, * new_sel_ptr;

   new_sel_ptr = (struct SelRec *) calloc (1, sizeof(struct SelRec));
   new_sel_ptr->obj = ObjPtr;

   for ( ; sel_ptr != NULL && obj_ptr != ObjPtr; obj_ptr = obj_ptr->next)
      if (obj_ptr == sel_ptr->obj)
         sel_ptr = sel_ptr->next;

   if (sel_ptr == NULL)
   {  /* the object is below the last selected object */
      if (botSel == NULL)
         topSel = new_sel_ptr;
      else
         botSel->next = new_sel_ptr;

      new_sel_ptr->prev = botSel;
      new_sel_ptr->next = NULL;
      botSel = new_sel_ptr;
   }
   else
   {  /* the object is between sel_ptr and sel_ptr->prev */
      if (sel_ptr->prev == NULL)
         topSel = new_sel_ptr;
      else
         sel_ptr->prev->next = new_sel_ptr;

      new_sel_ptr->next = sel_ptr;
      new_sel_ptr->prev = sel_ptr->prev;
      sel_ptr->prev = new_sel_ptr;
   }
}

void UpdSelBBox ()
   /* update selLtX, selLtY, selRbX, selRbY */
{
   register struct ObjRec	* obj_ptr;
   register struct SelRec	* sel_ptr;

   if ((sel_ptr = topSel) == NULL) return;

   selLtX = sel_ptr->obj->bbox.ltx;
   selLtY = sel_ptr->obj->bbox.lty;
   selRbX = sel_ptr->obj->bbox.rbx;
   selRbY = sel_ptr->obj->bbox.rby;
   selObjLtX = sel_ptr->obj->obbox.ltx;
   selObjLtY = sel_ptr->obj->obbox.lty;
   selObjRbX = sel_ptr->obj->obbox.rbx;
   selObjRbY = sel_ptr->obj->obbox.rby;

   for (sel_ptr = topSel->next; sel_ptr != NULL; sel_ptr = sel_ptr->next)
   {
      obj_ptr = sel_ptr->obj;
      if (obj_ptr->bbox.ltx < selLtX) selLtX = obj_ptr->bbox.ltx;
      if (obj_ptr->bbox.lty < selLtY) selLtY = obj_ptr->bbox.lty;
      if (obj_ptr->bbox.rbx > selRbX) selRbX = obj_ptr->bbox.rbx;
      if (obj_ptr->bbox.rby > selRbY) selRbY = obj_ptr->bbox.rby;
      if (obj_ptr->obbox.ltx < selObjLtX) selObjLtX = obj_ptr->obbox.ltx;
      if (obj_ptr->obbox.lty < selObjLtY) selObjLtY = obj_ptr->obbox.lty;
      if (obj_ptr->obbox.rbx > selObjRbX) selObjRbX = obj_ptr->obbox.rbx;
      if (obj_ptr->obbox.rby > selObjRbY) selObjRbY = obj_ptr->obbox.rby;
   }
}
 
static
struct SelRec * SelectOneObj (X, Y)
   int	X, Y;
   /* X and Y are absolute coordinates */
{
   register struct ObjRec	* obj_ptr;

   RemoveAllSel ();
   if ((obj_ptr = FindAnObj (X, Y)) == NULL) return (NULL);

   topSel = (struct SelRec *) calloc (1, sizeof(struct SelRec));
   topSel->next = NULL;
   topSel->obj = obj_ptr;
   topSel->prev = NULL;
   botSel = topSel;
   UpdSelBBox ();

   return (topSel);
}

static
struct SelRec * FindObjects (X1, Y1, X2, Y2)
   int	X1, Y1, X2, Y2;
   /* X1, Y1, X2, Y2 are absolute coordinates */
{
   register struct ObjRec	* obj_ptr;
   register struct SelRec	* sel_ptr;

   RemoveAllSel ();

   for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
      if (X1 <= obj_ptr->bbox.ltx && X2 >= obj_ptr->bbox.rbx &&
            Y1 <= obj_ptr->bbox.lty && Y2 >= obj_ptr->bbox.rby)
      {
         sel_ptr = (struct SelRec *) calloc (1, sizeof(struct SelRec));
         sel_ptr->next = topSel;
         sel_ptr->obj = obj_ptr;
         sel_ptr->prev = NULL;
         if (topSel == NULL)
            botSel = sel_ptr;
         else
            topSel->prev = sel_ptr;
         topSel = sel_ptr;
      }

   UpdSelBBox ();
   return (topSel);
}
 
static XPoint sv[5];

void SelBox (window, gc, x1, y1, x2, y2)
   Window	window;
   GC		gc;
   int		x1, y1, x2, y2;
{
   sv[0].x = (short)x1; sv[0].y = (short)y1;
   sv[1].x = (short)x1; sv[1].y = (short)y2;
   sv[2].x = (short)x2; sv[2].y = (short)y2;
   sv[3].x = (short)x2; sv[3].y = (short)y1;
   sv[4].x = (short)x1; sv[4].y = (short)y1;
   XDrawLines (mainDisplay, window, gc, sv, 5, CoordModeOrigin);
}

static
struct ObjRec * PtInObjList (X, Y, FirstObjPtr)
   int			X, Y;
   struct ObjRec	* FirstObjPtr;
   /* X and Y are absolute coordinates */
{
   register struct ObjRec	* obj_ptr, * obj_ptr1;

   for (obj_ptr = FirstObjPtr; obj_ptr != NULL; obj_ptr = obj_ptr->next)
   {
      if (X >= obj_ptr->bbox.ltx-(3<<zoomScale) &&
            X <= obj_ptr->bbox.rbx+(3<<zoomScale) &&
            Y >= obj_ptr->bbox.lty-(3<<zoomScale) &&
            Y <= obj_ptr->bbox.rby+(3<<zoomScale))
      {
         switch (obj_ptr->type)
         {
            case OBJ_TEXT: return (obj_ptr);
            case OBJ_BOX:
               if (FindGoodBox (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_OVAL:
               if (FindGoodOval (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_POLY:
               if (FindGoodPoly (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_POLYGON:
               if (FindGoodPolygon (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_GROUP:
            case OBJ_SYM:
            case OBJ_ICON:
               obj_ptr1 = PtInObjList (X, Y, obj_ptr->detail.r->first);
               if (obj_ptr1 != NULL) return (obj_ptr1);
               break;
         }
      }
   }
   return (NULL);
}

static
struct ObjRec * PtInSelected (X, Y)
   int	X, Y;
   /* X and Y are absolute coordinates */
{
   register struct SelRec	* sel_ptr;
   register struct ObjRec	* obj_ptr, * obj_ptr1;

   for (sel_ptr = topSel; sel_ptr != NULL; sel_ptr = sel_ptr->next)
   {
      obj_ptr = sel_ptr->obj;
      if (X >= obj_ptr->bbox.ltx-(3<<zoomScale) &&
            X <= obj_ptr->bbox.rbx+(3<<zoomScale) &&
            Y >= obj_ptr->bbox.lty-(3<<zoomScale) &&
            Y <= obj_ptr->bbox.rby+(3<<zoomScale))
      {
         switch (obj_ptr->type)
         {
            case OBJ_TEXT: return (obj_ptr);
            case OBJ_BOX:
               if (FindGoodBox (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_OVAL:
               if (FindGoodOval (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_POLY:
               if (FindGoodPoly (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_POLYGON:
               if (FindGoodPolygon (X, Y, obj_ptr)) return (obj_ptr); break;
            case OBJ_GROUP:
            case OBJ_SYM:
            case OBJ_ICON:
               obj_ptr1 = PtInObjList (X, Y, obj_ptr->detail.r->first);
               if (obj_ptr1 != NULL) return (obj_ptr1);
               break;
         }
      }
   }
   return (NULL);
}

static
void ContinueSel (XOff, YOff)
   int 	XOff, YOff;
   /* XOff and YOff are screen offsets, and they are not on grid */
{
   Window	root_win, child_win;
   int		root_x, root_y;
   unsigned int	status;
   int 		end_x, end_y;
   XButtonEvent	input;
   int 		done = FALSE, ltx, lty, rbx, rby, dx, dy;
   int 		new_end_x, new_end_y;

   end_x = XOff;
   end_y = YOff; 

   SelBox (drawWindow, revDefaultGC, XOff, YOff, end_x, end_y);
   XGrabPointer (mainDisplay, drawWindow, FALSE, ButtonReleaseMask,
         GrabModeAsync, GrabModeAsync, None, handCursor, CurrentTime);
   
   while (!done)
   {
      XQueryPointer (mainDisplay, drawWindow, &root_win, &child_win,
            &root_x, &root_y, &new_end_x, &new_end_y, &status);

      if (new_end_x==end_x && new_end_y==end_y && (status & Button1Mask)!=0)
         continue;

      SelBox (drawWindow, revDefaultGC, XOff, YOff, end_x, end_y);
      end_x = new_end_x; end_y = new_end_y;

      SelBox (drawWindow, revDefaultGC, XOff, YOff, end_x, end_y);

      if ((status & Button1Mask) == 0)
      {
         XUngrabPointer (mainDisplay, CurrentTime);
         MarkRulers (end_x, end_y);
         SelBox (drawWindow, revDefaultGC, XOff, YOff, end_x, end_y);
         done = TRUE;
      }
      else
         MarkRulers (end_x, end_y);
   }

   if (topSel != NULL) HighLightReverse ();

   dx = abs (XOff - end_x);
   dy = abs (YOff - end_y);
   if (dx <= 2 && dy <= 2)
   {
      if (SelectOneObj ((XOff<<zoomScale)+drawOrigX,
            (YOff<<zoomScale)+drawOrigY) != NULL)
         HighLightForward ();
   }
   else if (XOff != end_x && YOff != end_y)
   {
      CalcBBox (XOff, YOff, end_x, end_y, &ltx, &lty, &rbx, &rby);
      if (FindObjects ((ltx<<zoomScale)+drawOrigX, (lty<<zoomScale)+drawOrigY,
            (rbx<<zoomScale)+drawOrigX, (rby<<zoomScale)+drawOrigY) != NULL)
         HighLightForward ();
   }
   else
      RemoveAllSel ();
}

void Select (input)
   XEvent	* input;
{
   Window		root_win, child_win;
   XButtonEvent		* button_ev;
   struct SelRec	* sel_ptr;
   struct ObjRec	* obj_ptr;
   int			mouse_x, mouse_y, grid_x, grid_y, corner;
   int			root_x, root_y;
   unsigned int		status;

   if (input->type != ButtonPress) return;

   button_ev = &(input->xbutton);
   if (button_ev->button == Button1)
   {
      XQueryPointer (mainDisplay, drawWindow, &root_win, &child_win,
            &root_x, &root_y, &mouse_x, &mouse_y, &status);
      if (button_ev->state & ShiftMask)
      {
         obj_ptr = FindAnObj ((mouse_x<<zoomScale)+drawOrigX,
               (mouse_y<<zoomScale)+drawOrigY);
         if (obj_ptr != NULL)
         {
            if ((sel_ptr = AlreadySelected (obj_ptr)) != NULL)
            {
               /* de-select an object */
               HighLightAnObj (obj_ptr);

               if (sel_ptr->prev == NULL)
                  topSel = sel_ptr->next;
               else
                  sel_ptr->prev->next = sel_ptr->next;

               if (sel_ptr->next == NULL)
                  botSel = sel_ptr->prev;
               else
                  sel_ptr->next->prev = sel_ptr->prev;

               cfree (sel_ptr);
            }
            else
            {
               /* add a newly selected object */
               AddNewSelObj (obj_ptr);
               HighLightAnObj (obj_ptr);
            }
            UpdSelBBox ();
         }
         return;
      }
      GridXY (mouse_x, mouse_y, &grid_x, &grid_y);
      if (topSel != NULL)
      {
         if ((sel_ptr = PtInSelMark (mouse_x, mouse_y, &corner)) != NULL)
         {
            StretchSel (grid_x, grid_y, sel_ptr->obj, corner);
            justDupped = FALSE;
            return;
         }
         else if ((obj_ptr = PtInSelected ((mouse_x<<zoomScale)+drawOrigX,
               (mouse_y<<zoomScale)+drawOrigY)) != NULL)
         {
            MoveSel (grid_x, grid_y, obj_ptr);
            return;
         }
      }
      ContinueSel (mouse_x, mouse_y);
      justDupped = FALSE;
   }
}

void SelAllObj ()
{
   register struct SelRec	* sel_ptr;
   register struct ObjRec	* obj_ptr;

   TieLooseEnds ();
   SetCurChoice (NOTHING);

   if (topSel != NULL)
   {
      HighLightReverse ();
      RemoveAllSel ();
   }
   for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      sel_ptr = (struct SelRec *) calloc (1, sizeof(struct SelRec));
      sel_ptr->next = topSel;
      sel_ptr->obj = obj_ptr;
      sel_ptr->prev = NULL;
      if (topSel == NULL)
         botSel = sel_ptr;
      else
         topSel->prev = sel_ptr;
      topSel = sel_ptr;
   }
   UpdSelBBox ();
   HighLightForward ();
   justDupped = FALSE;
}

static struct ObjRec	* tmpTopObj, * tmpBotObj;

static
void PushTmpObj (ObjPtr)
   struct ObjRec	* ObjPtr;
{
   ObjPtr->next = tmpTopObj;
   ObjPtr->prev = NULL;

   if (tmpBotObj == NULL)
      tmpBotObj = ObjPtr;
   else
      tmpTopObj->prev = ObjPtr;

   tmpTopObj = ObjPtr;
}

static
void BreakSel ()
   /* break off selected objects from the main stream objects           */
   /* when returns, tmpTopObj points to the top of the selected objects */
   /*    and tmpBotObj points to the bottom of the selected objects     */
{
   register struct SelRec	* sel_ptr;

   for (sel_ptr = botSel; sel_ptr != NULL; sel_ptr = sel_ptr->prev)
   {
      UnlinkObj (sel_ptr->obj);
      PushTmpObj (sel_ptr->obj);
   }
}

void MoveSelToTop ()
{
   if (topSel == NULL) return;

   tmpTopObj = tmpBotObj = NULL;

   BreakSel ();
   tmpBotObj->next = topObj;
   if (topObj == NULL)
      botObj = tmpBotObj;
   else
      topObj->prev = tmpBotObj;
   topObj = tmpTopObj;
}

void MoveSelToBot ()
{
   if (topSel == NULL) return;

   tmpTopObj = tmpBotObj = NULL;

   BreakSel ();
   tmpTopObj->prev = botObj;
   if (topObj == NULL)
      topObj = tmpTopObj;
   else
      botObj->next = tmpTopObj;
   botObj = tmpBotObj;
}

static
void DelCutSel ()
{
   register struct SelRec	* sel_ptr;

   if ((sel_ptr = topCutSel) == NULL) return;

   for ( ; sel_ptr != NULL; sel_ptr = sel_ptr->next)
   {
      DelObj (sel_ptr->obj);
      cfree (sel_ptr);
   }
   topCutSel = botCutSel = NULL;
}

void UndoDelete ()
{
   if (topCutSel == NULL) return;

   HighLightReverse ();
   RemoveAllSel ();

   topSel = topCutSel;
   botSel = botCutSel;

   botSel->obj->next = topObj;
   if (topObj == NULL)
      botObj = botSel->obj;
   else
      topObj->prev = botSel->obj;
   topObj = topSel->obj;

   topCutSel = botCutSel = NULL;

   UpdSelBBox ();
   RedrawAnArea (botObj, selLtX-(1<<zoomScale), selLtY-(1<<zoomScale),
         selRbX+(1<<zoomScale), selRbY+(1<<zoomScale));
   HighLightForward ();
   SetCurChoice (NOTHING);
   fileModified = TRUE;
   justDupped = FALSE;
}

void DelAllSelObj ()
{
   register struct SelRec	* sel_ptr;

   if (topSel == NULL) return;

   HighLightReverse ();
   tmpTopObj = tmpBotObj = NULL;
   BreakSel ();
   topCutSel = topSel;
   botCutSel = botSel;

   topSel = botSel = NULL;
   RedrawAnArea (botObj, selLtX-(1<<zoomScale), selLtY-(1<<zoomScale),
         selRbX+(1<<zoomScale), selRbY+(1<<zoomScale));
   fileModified = TRUE;
   justDupped = FALSE;
}

void GroupSelObj ()
{
   if (topSel == NULL) return; /* no object seleted */
   if (topSel == botSel) return; /* single object selected */

   tmpTopObj = tmpBotObj = NULL;

   HighLightReverse ();
   BreakSel ();

   CreateGroupObj (tmpTopObj, tmpBotObj);

   RemoveAllSel ();
   topSel = botSel = (struct SelRec *) calloc (1, sizeof(struct SelRec));
   topSel->obj = topObj;
   topSel->next = topSel->prev = NULL;

   RedrawAnArea (botObj, selLtX-(1<<zoomScale), selLtY-(1<<zoomScale),
         selRbX+(1<<zoomScale), selRbY+(1<<zoomScale));
   HighLightForward ();
   fileModified = TRUE;
   justDupped = FALSE;
}

void SelectTopObj ()
{
   if (topObj == NULL) return;

   topSel = botSel = (struct SelRec *) calloc (1, sizeof(struct SelRec));
   topSel->obj = topObj;
   topSel->next = topSel->prev = NULL;
   UpdSelBBox ();

   HighLightForward ();
   justDupped = FALSE;
}
