/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/special.c,v 1.1 90/04/01 22:16:42 william Exp $";
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "animate.e"
#include "attr.e"
#include "choice.e"
#include "color.e"
#include "cursor.e"
#include "dialog.e"
#include "drawing.e"
#include "dup.e"
#include "file.e"
#include "grid.e"
#include "group.e"
#include "mark.e"
#include "msg.e"
#include "move.e"
#include "names.e"
#include "obj.e"
#include "raster.e"
#include "select.e"
#include "setup.e"
#include "stk.e"

#define SPECIAL_SYM 0
#define SPECIAL_UNSYM 1
#define SPECIAL_INST 2
#define SPECIAL_ICON 3
#define SPECIAL_UNICON 4
#define SPECIAL_PUSH 5
#define SPECIAL_POP 6
#define SPECIAL_ADDATTR 7
#define SPECIAL_DETACHATTR 8
#define SPECIAL_SHOWATTRNAME 9
#define SPECIAL_HIDEATTRNAME 10
#define SPECIAL_MOVEATTR 11
#define SPECIAL_ANIMATESEND 12
#define SPECIAL_ANIMATEFLASH 13

#define MAXSPECIALMENUS 14
static char * specialMenuStr[] =
{
   "MakeSymbolic   ^#M",
   "UnMakeSymbolic ^#N",
   "Instantiate     ^I",
   "MakeIconic     ^#I",
   "UnMakeIconic   ^#J",
   "Push            ^V",
   "Pop             ^K",
   "AddAttr         #A",
   "DetachAttrs     #T",
   "ShowAttrName    #N",
   "HideAttrName    #J",
   "MoveAttr        #M",
   "AnimateSend     #E",
   "AnimateFlash    #F"
};

struct ObjRec * ReadSymbol (FP)
   FILE	* FP;
{
   struct ObjRec	* obj_ptr;
   struct AttrRec       * attr_ptr, * top_attr = NULL, * bot_attr = NULL;
   char			inbuf[255], obj_name[10];
   int			len;

   while (fgets (inbuf, 255, FP) != NULL)
   {
      len = strlen (inbuf) - 1;
      inbuf[len] = '\0';
      ParseStr (inbuf, '(', obj_name);
      if (strcmp (obj_name, "sym") == 0)
      {
         ReadGroupObj (FP, &obj_ptr);
         obj_ptr->type = OBJ_ICON;
         while (ReadAttr (FP, &attr_ptr))
         {
            attr_ptr->owner = obj_ptr;
            attr_ptr->prev = NULL;
            attr_ptr->next = top_attr;
            attr_ptr->inherited = TRUE;
            if (top_attr == NULL)
               bot_attr = attr_ptr;
            else
               top_attr->prev = attr_ptr;
            top_attr = attr_ptr;
         }
         if (bot_attr != NULL) bot_attr->next = NULL;
         obj_ptr->detail.r->fattr = top_attr;
         obj_ptr->detail.r->lattr = bot_attr;
         return (obj_ptr);
      }
      else if (strcmp (obj_name, "state") == 0)
      {
         importingFile = TRUE; /* ignore 'state' info but set fileVersion */
         ReadState (inbuf);
         importingFile = FALSE;
      }
   }
   return (NULL);
}

void PlaceTopObj ()
{
   int		x, y, grid_x, grid_y, dx, dy, placing = TRUE, moved;
   int		ltx, lty, rbx, rby, orig_x, orig_y;
   Window	root_win, child_win;
   int		root_x, root_y;
   unsigned int	status;
   XGCValues	values;

   while (TRUE)
   {
      XQueryPointer (mainDisplay, drawWindow, &root_win, &child_win,
            &root_x, &root_y, &ltx, &lty, &status);
      if (!(status & (Button1Mask | Button2Mask | Button3Mask))) break;
   }

   orig_x = ltx;
   orig_y = lty;
   rbx = ((topObj->obbox.rbx - topObj->obbox.ltx) >> zoomScale) + ltx;
   rby = ((topObj->obbox.rby - topObj->obbox.lty) >> zoomScale) + lty;
   SelBox (drawWindow, revDefaultGC, ltx, lty, rbx, rby);

   XGrabPointer (mainDisplay, drawWindow, FALSE, ButtonReleaseMask,
         GrabModeAsync, GrabModeAsync, None, handCursor, CurrentTime);

   dx = dy = 0;

   while (placing)
   {
      XQueryPointer (mainDisplay, drawWindow, &root_win, &child_win,
            &root_x, &root_y, &x, &y, &status);
      GridXY (x, y, &grid_x, &grid_y);

      if (moved = (grid_x != ltx+dx || grid_y != lty+dy))
      {
         SelBox (drawWindow, revDefaultGC, ltx+dx, lty+dy, rbx+dx, rby+dy);
         MarkRulers (grid_x, grid_y);
      }

      dx = grid_x - orig_x;
      dy = grid_y - orig_y;

      if (status & (Button1Mask | Button2Mask | Button3Mask))
      {
         XUngrabPointer (mainDisplay, CurrentTime);
         placing = FALSE;
         if (!moved)
         {
            SelBox (drawWindow, revDefaultGC, ltx+dx, lty+dy, rbx+dx, rby+dy);
            MarkRulers (grid_x, grid_y);
         }
         if (grid_x>=0 && grid_y>=0 && grid_x<drawWinW && grid_y<drawWinH)
         {
            dx = (grid_x << zoomScale) + drawOrigX - topObj->obbox.ltx;
            dy = (grid_y << zoomScale) + drawOrigY - topObj->obbox.lty;
            MoveObj (topObj, dx, dy);
            DrawObj (drawWindow, topObj);
            return;
         }
         else
         {
            FreeGroupObj (topObj);
            topObj = botObj = NULL;
         }
      }
      else if (moved)
         SelBox (drawWindow, revDefaultGC, ltx+dx, lty+dy, rbx+dx, rby+dy);
   }
}

void Instantiate ()
{
   char			file_name[80], sym_name[80], full_name[MAXPATHLENGTH];
   struct ObjRec	* obj_ptr;
   FILE			* fp;

   if (SelectFileName ("sym", "PLEASE SELECT A SYMBOL TO INSTANTIATE ...",
         sym_name) == INVALID) return;

   TieLooseEnds ();
   SetCurChoice (NOTHING);

   sprintf (file_name, "%s.sym", sym_name);
   strcpy (full_name, curDomainName);
   if (*curDomainName != '\0') strcat (full_name, "/");
   strcat (full_name, file_name);
   if ((fp = fopen (full_name, "r")) == NULL)
   { printf ("Can not open %s\n", file_name); return; }

   if ((obj_ptr = ReadSymbol (fp)) != NULL)
   {
      if (topSel != NULL)
      {
         HighLightReverse ();
         RemoveAllSel ();
      }
      SetNullCursor (drawWindow);

      obj_ptr->detail.r->id = objId++;
      obj_ptr->detail.r->dirty = FALSE;
      strcpy (obj_ptr->detail.r->s, sym_name);
      AddObj (NULL, topObj, obj_ptr);
      AdjObjBBox (obj_ptr);

      PlaceTopObj ();
      if (topObj != NULL) SelectTopObj ();
      ShowCursor ();
      fileModified = TRUE;
      justDupped = FALSE;
   }
   fclose (fp);
}

void MakeSymbolic ()
{
   if (topSel!=NULL && topSel==botSel &&
      (topSel->obj->type==OBJ_GROUP || topSel->obj->type==OBJ_ICON))
   {
      HighLightReverse ();
      topSel->obj->type = OBJ_SYM;
      Msg ("Selected GROUP object is now SYMBOLIC.");
      fileModified = TRUE;
      AdjObjBBox (topSel->obj);
      UpdSelBBox ();
      RedrawAnArea (botObj,
            botSel->obj->obbox.ltx-QUARTER_INCH-(1<<zoomScale),
            botSel->obj->obbox.lty-QUARTER_INCH-(1<<zoomScale),
            botSel->obj->obbox.rbx+QUARTER_INCH+(1<<zoomScale),
            botSel->obj->obbox.rby+QUARTER_INCH+(1<<zoomScale));
      HighLightForward ();
   }
   else
      Msg ("Please select one GROUP or ICON object to make it Symbolic.");
}

void UnMakeSymbolic ()
{
   register struct ObjRec	* obj_ptr;
   register int			ltx, lty, rbx, rby;
   struct SelRec		* sel_ptr = topSel;
   int				modified = FALSE;

   for ( ; sel_ptr != NULL; sel_ptr = sel_ptr->next)
   {
      obj_ptr = sel_ptr->obj;
      if (obj_ptr->type == OBJ_SYM)
      {
         obj_ptr->type = OBJ_GROUP;
         AdjObjBBox (obj_ptr);
         if (modified)
         {
            if (ltx < obj_ptr->bbox.ltx) ltx = obj_ptr->bbox.ltx;
            if (lty < obj_ptr->bbox.lty) lty = obj_ptr->bbox.lty;
            if (rbx > obj_ptr->bbox.rbx) rbx = obj_ptr->bbox.rbx;
            if (rby > obj_ptr->bbox.rby) rby = obj_ptr->bbox.rby;
         }
         else
         {
            ltx = obj_ptr->bbox.ltx; lty = obj_ptr->bbox.lty;
            rbx = obj_ptr->bbox.rbx; rby = obj_ptr->bbox.rby;
            modified = TRUE;
         }
      }
   }
   if (modified)
   {
      HighLightReverse ();
      fileModified = TRUE;
      UpdSelBBox ();
      RedrawAnArea (botObj,
            ltx-QUARTER_INCH-(1<<zoomScale), lty-QUARTER_INCH-(1<<zoomScale),
            rbx+QUARTER_INCH+(1<<zoomScale), rby+QUARTER_INCH+(1<<zoomScale));
      HighLightForward ();
   }
}

void MakeIconic ()
{
   char 		icon_name[80], file_name[80], s[255];
   FILE			* fp;
   struct ObjRec	* saved_obj_ptr;

   if (topSel!=NULL && topSel==botSel && topSel->obj->type==OBJ_GROUP)
   {
      Dialog ("Please Enter Name of the Icon:", icon_name);
      if (*icon_name == '\0')
      {
         Msg ("Name not specified, icon not created.");
         return;
      }
      if (*curDomainName != '\0')
         sprintf (file_name, "%s/%s.sym", curDomainName, icon_name);
      else
         sprintf (file_name, "%s.sym", icon_name);
      if (!OkayToCreateFile (file_name)) return;
      if ((fp = fopen (file_name, "w")) == NULL)
      {
         sprintf (s, "Can not create '%s', icon not created.", file_name);
         Msg (s);
         return;
      }
      sprintf (s, "Creating '%s' ...", file_name);
      Msg (s);
      topSel->obj->type = OBJ_SYM;
      strcpy (topSel->obj->detail.r->s, icon_name);

      saved_obj_ptr = topSel->obj->prev;
      topSel->obj->prev = NULL;
      Save (fp, topSel->obj, 0);
      topSel->obj->prev = saved_obj_ptr;

      sprintf (s, "File '%s' created.", file_name);
      Msg (s);
      fclose (fp);

      HighLightReverse ();
      topSel->obj->type = OBJ_ICON;
      topSel->obj->detail.r->id = objId++;
      Msg ("Selected GROUP object is now ICONIC.");
      fileModified = TRUE;
      AdjObjBBox (topSel->obj);
      UpdSelBBox ();
      RedrawAnArea (botObj,
            botSel->obj->bbox.ltx-(1<<zoomScale),
            botSel->obj->bbox.lty-(1<<zoomScale),
            botSel->obj->bbox.rbx+(1<<zoomScale),
            botSel->obj->bbox.rby+(1<<zoomScale));
      HighLightForward ();
   }
   else
      Msg ("Please select one GROUP object to make it ICONIC.");
}

void UnMakeIconic ()
{
   register struct ObjRec	* obj_ptr;
   struct SelRec		* sel_ptr = topSel;
   int				modified = FALSE;

   HighLightReverse ();
   for ( ; sel_ptr != NULL; sel_ptr = sel_ptr->next)
   {
      obj_ptr = sel_ptr->obj;
      if (obj_ptr->type == OBJ_ICON)
      {
         obj_ptr->type = OBJ_GROUP;
         AdjObjBBox (obj_ptr);
         modified = TRUE;
      }
   }
   if (modified)
   {
      Msg ("Selected ICONIC objects are GROUP objects now.");
      fileModified = TRUE;
      UpdSelBBox ();
      RedrawAnArea (botObj, selLtX-(1<<zoomScale), selLtY-(1<<zoomScale),
            selRbX+(1<<zoomScale), selRbY+(1<<zoomScale));
   }
   HighLightForward ();
}

void SpecialMenu (X, Y)
   int	X, Y;
{
   register int	index;
   char		s[80];
   int		* fore_colors, * valid;

   DefaultColorArrays (MAXSPECIALMENUS, &fore_colors, &valid);
   index = TextMenuLoop (X, Y, specialMenuStr, MAXSPECIALMENUS, fore_colors,
         valid, SINGLECOLOR);
   switch (index)
   {
      case SPECIAL_SYM: MakeSymbolic (); break;
      case SPECIAL_UNSYM: UnMakeSymbolic (); break;
      case SPECIAL_INST: Instantiate (); break;
      case SPECIAL_ICON: MakeIconic (); break;
      case SPECIAL_UNICON: UnMakeIconic (); break;
      case SPECIAL_PUSH: PushIcon (); break;
      case SPECIAL_POP: PopIcon (); break;
      case SPECIAL_ADDATTR: AddAttrs (); break;
      case SPECIAL_SHOWATTRNAME: ShowAllAttrNames (); break;
      case SPECIAL_HIDEATTRNAME: HideAllAttrNames (); break;
      case SPECIAL_DETACHATTR: DetachAttrs (); break;
      case SPECIAL_MOVEATTR: MoveAttr (); break;
      case SPECIAL_ANIMATESEND: AnimateSel (); break;
      case SPECIAL_ANIMATEFLASH: FlashSelColor (); break;
   }
}
