/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/stk.c,v 1.1 90/04/01 22:16:44 william Exp $";
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "align.e"
#include "button.e"
#include "choice.e"
#include "color.e"
#include "dup.e"
#include "file.e"
#include "font.e"
#include "grid.e"
#include "mark.e"
#include "menu.e"
#include "msg.e"
#include "names.e"
#include "obj.e"
#include "pattern.e"
#include "ruler.e"
#include "scroll.e"
#include "select.e"
#include "setup.e"
#include "text.e"

static struct StkRec	* topStk = NULL;

int AncesterModified ()
{
   struct StkRec	* stk_ptr;

   for (stk_ptr = topStk; stk_ptr != NULL; stk_ptr = stk_ptr->next)
      if (stk_ptr->file_mod) return (TRUE);
   return (FALSE);
}

void PushIcon ()
{
   struct StkRec	* stk_ptr;
   struct ObjRec	* obj_ptr;
   char			file_name[80], s[80], full_name[MAXPATHLENGTH];
   FILE			* fp;

   if (topSel == NULL || topSel != botSel || topSel->obj->type != OBJ_ICON)
   {
      Msg ("Please select one ICON object to push into.");
      return;
   }

   sprintf (file_name, "%s.sym", topSel->obj->detail.r->s);
   strcpy (full_name, curDomainName);
   if (*curDomainName != '\0') strcat (full_name, "/");
   strcat (full_name, file_name);

   if ((fp = fopen (full_name, "r")) == NULL)
   {
      sprintf (s, "Can not open '%s', icon not pushed into.", file_name);
      Msg (s);
      return;
   }

   HighLightReverse ();

   stk_ptr = (struct StkRec *) calloc (1, sizeof(struct StkRec));
   stk_ptr->next = topStk;
   stk_ptr->sel = topSel->obj;
   stk_ptr->first = topObj;
   stk_ptr->last = botObj;
   stk_ptr->file_mod = fileModified;
   stk_ptr->id = objId;
   stk_ptr->page_style = pageStyle;

   stk_ptr->orig_x = drawOrigX;
   stk_ptr->orig_y = drawOrigY;
   stk_ptr->zoom = zoomScale;
   stk_ptr->grid = xyGrid;
   stk_ptr->grid_on = gridOn;
   stk_ptr->color = colorIndex;
   stk_ptr->h_align = horiAlign;
   stk_ptr->v_align = vertAlign;
   stk_ptr->line_w = lineWidth;
   stk_ptr->line_s = lineStyle;
   stk_ptr->fill = objFill;
   stk_ptr->pen = penPat;
   stk_ptr->just = textJust;
   stk_ptr->font = curFont;
   stk_ptr->font_s = curStyle;
   stk_ptr->f_size = curSize;

   topObj = botObj = NULL;
   if (stk_ptr->name_valid = curFileDefined)
      strcpy (stk_ptr->name, curFileName);
   strcpy (stk_ptr->domain, curDomainName);
   topStk = stk_ptr;

   sprintf (s, "Pushing into '%s' ...", file_name);
   Msg (s);
   CleanUpDrawingWindow ();
   XClearWindow (mainDisplay, drawWindow);

   while (ReadObj (fp, &obj_ptr))
      if (obj_ptr != NULL)
      {
         AddObj (NULL, topObj, obj_ptr);
         if (PointInBBox (obj_ptr->x, obj_ptr->y, drawWinBBox) ||
               BBoxIntersect (obj_ptr->bbox, drawWinBBox))
            DrawObj (drawWindow, obj_ptr);
      }

   if (topObj == NULL)
   {
      DrawPaperBoundary ();
      RedrawGridLines ();
   }

   fclose (fp);
   strcpy (curFileName, file_name);
   curFileDefined = TRUE;

   fileModified = FALSE;
   sprintf (s, "Current file is '%s'.", file_name);
   Msg (s);

   RedrawTitleWindow ();
}

void PopIcon ()
{
   char	dummy[80];

   if (topStk == NULL)
   {
      Msg ("Already at top level.");
      return;
   }
   if (fileModified)
   {
      switch (YesNoCancel ("File modified, save file before quit? [ync](y)"))
      {
         case CONFIRM_YES: SaveFile (); break;
         case CONFIRM_NO: break;
         case CONFIRM_CANCEL: return;
      }
   }
   fileModified = topStk->file_mod;
   objId = topStk->id;

   CleanUpDrawingWindow ();

   drawOrigX = topStk->orig_x;
   drawOrigY = topStk->orig_y;
   zoomScale = topStk->zoom;
   xyGrid = topStk->grid;
   gridOn = topStk->grid_on;
   colorIndex = topStk->color;
   horiAlign = topStk->h_align;
   vertAlign = topStk->v_align;
   lineWidth = topStk->line_w;
   lineStyle = topStk->line_s;
   objFill = topStk->fill;
   penPat = topStk->pen;
   textJust = topStk->just;
   curFont = topStk->font;
   curStyle = topStk->font_s;
   curSize = topStk->f_size;

   topObj = topStk->first;
   botObj = topStk->last;
   if (curFileDefined = topStk->name_valid)
   {
      strcpy (curFileName, topStk->name);
      sprintf (dummy, "Poping back to '%s'.", curFileName);
      Msg (dummy);
   }
   else
   {
      sprintf (dummy, "Poping back to parent level.  Current file undefined.");
      Msg (dummy);
   }
   strcpy (curDomainName, topStk->domain);
   topSel = botSel = (struct SelRec *) calloc (1, sizeof(struct SelRec));
   topSel->next = NULL;
   topSel->prev = NULL;
   topSel->obj = topStk->sel;
   UpdSelBBox ();

   if (pageStyle != topStk->page_style) UpdPageStyle (topStk->page_style);
   UpdDrawWinWH ();
   SetCanvasFont ();
   RedrawRulers ();
   if (pageStyle == topStk->page_style) RedrawScrollBars ();
   RedrawChoiceWindow ();
   RedrawTitleWindow ();
   UpdDrawWinBBox ();

   cfree (topStk);
   topStk = topStk->next;
   ClearAndRedrawDrawWindow ();
   XSync (mainDisplay, TRUE);
   justDupped = FALSE;
}
