%
% Consult a '.obj' file or a '.sym' file, then try to trace the following
%	code.
%
% @(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/tgif.pl,v 1.2 90/05/10 15:22:58 william Exp $
%

tgif_state(FileVersion) :-
	%	Note:  This works for FileVersion>=6.
	state(_PageStyle,FileVersion,_OrigX,_OrigY,_Zoom,_GridSize,_Grid,
		_Color,_HoriAlign,_VertAlign,_LineWidth,_Spline,_LineStyle,
		_ObjFill,_PenPat,_TextJust,_Font,_TextStyle,_TextSize).

tgif_text(FileVersion, Obj) :-
	FileVersion >= 6,
	(	var(Obj) ->
		Obj = text(_Color,_X,_Y,_Font,_TextStyle,_TextSize,_NumLines,
			_TextJust,_TextRotate,_PenPat,StrList),
		call(Obj)
	;	Obj = text(_Color,_X,_Y,_Font,_TextStyle,_TextSize,_NumLines,
			_TextJust,_TextRotate,_PenPat,StrList)
	),
	tgif_strs(FileVersion, StrList).

tgif_box(FileVersion, Obj) :-
	FileVersion >= 6,
	(	var(Obj) ->
		Obj = box(_Color,_X1,_Y1,_X2,_Y2,_ObjFill,_LineWidth,_PenPat),
		call(Obj)
	;	Obj = box(_Color,_X1,_Y1,_X2,_Y2,_ObjFill,_LineWidth,_PenPat)
	).

tgif_oval(FileVersion, Obj) :-
	FileVersion >= 6,
	(	var(Obj) ->
		Obj = oval(_Color,_LeftTopX,_LeftTopY,_RightBotX,_RightBotY,
			_ObjFill,_LineWidth,_PenPat),
		call(Obj)
	;	Obj = oval(_Color,_LeftTopX,_LeftTopY,_RightBotX,_RightBotY,
			_ObjFill,_LineWidth,_PenPat)
	).

tgif_poly(FileVersion, Obj) :-
	FileVersion >= 6,
	(	var(Obj) ->
		Obj = poly(_Color,_NumVs,_Vs,_LineStyle,_LineWidth,_PenPat,_Id,
			_Spline,_ObjFill,AttrList),
		call(Obj)
	;	Obj = poly(_Color,_NumVs,_Vs,_LineStyle,_LineWidth,_PenPat,_Id,
			_Spline,_ObjFill,AttrList)
	),
	tgif_attrs(FileVersion, AttrList).

tgif_polygon(FileVersion, Obj) :-
	FileVersion >= 6,
	(	var(Obj) ->
		Obj = polygon(_Color,_NumVs,_Vs,_ObjFill,_LineWidth,_PenPat,
			_Spline),
		call(Obj)
	;	Obj = polygon(_Color,_NumVs,_Vs,_ObjFill,_LineWidth,_PenPat,
			_Spline)
	).

tgif_group(FileVersion, Obj) :-
	FileVersion >= 6,
	(	var(Obj) ->
		Obj = group(ObjList,AttrList),
		call(Obj)
	;	Obj = group(ObjList,AttrList)
	),
	tgif_objs(FileVersion, ObjList),
	tgif_attrs(FileVersion, AttrList).

tgif_sym(FileVersion, Obj) :-
	FileVersion >= 6,
	(	var(Obj) ->
		Obj = sym(ObjList,AttrList),
		call(Obj)
	;	Obj = sym(ObjList,AttrList)
	),
	tgif_objs(FileVersion, ObjList),
	tgif_attrs(FileVersion, AttrList).

tgif_icon(FileVersion, Obj) :-
	FileVersion >= 6,
	(	var(Obj) ->
		Obj = icon(ObjList,_Name,_Id,AttrList),
		call(Obj)
	;	Obj = icon(ObjList,_Name,_Id,AttrList)
	),
	tgif_objs(FileVersion, ObjList),
	tgif_attrs(FileVersion, AttrList).

% ----------------------------- support routines --------------------------- %

tgif_strs(_FileVersion, []) :- !.
tgif_strs(FileVersion, [_Str|Strs]) :-
	FileVersion >= 6,
	!, tgif_strs(FileVersion, Strs).

tgif_objs(_FileVersion, []) :- !.
tgif_objs(FileVersion, [Obj|Objs]) :-
	FileVersion >= 6,
	!,
	functor(Obj,Functor,_Arity),
	(	Functor == text ->
		tgif_text(FileVersion, Obj)
	;	Functor == box ->
		tgif_box(FileVersion, Obj)
	;	Functor == oval ->
		tgif_oval(FileVersion, Obj)
	;	Functor == poly ->
		tgif_poly(FileVersion, Obj)
	;	Functor == polygon ->
		tgif_polygon(FileVersion, Obj)
	;	Functor == group ->
		tgif_group(FileVersion, Obj)
	;	Functor == sym ->
		tgif_sym(FileVersion, Obj)
	;	Functor == icon ->
		tgif_icon(FileVersion, Obj)
	;	true
	),
	!, tgif_objs(FileVersion, Objs).

tgif_attrs(_FileVersion, []) :- !.
tgif_attrs(FileVersion, [Attr|Attrs]) :-
	FileVersion >= 6,
	!,
	Attr = attr(_Name,_Value,_ShowAll,_ShowNameOnly,_Inherited,_TextObj),
	!, tgif_attrs(FileVersion, Attrs).
