/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, William Cheng.
 *
 * @(#)$Header: /n/kona/u/tangram/u/william/X11/TGIF/RCS/types.h,v 1.1 90/04/01 22:17:28 william Exp $
 */

typedef struct BBRec {
   int	ltx, lty, rbx, rby;
} * BBRecPtr;

typedef struct PtRec {
   int		x, y;
   struct PtRec	* next;
} * PtRecPtr;

typedef struct ObjRec {
   int			x, y, type, color;
   struct BBRec		obbox;	/* object bounding box */
   struct BBRec		bbox;	/* real bounding box */
   struct ObjRec	* next, * prev;
   union {
      struct GroupRec	* r;
      struct PolyRec	* p;
      struct PolygonRec	* g;
      struct BoxRec	* b;
      struct OvalRec	* o;
      struct TextRec	* t;
   } detail;
} * ObjRecPtr;

typedef struct AttrRec {
   char			name[80]; /* attribute name */
   char			s[80]; /* attribute value string */
   short		shown; /* TRUE if the attribute is shown */
   short		nameshown; /* TRUE if the attr name is also shown */
   short		inherited; /* TRUE if attr was inherited */
   struct ObjRec	* obj; /* the OBJ_TEXT object that represent the attr */
   struct ObjRec        * owner; /* ptr to owner obj of the record */
   struct AttrRec	* next, * prev; /* next and prev attributes */
} * AttrRecPtr;

typedef struct GroupRec {
   struct ObjRec	* first, * last;
   char			s[80];
   int			id, dirty;
   struct AttrRec	* fattr, *lattr; /* first and last attributes */
			/* OBJ_GROUP, OBJ_SYM, OBJ_ICON can have attr */
} * GroupRecPtr;

typedef struct PolyRec {
   int			n;	/* number of points in the polyline */
   XPoint		* vlist;
   int			sn;	/* number of points in the spline polyline */
   XPoint		* svlist;
   int			style, width, pen, curved, id, fill;
   struct AttrRec	* fattr, *lattr; /* OBJ_POLY can have attr */
} * PolyRecPtr;

typedef struct PolygonRec {
   int		n;	/* number of points in the polygon */
   XPoint	* vlist;
   int		sn;	/* number of points in the spline polygon */
   XPoint	* svlist;
   int		fill, width, pen, curved;
} * PolygonRecPtr;

typedef struct BoxRec {
   int	fill, width, pen;
} * BoxRecPtr;

typedef struct OvalRec {
   int	fill, width, pen;
} * OvalRecPtr;

typedef struct StrRec {
   char			s[80];
   struct StrRec	* next, * prev;
} * StrRecPtr;

typedef struct TextRec {
   int			just;
   int			lines;
   int			font, style, size, rotate, pen;
   struct AttrRec       * attr; /* ptr to attr record if text obj is an attr */
   struct StrRec	* first, * last;
} * TextRecPtr;

typedef struct SelRec {
   struct ObjRec	* obj;
   struct SelRec	* next, * prev;
} * SelRecPtr;

typedef struct StkRec {
   struct ObjRec	* first, * last, * sel;
   struct StkRec	* next;
   int			name_valid, file_mod, id, page_style;
   int			orig_x, orig_y, zoom, grid, grid_on;
   int			color, h_align, v_align, line_w, line_s;
   int			fill, pen, just, font, font_s, f_size, f_comp_size;
   char			name[80], domain[256];
} * StkRecPtr;

typedef struct _DspItem {
   char		s[256];
} DspItem;

typedef struct _DspList {
   char			itemstr[256];
   struct _DspList	* next;
} DspList;
